@@ -25,7 +25,6 @@ import (
 	"encoding/base64"
 	"errors"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"regexp"
@@ -137,7 +136,7 @@ func (dbs *databaseSuite) TestImportKey(c *C) {
 	c.Check(info.Mode().Perm(), Equals, os.FileMode(0600)) // secret
 	// too much "clear box" testing? ok at least until we have
 	// more functionality
-	privKey, err := ioutil.ReadFile(keyPath)
+	privKey, err := os.ReadFile(keyPath)
 	c.Assert(err, IsNil)
 
 	privKeyFromDisk, err := asserts.DecodePrivateKeyInTest(privKey)
@@ -1576,6 +1575,80 @@ func (safs *signAddFindSuite) TestFindSequence(c *C) {
 
 }
 
+func (safs *signAddFindSuite) TestCheckConstraints(c *C) {
+	headers := map[string]interface{}{
+		"type":         "account",
+		"authority-id": "canonical",
+		"account-id":   "my-brand",
+		"display-name": "My Brand",
+		"validation":   "verified",
+		"timestamp":    time.Now().Format(time.RFC3339),
+	}
+	acct, err := safs.signingDB.Sign(asserts.AccountType, headers, nil, safs.signingKeyID)
+	c.Assert(err, IsNil)
+
+	err = safs.db.Add(acct)
+	c.Check(err, IsNil)
+
+	pubKey1 := testPrivKey1.PublicKey()
+	pubKey1Encoded, err := asserts.EncodePublicKey(pubKey1)
+	c.Assert(err, IsNil)
+
+	now := time.Now().UTC()
+	headers = map[string]interface{}{
+		"authority-id":        "canonical",
+		"format":              "1",
+		"account-id":          "my-brand",
+		"public-key-sha3-384": pubKey1.ID(),
+		"name":                "default",
+		"since":               now.Format(time.RFC3339),
+		"until":               now.AddDate(1, 0, 0).Format(time.RFC3339),
+		"constraints": []interface{}{
+			map[string]interface{}{
+				"headers": map[string]interface{}{
+					"type":  "model",
+					"model": "foo-.*",
+				},
+			},
+		},
+	}
+	accKey, err := safs.signingDB.Sign(asserts.AccountKeyType, headers, []byte(pubKey1Encoded), safs.signingKeyID)
+	c.Assert(err, IsNil)
+
+	err = safs.db.Add(accKey)
+	c.Check(err, IsNil)
+
+	headers = map[string]interface{}{
+		"type":         "model",
+		"authority-id": "my-brand",
+		"brand-id":     "my-brand",
+		"series":       "16",
+		"model":        "foo-200",
+		"classic":      "true",
+		"timestamp":    now.Format(time.RFC3339),
+	}
+	mfoo, err := asserts.AssembleAndSignInTest(asserts.ModelType, headers, nil, testPrivKey1)
+	c.Assert(err, IsNil)
+
+	err = safs.db.Add(mfoo)
+	c.Check(err, IsNil)
+
+	headers = map[string]interface{}{
+		"type":         "model",
+		"authority-id": "my-brand",
+		"brand-id":     "my-brand",
+		"series":       "16",
+		"model":        "goo-200",
+		"classic":      "true",
+		"timestamp":    now.Format(time.RFC3339),
+	}
+	mnotfoo, err := asserts.AssembleAndSignInTest(asserts.ModelType, headers, nil, testPrivKey1)
+	c.Assert(err, IsNil)
+
+	err = safs.db.Add(mnotfoo)
+	c.Check(err, ErrorMatches, `assertion does not match signing constraints for public key ".*" from "my-brand"`)
+}
+
 type revisionErrorSuite struct{}
 
 func (res *revisionErrorSuite) TestErrorText(c *C) {
