@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2017 Canonical Ltd
+ * Copyright (C) 2015-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -118,43 +118,14 @@ func (s *attrConstraintsSuite) TestSimple(c *C) {
 	c.Check(err, ErrorMatches, `attribute "bar" has constraints but is unset`)
 }
 
-func (s *attrConstraintsSuite) TestSimpleAnchorsVsRegexpAlt(c *C) {
+func (s *attrConstraintsSuite) TestMissingCheck(c *C) {
 	m, err := asserts.ParseHeaders([]byte(`attrs:
-  bar: BAR|BAZ`))
+  foo: $MISSING`))
 	c.Assert(err, IsNil)
 
 	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
 	c.Assert(err, IsNil)
-
-	plug := attrerObject(map[string]interface{}{
-		"bar": "BAR",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, IsNil)
-
-	plug = attrerObject(map[string]interface{}{
-		"bar": "BARR",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `attribute "bar" value "BARR" does not match \^\(BAR|BAZ\)\$`)
-
-	plug = attrerObject(map[string]interface{}{
-		"bar": "BBAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `attribute "bar" value "BAZZ" does not match \^\(BAR|BAZ\)\$`)
-
-	plug = attrerObject(map[string]interface{}{
-		"bar": "BABAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `attribute "bar" value "BABAZ" does not match \^\(BAR|BAZ\)\$`)
-
-	plug = attrerObject(map[string]interface{}{
-		"bar": "BARAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `attribute "bar" value "BARAZ" does not match \^\(BAR|BAZ\)\$`)
+	c.Check(asserts.RuleFeature(cstrs, "dollar-attr-constraints"), Equals, true)
 }
 
 func (s *attrConstraintsSuite) TestNested(c *C) {
@@ -207,111 +178,6 @@ baz: BAZ
 	c.Check(err, ErrorMatches, `attribute "bar\.bar2" must be a scalar or list`)
 }
 
-func (s *attrConstraintsSuite) TestAlternative(c *C) {
-	m, err := asserts.ParseHeaders([]byte(`attrs:
-  -
-    foo: FOO
-    bar: BAR
-  -
-    foo: FOO
-    bar: BAZ`))
-	c.Assert(err, IsNil)
-
-	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].([]interface{}))
-	c.Assert(err, IsNil)
-
-	plug := attrerObject(map[string]interface{}{
-		"foo": "FOO",
-		"bar": "BAR",
-		"baz": "BAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, IsNil)
-
-	plug = attrerObject(map[string]interface{}{
-		"foo": "FOO",
-		"bar": "BAZ",
-		"baz": "BAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, IsNil)
-
-	plug = attrerObject(map[string]interface{}{
-		"foo": "FOO",
-		"bar": "BARR",
-		"baz": "BAR",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `no alternative matches: attribute "bar" value "BARR" does not match \^\(BAR\)\$`)
-}
-
-func (s *attrConstraintsSuite) TestNestedAlternative(c *C) {
-	m, err := asserts.ParseHeaders([]byte(`attrs:
-  foo: FOO
-  bar:
-    bar1: BAR1
-    bar2:
-      - BAR2
-      - BAR22`))
-	c.Assert(err, IsNil)
-
-	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
-	c.Assert(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: FOO
-bar:
-  bar1: BAR1
-  bar2: BAR2
-`), nil)
-	c.Check(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: FOO
-bar:
-  bar1: BAR1
-  bar2: BAR22
-`), nil)
-	c.Check(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: FOO
-bar:
-  bar1: BAR1
-  bar2: BAR3
-`), nil)
-	c.Check(err, ErrorMatches, `no alternative for attribute "bar\.bar2" matches: attribute "bar\.bar2" value "BAR3" does not match \^\(BAR2\)\$`)
-}
-
-func (s *attrConstraintsSuite) TestAlternativeMatchingStringList(c *C) {
-	toMatch := attrs(`
-write:
- - /var/tmp
- - /var/lib/snapd/snapshots
-`)
-	m, err := asserts.ParseHeaders([]byte(`attrs:
-  write: /var/(tmp|lib/snapd/snapshots)`))
-	c.Assert(err, IsNil)
-
-	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
-	c.Assert(err, IsNil)
-
-	err = cstrs.Check(toMatch, nil)
-	c.Check(err, IsNil)
-
-	m, err = asserts.ParseHeaders([]byte(`attrs:
-  write:
-    - /var/tmp
-    - /var/lib/snapd/snapshots`))
-	c.Assert(err, IsNil)
-
-	cstrsLst, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
-	c.Assert(err, IsNil)
-
-	err = cstrsLst.Check(toMatch, nil)
-	c.Check(err, IsNil)
-}
-
 func (s *attrConstraintsSuite) TestAlternativeMatchingComplex(c *C) {
 	toMatch := attrs(`
 mnt: [{what: "/dev/x*", where: "/foo/*", options: ["rw", "nodev"]}, {what: "/bar/*", where: "/baz/*", options: ["rw", "bind"]}]
@@ -390,13 +256,6 @@ foo: 1
 bar: true
 `), nil)
 	c.Check(err, IsNil)
-
-	plug := attrerObject(map[string]interface{}{
-		"foo": int64(1),
-		"bar": true,
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, IsNil)
 }
 
 func (s *attrConstraintsSuite) TestCompileErrors(c *C) {
@@ -405,16 +264,6 @@ func (s *attrConstraintsSuite) TestCompileErrors(c *C) {
 	})
 	c.Check(err, ErrorMatches, `cannot compile "foo" constraint "\[": error parsing regexp:.*`)
 
-	_, err = asserts.CompileAttributeConstraints(map[string]interface{}{
-		"foo": []interface{}{"foo", "["},
-	})
-	c.Check(err, ErrorMatches, `cannot compile "foo/alt#2/" constraint "\[": error parsing regexp:.*`)
-
-	_, err = asserts.CompileAttributeConstraints(map[string]interface{}{
-		"foo": []interface{}{"foo", []interface{}{"bar", "baz"}},
-	})
-	c.Check(err, ErrorMatches, `cannot nest alternative constraints directly at "foo/alt#2/"`)
-
 	_, err = asserts.CompileAttributeConstraints("FOO")
 	c.Check(err, ErrorMatches, `first level of non alternative constraints must be a set of key-value contraints`)
 
@@ -437,45 +286,6 @@ func (s *attrConstraintsSuite) TestCompileErrors(c *C) {
 	}
 }
 
-func (s *attrConstraintsSuite) TestMatchingListsSimple(c *C) {
-	m, err := asserts.ParseHeaders([]byte(`attrs:
-  foo: /foo/.*`))
-	c.Assert(err, IsNil)
-
-	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
-	c.Assert(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: ["/foo/x", "/foo/y"]
-`), nil)
-	c.Check(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: ["/foo/x", "/foo"]
-`), nil)
-	c.Check(err, ErrorMatches, `attribute "foo\.1" value "/foo" does not match \^\(/foo/\.\*\)\$`)
-}
-
-func (s *attrConstraintsSuite) TestMissingCheck(c *C) {
-	m, err := asserts.ParseHeaders([]byte(`attrs:
-  foo: $MISSING`))
-	c.Assert(err, IsNil)
-
-	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
-	c.Assert(err, IsNil)
-	c.Check(asserts.RuleFeature(cstrs, "dollar-attr-constraints"), Equals, true)
-
-	err = cstrs.Check(attrs(`
-bar: baz
-`), nil)
-	c.Check(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: ["x"]
-`), nil)
-	c.Check(err, ErrorMatches, `attribute "foo" is constrained to be missing but is set`)
-}
-
 type testEvalAttr struct {
 	comp func(side string, arg string) (interface{}, error)
 }
@@ -548,30 +358,6 @@ bar: bar.baz
 	c.Check(err, ErrorMatches, `attribute "foo" does not match \$SLOT\(foo\): foo != other-value`)
 }
 
-func (s *attrConstraintsSuite) TestMatchingListsMap(c *C) {
-	m, err := asserts.ParseHeaders([]byte(`attrs:
-  foo:
-    p: /foo/.*`))
-	c.Assert(err, IsNil)
-
-	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
-	c.Assert(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: [{p: "/foo/x"}, {p: "/foo/y"}]
-`), nil)
-	c.Check(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: [{p: "zzz"}, {p: "/foo/y"}]
-`), nil)
-	c.Check(err, ErrorMatches, `attribute "foo\.0\.p" value "zzz" does not match \^\(/foo/\.\*\)\$`)
-}
-
-func (s *attrConstraintsSuite) TestAlwaysMatchAttributeConstraints(c *C) {
-	c.Check(asserts.AlwaysMatchAttributes.Check(nil, nil), IsNil)
-}
-
 func (s *attrConstraintsSuite) TestNeverMatchAttributeConstraints(c *C) {
 	c.Check(asserts.NeverMatchAttributes.Check(nil, nil), NotNil)
 }
@@ -892,6 +678,7 @@ func (s *plugSlotRulesSuite) TestCompilePlugRuleInstalationConstraintsIDConstrai
 	rule, err := asserts.CompilePlugRule("iface", map[string]interface{}{
 		"allow-installation": map[string]interface{}{
 			"plug-snap-type": []interface{}{"core", "kernel", "gadget", "app"},
+			"plug-snap-id":   []interface{}{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"},
 		},
 	})
 	c.Assert(err, IsNil)
@@ -899,6 +686,7 @@ func (s *plugSlotRulesSuite) TestCompilePlugRuleInstalationConstraintsIDConstrai
 	c.Assert(rule.AllowInstallation, HasLen, 1)
 	cstrs := rule.AllowInstallation[0]
 	c.Check(cstrs.PlugSnapTypes, DeepEquals, []string{"core", "kernel", "gadget", "app"})
+	c.Check(cstrs.PlugSnapIDs, DeepEquals, []string{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"})
 }
 
 func (s *plugSlotRulesSuite) TestCompilePlugRuleInstallationConstraintsOnClassic(c *C) {
@@ -1736,6 +1524,7 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleInstallationConstraintsIDConstra
 	rule, err := asserts.CompileSlotRule("iface", map[string]interface{}{
 		"allow-installation": map[string]interface{}{
 			"slot-snap-type": []interface{}{"core", "kernel", "gadget", "app"},
+			"slot-snap-id":   []interface{}{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"},
 		},
 	})
 	c.Assert(err, IsNil)
@@ -1743,6 +1532,7 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleInstallationConstraintsIDConstra
 	c.Assert(rule.AllowInstallation, HasLen, 1)
 	cstrs := rule.AllowInstallation[0]
 	c.Check(cstrs.SlotSnapTypes, DeepEquals, []string{"core", "kernel", "gadget", "app"})
+	c.Check(cstrs.SlotSnapIDs, DeepEquals, []string{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"})
 }
 
 func (s *plugSlotRulesSuite) TestCompileSlotRuleInstallationConstraintsOnClassic(c *C) {
@@ -1896,6 +1686,7 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleInstallationConstraintsSlotNames
 func (s *plugSlotRulesSuite) TestCompileSlotRuleConnectionConstraintsIDConstraints(c *C) {
 	rule, err := asserts.CompileSlotRule("iface", map[string]interface{}{
 		"allow-connection": map[string]interface{}{
+			"slot-snap-type":    []interface{}{"core"},
 			"plug-snap-type":    []interface{}{"core", "kernel", "gadget", "app"},
 			"plug-snap-id":      []interface{}{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"},
 			"plug-publisher-id": []interface{}{"pubidpubidpubidpubidpubidpubid09", "canonical", "$SAME"},
@@ -1905,6 +1696,7 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleConnectionConstraintsIDConstrain
 
 	c.Assert(rule.AllowConnection, HasLen, 1)
 	cstrs := rule.AllowConnection[0]
+	c.Check(cstrs.SlotSnapTypes, DeepEquals, []string{"core"})
 	c.Check(cstrs.PlugSnapTypes, DeepEquals, []string{"core", "kernel", "gadget", "app"})
 	c.Check(cstrs.PlugSnapIDs, DeepEquals, []string{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"})
 	c.Check(cstrs.PlugPublisherIDs, DeepEquals, []string{"pubidpubidpubidpubidpubidpubid09", "canonical", "$SAME"})
@@ -2188,6 +1980,10 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleErrors(c *C) {
     plug-snap-id:
       - foo`, `plug-snap-id in allow-connection in slot rule for interface "iface" contains an invalid element: "foo"`},
 		{`iface:
+  allow-connection:
+    slot-snap-type:
+      - foo`, `slot-snap-type in allow-connection in slot rule for interface "iface" contains an invalid element: "foo"`},
+		{`iface:
   allow-connection:
     plug-snap-type:
       - foo`, `plug-snap-type in allow-connection in slot rule for interface "iface" contains an invalid element: "foo"`},
@@ -2206,19 +2002,19 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleErrors(c *C) {
 		{`iface:
   allow-connection:
     plug-snap-ids:
-      - foo`, `allow-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `allow-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
 		{`iface:
   deny-connection:
     plug-snap-ids:
-      - foo`, `deny-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `deny-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
 		{`iface:
   allow-auto-connection:
     plug-snap-ids:
-      - foo`, `allow-auto-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `allow-auto-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
 		{`iface:
   deny-auto-connection:
     plug-snap-ids:
-      - foo`, `deny-auto-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `deny-auto-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
 		{`iface:
   allow-connect: true`, `slot rule for interface "iface" must specify at least one of allow-installation, deny-installation, allow-connection, deny-connection, allow-auto-connection, deny-auto-connection`},
 		{`iface:
@@ -2355,6 +2151,7 @@ func (s *plugSlotRulesSuite) TestSlotRuleFeatures(c *C) {
 }
 
 func (s *plugSlotRulesSuite) TestValidOnStoreBrandModel(c *C) {
+	// more extensive testing is now done in deviceScopeConstraintSuite
 	tests := []struct {
 		constr string
 		value  string
@@ -2364,43 +2161,20 @@ func (s *plugSlotRulesSuite) TestValidOnStoreBrandModel(c *C) {
 		{"on-store", "foo", true},
 		{"on-store", "F_o-O88", true},
 		{"on-store", "foo!", false},
-		{"on-store", "foo.", false},
-		{"on-store", "foo/", false},
 		{"on-brand", "", false},
 		// custom set brands (length 2-28)
 		{"on-brand", "dwell", true},
 		{"on-brand", "Dwell", false},
 		{"on-brand", "dwell-88", true},
 		{"on-brand", "dwell_88", false},
-		{"on-brand", "dwell.88", false},
-		{"on-brand", "dwell:88", false},
-		{"on-brand", "dwell!88", false},
-		{"on-brand", "a", false},
-		{"on-brand", "ab", true},
 		{"on-brand", "0123456789012345678901234567", true},
 		// snappy id brands (fixed length 32)
 		{"on-brand", "01234567890123456789012345678", false},
-		{"on-brand", "012345678901234567890123456789", false},
-		{"on-brand", "0123456789012345678901234567890", false},
 		{"on-brand", "01234567890123456789012345678901", true},
-		{"on-brand", "abcdefghijklmnopqrstuvwxyz678901", true},
-		{"on-brand", "ABCDEFGHIJKLMNOPQRSTUVWCYZ678901", true},
-		{"on-brand", "ABCDEFGHIJKLMNOPQRSTUVWCYZ678901X", false},
-		{"on-brand", "ABCDEFGHIJKLMNOPQ!STUVWCYZ678901", false},
-		{"on-brand", "ABCDEFGHIJKLMNOPQ_STUVWCYZ678901", false},
-		{"on-brand", "ABCDEFGHIJKLMNOPQ-STUVWCYZ678901", false},
 		{"on-model", "", false},
-		{"on-model", "/", false},
 		{"on-model", "dwell/dwell1", true},
 		{"on-model", "dwell", false},
 		{"on-model", "dwell/", false},
-		{"on-model", "dwell//dwell1", false},
-		{"on-model", "dwell/-dwell1", false},
-		{"on-model", "dwell/dwell1-", false},
-		{"on-model", "dwell/dwell1-23", true},
-		{"on-model", "dwell/dwell1!", false},
-		{"on-model", "dwell/dwe_ll1", false},
-		{"on-model", "dwell/dwe.ll1", false},
 	}
 
 	check := func(constr, value string, valid bool) {
