@@ -300,14 +300,19 @@ func MakeRecoverySystemBootable(model *asserts.Model, rootdir string, relativeRe
 	recoveryBlVars := map[string]string{
 		"snapd_recovery_kernel": filepath.Join("/", kernelPath),
 	}
-	if _, ok := bl.(bootloader.TrustedAssetsBootloader); ok {
+	if tbl, ok := bl.(bootloader.TrustedAssetsBootloader); ok {
 		// Look at gadget default values for system.kernel.*cmdline-append options
 		cmdlineAppend, err := buildOptionalKernelCommandLine(model, bootWith.GadgetSnapOrDir)
 		if err != nil {
 			return fmt.Errorf("while retrieving system.kernel.*cmdline-append defaults: %v", err)
 		}
+		candidate := false
+		defaultCmdLine, err := tbl.DefaultCommandLine(candidate)
+		if err != nil {
+			return err
+		}
 		// to set cmdlineAppend.
-		recoveryCmdlineArgs, err := bootVarsForTrustedCommandLineFromGadget(bootWith.GadgetSnapOrDir, cmdlineAppend)
+		recoveryCmdlineArgs, err := bootVarsForTrustedCommandLineFromGadget(bootWith.GadgetSnapOrDir, cmdlineAppend, defaultCmdLine, model)
 		if err != nil {
 			return fmt.Errorf("cannot obtain recovery system command line: %v", err)
 		}
@@ -325,6 +330,8 @@ func MakeRecoverySystemBootable(model *asserts.Model, rootdir string, relativeRe
 type makeRunnableOptions struct {
 	Standalone     bool
 	AfterDataReset bool
+	SeedDir        string
+	StateUnlocker  Unlocker
 }
 
 func copyBootSnap(orig string, dstInfo *snap.Info, dstSnapBlobDir string) error {
@@ -357,6 +364,8 @@ func makeRunnableSystem(model *asserts.Model, bootWith *BootableSet, sealer *Tru
 	if bootWith.RecoverySystemDir != "" {
 		return fmt.Errorf("internal error: RecoverySystemDir unexpectedly set for MakeRunnableSystem")
 	}
+	modeenvLock()
+	defer modeenvUnlock()
 
 	// TODO:UC20:
 	// - figure out what to do for uboot gadgets, currently we require them to
@@ -480,7 +489,7 @@ func makeRunnableSystem(model *asserts.Model, bootWith *BootableSet, sealer *Tru
 		return fmt.Errorf("cannot set run system environment: %v", err)
 	}
 
-	_, ok = bl.(bootloader.TrustedAssetsBootloader)
+	tbl, ok := bl.(bootloader.TrustedAssetsBootloader)
 	if ok {
 		// the bootloader can manage its boot config
 
@@ -501,7 +510,14 @@ func makeRunnableSystem(model *asserts.Model, bootWith *BootableSet, sealer *Tru
 		if err != nil {
 			return fmt.Errorf("while retrieving system.kernel.*cmdline-append defaults: %v", err)
 		}
-		cmdlineVars, err := bootVarsForTrustedCommandLineFromGadget(bootWith.UnpackedGadgetDir, cmdlineAppend)
+
+		candidate := false
+		defaultCmdLine, err := tbl.DefaultCommandLine(candidate)
+		if err != nil {
+			return err
+		}
+
+		cmdlineVars, err := bootVarsForTrustedCommandLineFromGadget(bootWith.UnpackedGadgetDir, cmdlineAppend, defaultCmdLine, model)
 		if err != nil {
 			return fmt.Errorf("cannot prepare bootloader variables for kernel command line: %v", err)
 		}
@@ -525,6 +541,8 @@ func makeRunnableSystem(model *asserts.Model, bootWith *BootableSet, sealer *Tru
 		flags := sealKeyToModeenvFlags{
 			HasFDESetupHook: hasHook,
 			FactoryReset:    makeOpts.AfterDataReset,
+			SeedDir:         makeOpts.SeedDir,
+			StateUnlocker:   makeOpts.StateUnlocker,
 		}
 		if makeOpts.Standalone {
 			flags.SnapsDir = snapBlobDir
@@ -595,17 +613,32 @@ func buildOptionalKernelCommandLine(model *asserts.Model, gadgetSnapOrDir string
 // setting up a run system and actually transitioning to it, with hooks, etc.
 // running in between.
 func MakeRunnableSystem(model *asserts.Model, bootWith *BootableSet, sealer *TrustedAssetsInstallObserver) error {
-	return makeRunnableSystem(model, bootWith, sealer, makeRunnableOptions{})
+	return makeRunnableSystem(model, bootWith, sealer, makeRunnableOptions{
+		SeedDir: dirs.SnapSeedDir,
+	})
 }
 
 // MakeRunnableStandaloneSystem operates like MakeRunnableSystem but does
-// assume that the run system being set up is related to the current
+// not assume that the run system being set up is related to the current
 // system. This is appropriate e.g when installing from a classic installer.
-func MakeRunnableStandaloneSystem(model *asserts.Model, bootWith *BootableSet, sealer *TrustedAssetsInstallObserver) error {
+func MakeRunnableStandaloneSystem(model *asserts.Model, bootWith *BootableSet, sealer *TrustedAssetsInstallObserver, unlocker Unlocker) error {
+	// TODO consider merging this back into MakeRunnableSystem but need
+	// to consider the properties of the different input used for sealing
+	return makeRunnableSystem(model, bootWith, sealer, makeRunnableOptions{
+		Standalone:    true,
+		SeedDir:       dirs.SnapSeedDir,
+		StateUnlocker: unlocker,
+	})
+}
+
+// MakeRunnableStandaloneSystemFromInitrd is the same as MakeRunnableStandaloneSystem
+// but uses seed dir path expected in initrd.
+func MakeRunnableStandaloneSystemFromInitrd(model *asserts.Model, bootWith *BootableSet, sealer *TrustedAssetsInstallObserver) error {
 	// TODO consider merging this back into MakeRunnableSystem but need
 	// to consider the properties of the different input used for sealing
 	return makeRunnableSystem(model, bootWith, sealer, makeRunnableOptions{
 		Standalone: true,
+		SeedDir:    filepath.Join(InitramfsRunMntDir, "ubuntu-seed"),
 	})
 }
 
@@ -615,5 +648,6 @@ func MakeRunnableStandaloneSystem(model *asserts.Model, bootWith *BootableSet, s
 func MakeRunnableSystemAfterDataReset(model *asserts.Model, bootWith *BootableSet, sealer *TrustedAssetsInstallObserver) error {
 	return makeRunnableSystem(model, bootWith, sealer, makeRunnableOptions{
 		AfterDataReset: true,
+		SeedDir:        dirs.SnapSeedDir,
 	})
 }
