@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2020-2022 Canonical Ltd
+ * Copyright (C) 2020-2023 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -81,7 +81,7 @@ func MockSecbootResealKeys(f func(params *secboot.ResealKeysParams) error) (rest
 }
 
 // MockResealKeyToModeenv is only useful in testing.
-func MockResealKeyToModeenv(f func(rootdir string, modeenv *Modeenv, expectReseal bool) error) (restore func()) {
+func MockResealKeyToModeenv(f func(rootdir string, modeenv *Modeenv, expectReseal bool, unlocker Unlocker) error) (restore func()) {
 	osutil.MustBeTestBinary("resealKeyToModeenv only can be mocked in tests")
 	old := resealKeyToModeenv
 	resealKeyToModeenv = f
@@ -119,12 +119,21 @@ type sealKeyToModeenvFlags struct {
 	// SnapsDir is set to provide a non-default directory to find
 	// run mode snaps in.
 	SnapsDir string
+	// SeedDir is the path where to find mounted seed with
+	// essential snaps.
+	SeedDir string
+	// Unlocker is used unlock the snapd state for long operations
+	StateUnlocker Unlocker
 }
 
 // sealKeyToModeenvImpl seals the supplied keys to the parameters specified
 // in modeenv.
 // It assumes to be invoked in install mode.
 func sealKeyToModeenvImpl(key, saveKey keys.EncryptionKey, model *asserts.Model, modeenv *Modeenv, flags sealKeyToModeenvFlags) error {
+	if !isModeeenvLocked() {
+		return fmt.Errorf("internal error: cannot seal without the modeenv lock")
+	}
+
 	// make sure relevant locations exist
 	for _, p := range []string{
 		InitramfsSeedEncryptionKeyDir,
@@ -142,6 +151,10 @@ func sealKeyToModeenvImpl(key, saveKey keys.EncryptionKey, model *asserts.Model,
 		return sealKeyToModeenvUsingFDESetupHook(key, saveKey, model, modeenv, flags)
 	}
 
+	if flags.StateUnlocker != nil {
+		relock := flags.StateUnlocker()
+		defer relock()
+	}
 	return sealKeyToModeenvUsingSecboot(key, saveKey, model, modeenv, flags)
 }
 
@@ -228,7 +241,7 @@ func sealKeyToModeenvUsingSecboot(key, saveKey keys.EncryptionKey, model *assert
 		// tested, hence allow both recover and factory reset modes
 		modeenv.RecoverySystem: {ModeRecover, ModeFactoryReset},
 	}
-	recoveryBootChains, err := recoveryBootChainsForSystems(systems, modes, tbl, modeenv, includeTryModel)
+	recoveryBootChains, err := recoveryBootChainsForSystems(systems, modes, tbl, modeenv, includeTryModel, flags.SeedDir)
 	if err != nil {
 		return fmt.Errorf("cannot compose recovery boot chains: %v", err)
 	}
@@ -408,7 +421,11 @@ var resealKeyToModeenv = resealKeyToModeenvImpl
 // atomically.  In particular we want to avoid resealing against
 // transient/in-memory information with the risk that successive
 // reseals during in-progress operations produce diverging outcomes.
-func resealKeyToModeenvImpl(rootdir string, modeenv *Modeenv, expectReseal bool) error {
+func resealKeyToModeenvImpl(rootdir string, modeenv *Modeenv, expectReseal bool, unlocker Unlocker) error {
+	if !isModeeenvLocked() {
+		return fmt.Errorf("internal error: cannot reseal without the modeenv lock")
+	}
+
 	method, err := device.SealedKeysMethod(rootdir)
 	if err == device.ErrNoSealedKeys {
 		// nothing to do
@@ -421,6 +438,10 @@ func resealKeyToModeenvImpl(rootdir string, modeenv *Modeenv, expectReseal bool)
 	case device.SealingMethodFDESetupHook:
 		return resealKeyToModeenvUsingFDESetupHook(rootdir, modeenv, expectReseal)
 	case device.SealingMethodTPM, device.SealingMethodLegacyTPM:
+		if unlocker != nil {
+			// unlock/relock global state
+			defer unlocker()()
+		}
 		return resealKeyToModeenvSecboot(rootdir, modeenv, expectReseal)
 	default:
 		return fmt.Errorf("unknown key sealing method: %q", method)
@@ -473,7 +494,7 @@ func resealKeyToModeenvSecboot(rootdir string, modeenv *Modeenv, expectReseal bo
 	// accommodate the dynamics of a remodel
 	includeTryModel := true
 	recoveryBootChainsForRunKey, err := recoveryBootChainsForSystems(modeenv.CurrentRecoverySystems, modes, tbl,
-		modeenv, includeTryModel)
+		modeenv, includeTryModel, dirs.SnapSeedDir)
 	if err != nil {
 		return fmt.Errorf("cannot compose recovery boot chains for run key: %v", err)
 	}
@@ -491,7 +512,7 @@ func resealKeyToModeenvSecboot(rootdir string, modeenv *Modeenv, expectReseal bo
 	// use the current model as the recovery keys are not expected to be
 	// used during a remodel
 	includeTryModel = false
-	recoveryBootChains, err := recoveryBootChainsForSystems(testedRecoverySystems, modes, tbl, modeenv, includeTryModel)
+	recoveryBootChains, err := recoveryBootChainsForSystems(testedRecoverySystems, modes, tbl, modeenv, includeTryModel, dirs.SnapSeedDir)
 	if err != nil {
 		return fmt.Errorf("cannot compose recovery boot chains: %v", err)
 	}
@@ -648,13 +669,18 @@ func modesForSystems(modeenv *Modeenv) map[string][]string {
 
 // TODO:UC20: this needs to take more than one model to accommodate the remodel
 // scenario
-func recoveryBootChainsForSystems(systems []string, modesForSystems map[string][]string, trbl bootloader.TrustedAssetsBootloader, modeenv *Modeenv, includeTryModel bool) (chains []bootChain, err error) {
+func recoveryBootChainsForSystems(systems []string, modesForSystems map[string][]string, trbl bootloader.TrustedAssetsBootloader, modeenv *Modeenv, includeTryModel bool, seedDir string) (chains []bootChain, err error) {
+	trustedAssets, err := trbl.TrustedAssets()
+	if err != nil {
+		return nil, err
+	}
+
 	chainsForModel := func(model secboot.ModelForSealing) error {
 		modelID := modelUniqueID(model)
 		for _, system := range systems {
 			// get kernel and gadget information from seed
 			perf := timings.New(nil)
-			seedSystemModel, snaps, err := seedReadSystemEssential(dirs.SnapSeedDir, system, []snap.Type{snap.TypeKernel, snap.TypeGadget}, perf)
+			seedSystemModel, snaps, err := seedReadSystemEssential(seedDir, system, []snap.Type{snap.TypeKernel, snap.TypeGadget}, perf)
 			if err != nil {
 				return fmt.Errorf("cannot read system %q seed: %v", system, err)
 			}
@@ -684,7 +710,7 @@ func recoveryBootChainsForSystems(systems []string, modesForSystems map[string][
 			}
 			for _, mode := range modes {
 				// get the command line for this mode
-				cmdline, err := composeCommandLine(currentEdition, mode, system, seedGadget.Path)
+				cmdline, err := composeCommandLine(currentEdition, mode, system, seedGadget.Path, model)
 				if err != nil {
 					return fmt.Errorf("cannot obtain kernel command line for mode %q: %v", mode, err)
 				}
@@ -696,30 +722,47 @@ func recoveryBootChainsForSystems(systems []string, modesForSystems map[string][
 				kernelRev = seedKernel.SideInfo.Revision.String()
 			}
 
-			recoveryBootChain, err := trbl.RecoveryBootChain(seedKernel.Path)
+			recoveryBootChains, err := trbl.RecoveryBootChains(seedKernel.Path)
 			if err != nil {
 				return err
 			}
 
+			foundChain := false
+
 			// get asset chains
-			assetChain, kbf, err := buildBootAssets(recoveryBootChain, modeenv)
-			if err != nil {
-				return err
+			for _, recoveryBootChain := range recoveryBootChains {
+				assetChain, kbf, err := buildBootAssets(recoveryBootChain, modeenv, trustedAssets)
+				if err != nil {
+					return err
+				}
+				if assetChain == nil {
+					// This chain is not used as
+					// it is not in the modeenv,
+					// we expect another chain to
+					// work.
+					continue
+				}
+
+				chains = append(chains, bootChain{
+					BrandID: model.BrandID(),
+					Model:   model.Model(),
+					// TODO: test this
+					Classic:        model.Classic(),
+					Grade:          model.Grade(),
+					ModelSignKeyID: model.SignKeyID(),
+					AssetChain:     assetChain,
+					Kernel:         seedKernel.SnapName(),
+					KernelRevision: kernelRev,
+					KernelCmdlines: cmdlines,
+					kernelBootFile: kbf,
+				})
+
+				foundChain = true
 			}
 
-			chains = append(chains, bootChain{
-				BrandID: model.BrandID(),
-				Model:   model.Model(),
-				// TODO: test this
-				Classic:        model.Classic(),
-				Grade:          model.Grade(),
-				ModelSignKeyID: model.SignKeyID(),
-				AssetChain:     assetChain,
-				Kernel:         seedKernel.SnapName(),
-				KernelRevision: kernelRev,
-				KernelCmdlines: cmdlines,
-				kernelBootFile: kbf,
-			})
+			if !foundChain {
+				return fmt.Errorf("could not find any valid chain for this model")
+			}
 		}
 		return nil
 	}
@@ -744,6 +787,11 @@ func runModeBootChains(rbl, bl bootloader.Bootloader, modeenv *Modeenv, cmdlines
 	}
 	chains := make([]bootChain, 0, len(modeenv.CurrentKernels))
 
+	trustedAssets, err := tbl.TrustedAssets()
+	if err != nil {
+		return nil, err
+	}
+
 	chainsForModel := func(model secboot.ModelForSealing) error {
 		for _, k := range modeenv.CurrentKernels {
 			info, err := snap.ParsePlaceInfoFromSnapFileName(k)
@@ -756,33 +804,49 @@ func runModeBootChains(rbl, bl bootloader.Bootloader, modeenv *Modeenv, cmdlines
 			} else {
 				kernelPath = filepath.Join(runSnapsDir, info.Filename())
 			}
-			runModeBootChain, err := tbl.BootChain(bl, kernelPath)
+			runModeBootChains, err := tbl.BootChains(bl, kernelPath)
 			if err != nil {
 				return err
 			}
 
-			// get asset chains
-			assetChain, kbf, err := buildBootAssets(runModeBootChain, modeenv)
-			if err != nil {
-				return err
+			foundChain := false
+
+			for _, runModeBootChain := range runModeBootChains {
+				// get asset chains
+				assetChain, kbf, err := buildBootAssets(runModeBootChain, modeenv, trustedAssets)
+				if err != nil {
+					return err
+				}
+				if assetChain == nil {
+					// This chain is not used as
+					// it is not in the modeenv,
+					// we expect another chain to
+					// work.
+					continue
+				}
+				var kernelRev string
+				if info.SnapRevision().Store() {
+					kernelRev = info.SnapRevision().String()
+				}
+				chains = append(chains, bootChain{
+					BrandID: model.BrandID(),
+					Model:   model.Model(),
+					// TODO: test this
+					Classic:        model.Classic(),
+					Grade:          model.Grade(),
+					ModelSignKeyID: model.SignKeyID(),
+					AssetChain:     assetChain,
+					Kernel:         info.SnapName(),
+					KernelRevision: kernelRev,
+					KernelCmdlines: cmdlines,
+					kernelBootFile: kbf,
+				})
+				foundChain = true
 			}
-			var kernelRev string
-			if info.SnapRevision().Store() {
-				kernelRev = info.SnapRevision().String()
+
+			if !foundChain {
+				return fmt.Errorf("could not find any valid chain for this model")
 			}
-			chains = append(chains, bootChain{
-				BrandID: model.BrandID(),
-				Model:   model.Model(),
-				// TODO: test this
-				Classic:        model.Classic(),
-				Grade:          model.Grade(),
-				ModelSignKeyID: model.SignKeyID(),
-				AssetChain:     assetChain,
-				Kernel:         info.SnapName(),
-				KernelRevision: kernelRev,
-				KernelCmdlines: cmdlines,
-				kernelBootFile: kbf,
-			})
 		}
 		return nil
 	}
@@ -802,7 +866,7 @@ func runModeBootChains(rbl, bl bootloader.Bootloader, modeenv *Modeenv, cmdlines
 // produces corresponding bootAssets with the matching current asset
 // hashes from modeenv plus it returns separately the last BootFile
 // which is for the kernel.
-func buildBootAssets(bootFiles []bootloader.BootFile, modeenv *Modeenv) (assets []bootAsset, kernel bootloader.BootFile, err error) {
+func buildBootAssets(bootFiles []bootloader.BootFile, modeenv *Modeenv, trustedAssets map[string]string) (assets []bootAsset, kernel bootloader.BootFile, err error) {
 	if len(bootFiles) == 0 {
 		// useful in testing, when mocking is insufficient
 		return nil, bootloader.BootFile{}, fmt.Errorf("internal error: cannot build boot assets without boot files")
@@ -811,16 +875,25 @@ func buildBootAssets(bootFiles []bootloader.BootFile, modeenv *Modeenv) (assets
 
 	// the last element is the kernel which is not a boot asset
 	for i, bf := range bootFiles[:len(bootFiles)-1] {
-		name := filepath.Base(bf.Path)
+		path := bf.Path
+		name, ok := trustedAssets[path]
+		if !ok {
+			return nil, kernel, fmt.Errorf("internal error: asset '%s' is not considered a trusted asset for the bootloader", path)
+		}
 		var hashes []string
-		var ok bool
 		if bf.Role == bootloader.RoleRecovery {
 			hashes, ok = modeenv.CurrentTrustedRecoveryBootAssets[name]
 		} else {
 			hashes, ok = modeenv.CurrentTrustedBootAssets[name]
 		}
 		if !ok {
-			return nil, kernel, fmt.Errorf("cannot find expected boot asset %s in modeenv", name)
+			// We have not found an asset for this
+			// chain. There are chains expected to not
+			// exist. So we return without error.
+			// recoveryBootChainsForSystems and
+			// runModeBootChains will fail if no chain is
+			// found
+			return nil, kernel, nil
 		}
 		assets[i] = bootAsset{
 			Role:   bf.Role,
@@ -840,7 +913,8 @@ func sealKeyModelParams(pbc predictableBootChains, roleToBlName map[bootloader.R
 	for _, bc := range pbc {
 		modelForSealing := bc.modelForSealing()
 		modelID := modelUniqueID(modelForSealing)
-		loadChains, err := bootAssetsToLoadChains(bc.AssetChain, bc.kernelBootFile, roleToBlName)
+		const expectNew = false
+		loadChains, err := bootAssetsToLoadChains(bc.AssetChain, bc.kernelBootFile, roleToBlName, expectNew)
 		if err != nil {
 			return nil, fmt.Errorf("cannot build load chains with current boot assets: %s", err)
 		}
