@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2016-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -24,15 +24,16 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"mime"
 	"mime/multipart"
 	"net/http"
+	"os"
 	"path/filepath"
 
 	"gopkg.in/check.v1"
 
 	"github.com/snapcore/snapd/client"
+	"github.com/snapcore/snapd/testutil"
 )
 
 var chanName = "achan"
@@ -48,6 +49,8 @@ var ops = []struct {
 	{(*client.Client).Enable, "enable"},
 	{(*client.Client).Disable, "disable"},
 	{(*client.Client).Switch, "switch"},
+	{(*client.Client).HoldRefreshes, "hold"},
+	{(*client.Client).UnholdRefreshes, "unhold"},
 }
 
 var multiOps = []struct {
@@ -57,6 +60,8 @@ var multiOps = []struct {
 	{(*client.Client).RefreshMany, "refresh"},
 	{(*client.Client).InstallMany, "install"},
 	{(*client.Client).RemoveMany, "remove"},
+	{(*client.Client).HoldRefreshesMany, "hold"},
+	{(*client.Client).UnholdRefreshesMany, "unhold"},
 }
 
 func (cs *clientSuite) TestClientOpSnapServerError(c *check.C) {
@@ -144,7 +149,7 @@ func (cs *clientSuite) TestClientOpSnap(c *check.C) {
 		_, ok := cs.req.Context().Deadline()
 		c.Check(ok, check.Equals, true)
 
-		body, err := ioutil.ReadAll(cs.req.Body)
+		body, err := io.ReadAll(cs.req.Body)
 		c.Assert(err, check.IsNil, check.Commentf(s.action))
 		jsonBody := make(map[string]string)
 		err = json.Unmarshal(body, &jsonBody)
@@ -172,7 +177,7 @@ func (cs *clientSuite) TestClientMultiOpSnap(c *check.C) {
 
 		c.Assert(cs.req.Header.Get("Content-Type"), check.Equals, "application/json", check.Commentf(s.action))
 
-		body, err := ioutil.ReadAll(cs.req.Body)
+		body, err := io.ReadAll(cs.req.Body)
 		c.Assert(err, check.IsNil, check.Commentf(s.action))
 		jsonBody := make(map[string]interface{})
 		err = json.Unmarshal(body, &jsonBody)
@@ -186,6 +191,68 @@ func (cs *clientSuite) TestClientMultiOpSnap(c *check.C) {
 	}
 }
 
+func (cs *clientSuite) TestClientMultiOpSnapTransactional(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "d728",
+		"status-code": 202,
+		"type": "async"
+	}`
+	for _, s := range multiOps {
+		// Note body is essentially the same as TestClientMultiSnapshot; keep in sync
+		id, err := s.op(cs.cli, []string{pkgName},
+			&client.SnapOptions{Transaction: client.TransactionAllSnaps})
+		c.Assert(err, check.IsNil)
+
+		c.Assert(cs.req.Header.Get("Content-Type"), check.Equals, "application/json", check.Commentf(s.action))
+
+		body, err := io.ReadAll(cs.req.Body)
+		c.Assert(err, check.IsNil, check.Commentf(s.action))
+		jsonBody := make(map[string]interface{})
+		err = json.Unmarshal(body, &jsonBody)
+		c.Assert(err, check.IsNil, check.Commentf(s.action))
+		c.Check(jsonBody["action"], check.Equals, s.action, check.Commentf(s.action))
+		c.Check(jsonBody["snaps"], check.DeepEquals, []interface{}{pkgName}, check.Commentf(s.action))
+		c.Check(jsonBody["transaction"], check.Equals, string(client.TransactionAllSnaps),
+			check.Commentf(s.action))
+		c.Check(jsonBody, check.HasLen, 3, check.Commentf(s.action))
+
+		c.Check(cs.req.URL.Path, check.Equals, "/v2/snaps", check.Commentf(s.action))
+		c.Check(id, check.Equals, "d728", check.Commentf(s.action))
+	}
+}
+
+func (cs *clientSuite) TestClientMultiOpSnapIgnoreRunning(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "d728",
+		"status-code": 202,
+		"type": "async"
+	}`
+	for _, s := range multiOps {
+		// Note body is essentially the same as TestClientMultiSnapshot; keep in sync
+		id, err := s.op(cs.cli, []string{pkgName},
+			&client.SnapOptions{IgnoreRunning: true})
+		c.Assert(err, check.IsNil)
+
+		c.Assert(cs.req.Header.Get("Content-Type"), check.Equals, "application/json", check.Commentf(s.action))
+
+		body, err := io.ReadAll(cs.req.Body)
+		c.Assert(err, check.IsNil, check.Commentf(s.action))
+		jsonBody := make(map[string]interface{})
+		err = json.Unmarshal(body, &jsonBody)
+		c.Assert(err, check.IsNil, check.Commentf(s.action))
+		c.Check(jsonBody["action"], check.Equals, s.action, check.Commentf(s.action))
+		c.Check(jsonBody["snaps"], check.DeepEquals, []interface{}{pkgName}, check.Commentf(s.action))
+		c.Check(jsonBody["ignore-running"], check.Equals, true,
+			check.Commentf(s.action))
+		c.Check(jsonBody, check.HasLen, 3, check.Commentf(s.action))
+
+		c.Check(cs.req.URL.Path, check.Equals, "/v2/snaps", check.Commentf(s.action))
+		c.Check(id, check.Equals, "d728", check.Commentf(s.action))
+	}
+}
+
 func (cs *clientSuite) TestClientMultiSnapshot(c *check.C) {
 	// Note body is essentially the same as TestClientMultiOpSnap; keep in sync
 	cs.status = 202
@@ -199,7 +266,7 @@ func (cs *clientSuite) TestClientMultiSnapshot(c *check.C) {
 	c.Assert(err, check.IsNil)
 	c.Check(cs.req.Header.Get("Content-Type"), check.Equals, "application/json")
 
-	body, err := ioutil.ReadAll(cs.req.Body)
+	body, err := io.ReadAll(cs.req.Body)
 	c.Assert(err, check.IsNil)
 	jsonBody := make(map[string]interface{})
 	err = json.Unmarshal(body, &jsonBody)
@@ -222,21 +289,21 @@ func (cs *clientSuite) TestClientOpInstallPath(c *check.C) {
 	bodyData := []byte("snap-data")
 
 	snap := filepath.Join(c.MkDir(), "foo.snap")
-	err := ioutil.WriteFile(snap, bodyData, 0644)
+	err := os.WriteFile(snap, bodyData, 0644)
 	c.Assert(err, check.IsNil)
 
 	id, err := cs.cli.InstallPath(snap, "", nil)
 	c.Assert(err, check.IsNil)
 
-	body, err := ioutil.ReadAll(cs.req.Body)
+	body, err := io.ReadAll(cs.req.Body)
 	c.Assert(err, check.IsNil)
 
-	c.Assert(string(body), check.Matches, "(?s).*\r\nsnap-data\r\n.*")
-	c.Assert(string(body), check.Matches, "(?s).*Content-Disposition: form-data; name=\"action\"\r\n\r\ninstall\r\n.*")
+	c.Assert(string(body), testutil.Contains, "\r\nsnap-data\r\n")
+	c.Assert(string(body), testutil.Contains, "Content-Disposition: form-data; name=\"action\"\r\n\r\ninstall\r\n")
 
 	c.Check(cs.req.Method, check.Equals, "POST")
 	c.Check(cs.req.URL.Path, check.Equals, "/v2/snaps")
-	c.Assert(cs.req.Header.Get("Content-Type"), check.Matches, "multipart/form-data; boundary=.*")
+	c.Assert(cs.req.Header.Get("Content-Type"), testutil.Contains, "multipart/form-data; boundary=")
 	_, ok := cs.req.Context().Deadline()
 	c.Assert(ok, check.Equals, false)
 	c.Check(id, check.Equals, "66b3")
@@ -252,13 +319,13 @@ func (cs *clientSuite) TestClientOpInstallPathIgnoreRunning(c *check.C) {
 	bodyData := []byte("snap-data")
 
 	snap := filepath.Join(c.MkDir(), "foo.snap")
-	err := ioutil.WriteFile(snap, bodyData, 0644)
+	err := os.WriteFile(snap, bodyData, 0644)
 	c.Assert(err, check.IsNil)
 
 	id, err := cs.cli.InstallPath(snap, "", &client.SnapOptions{IgnoreRunning: true})
 	c.Assert(err, check.IsNil)
 
-	body, err := ioutil.ReadAll(cs.req.Body)
+	body, err := io.ReadAll(cs.req.Body)
 	c.Assert(err, check.IsNil)
 
 	c.Assert(string(body), check.Matches, "(?s).*\r\nsnap-data\r\n.*")
@@ -283,13 +350,13 @@ func (cs *clientSuite) TestClientOpInstallPathInstance(c *check.C) {
 	bodyData := []byte("snap-data")
 
 	snap := filepath.Join(c.MkDir(), "foo.snap")
-	err := ioutil.WriteFile(snap, bodyData, 0644)
+	err := os.WriteFile(snap, bodyData, 0644)
 	c.Assert(err, check.IsNil)
 
 	id, err := cs.cli.InstallPath(snap, "foo_bar", nil)
 	c.Assert(err, check.IsNil)
 
-	body, err := ioutil.ReadAll(cs.req.Body)
+	body, err := io.ReadAll(cs.req.Body)
 	c.Assert(err, check.IsNil)
 
 	c.Assert(string(body), check.Matches, "(?s).*\r\nsnap-data\r\n.*")
@@ -302,6 +369,143 @@ func (cs *clientSuite) TestClientOpInstallPathInstance(c *check.C) {
 	c.Check(id, check.Equals, "66b3")
 }
 
+func (cs *clientSuite) TestClientOpInstallPathMany(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "66b3",
+		"status-code": 202,
+		"type": "async"
+	}`
+
+	var paths []string
+	names := []string{"foo.snap", "bar.snap"}
+	for _, name := range names {
+		path := filepath.Join(c.MkDir(), name)
+		paths = append(paths, path)
+		c.Assert(os.WriteFile(path, []byte("snap-data"), 0644), check.IsNil)
+	}
+
+	id, err := cs.cli.InstallPathMany(paths, nil)
+	c.Assert(err, check.IsNil)
+
+	body, err := io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+
+	for _, name := range names {
+		c.Assert(string(body), check.Matches, fmt.Sprintf(`(?s).*Content-Disposition: form-data; name="snap"; filename="%s"\r\nContent-Type: application/octet-stream\r\n\r\nsnap-data\r\n.*`, name))
+
+	}
+	c.Assert(string(body), check.Matches, `(?s).*Content-Disposition: form-data; name="action"\r\n\r\ninstall\r\n.*`)
+
+	c.Check(cs.req.Method, check.Equals, "POST")
+	c.Check(cs.req.URL.Path, check.Equals, "/v2/snaps")
+	c.Assert(cs.req.Header.Get("Content-Type"), check.Matches, "multipart/form-data; boundary=.*")
+
+	_, ok := cs.req.Context().Deadline()
+	c.Assert(ok, check.Equals, false)
+	c.Check(id, check.Equals, "66b3")
+}
+
+func (cs *clientSuite) TestClientOpInstallPathManyTransactionally(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "66b3",
+		"status-code": 202,
+		"type": "async"
+	}`
+
+	var paths []string
+	names := []string{"foo.snap", "bar.snap"}
+	for _, name := range names {
+		path := filepath.Join(c.MkDir(), name)
+		paths = append(paths, path)
+		c.Assert(os.WriteFile(path, []byte("snap-data"), 0644), check.IsNil)
+	}
+
+	id, err := cs.cli.InstallPathMany(paths, &client.SnapOptions{Transaction: client.TransactionAllSnaps})
+	c.Assert(err, check.IsNil)
+
+	body, err := io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+
+	for _, name := range names {
+		c.Assert(string(body), check.Matches, fmt.Sprintf(`(?s).*Content-Disposition: form-data; name="snap"; filename="%s"\r\nContent-Type: application/octet-stream\r\n\r\nsnap-data\r\n.*`, name))
+
+	}
+	c.Assert(string(body), check.Matches, `(?s).*Content-Disposition: form-data; name="action"\r\n\r\ninstall\r\n.*`)
+	c.Assert(string(body), check.Matches, `(?s).*Content-Disposition: form-data; name="transaction"\r\n\r\nall-snaps\r\n.*`)
+
+	c.Check(cs.req.Method, check.Equals, "POST")
+	c.Check(cs.req.URL.Path, check.Equals, "/v2/snaps")
+	c.Assert(cs.req.Header.Get("Content-Type"), check.Matches, "multipart/form-data; boundary=.*")
+
+	_, ok := cs.req.Context().Deadline()
+	c.Assert(ok, check.Equals, false)
+	c.Check(id, check.Equals, "66b3")
+}
+
+func (cs *clientSuite) TestClientOpInstallPathManyWithOptions(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "66b3",
+		"status-code": 202,
+		"type": "async"
+	}`
+
+	var paths []string
+	for _, name := range []string{"foo.snap", "bar.snap"} {
+		path := filepath.Join(c.MkDir(), name)
+		paths = append(paths, path)
+		c.Assert(os.WriteFile(path, []byte("snap-data"), 0644), check.IsNil)
+	}
+
+	// InstallPathMany supports opts
+	_, err := cs.cli.InstallPathMany(paths, &client.SnapOptions{
+		Dangerous: true,
+		DevMode:   true,
+		Classic:   true,
+	})
+
+	c.Assert(err, check.IsNil)
+
+	body, err := io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+
+	c.Check(string(body), check.Matches, `(?s).*Content-Disposition: form-data; name="dangerous"\r\n\r\ntrue\r\n.*`)
+	c.Check(string(body), check.Matches, `(?s).*Content-Disposition: form-data; name="devmode"\r\n\r\ntrue\r\n.*`)
+	c.Check(string(body), check.Matches, `(?s).*Content-Disposition: form-data; name="classic"\r\n\r\ntrue\r\n.*`)
+}
+
+func (cs *clientSuite) TestClientOpInstallPathManyWithQuotaGroup(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "66b3",
+		"status-code": 202,
+		"type": "async"
+	}`
+
+	var paths []string
+	for _, name := range []string{"foo.snap", "bar.snap"} {
+		path := filepath.Join(c.MkDir(), name)
+		paths = append(paths, path)
+		c.Assert(os.WriteFile(path, []byte("snap-data"), 0644), check.IsNil)
+	}
+
+	// Verify that the quota group option is serialized as a part of multipart form.
+	_, err := cs.cli.InstallPathMany(paths, &client.SnapOptions{
+		Dangerous:      true,
+		QuotaGroupName: "foo-group",
+	})
+
+	c.Assert(err, check.IsNil)
+
+	body, err := io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+
+	c.Check(string(body), check.Matches, `(?s).*Content-Disposition: form-data; name="dangerous"\r\n\r\ntrue\r\n.*`)
+	c.Check(string(body), check.Matches, `(?s).*Content-Disposition: form-data; name="quota-group"\r\n\r\nfoo-group\r\n.*`)
+}
+
 func (cs *clientSuite) TestClientOpInstallDangerous(c *check.C) {
 	cs.status = 202
 	cs.rsp = `{
@@ -312,7 +516,7 @@ func (cs *clientSuite) TestClientOpInstallDangerous(c *check.C) {
 	bodyData := []byte("snap-data")
 
 	snap := filepath.Join(c.MkDir(), "foo.snap")
-	err := ioutil.WriteFile(snap, bodyData, 0644)
+	err := os.WriteFile(snap, bodyData, 0644)
 	c.Assert(err, check.IsNil)
 
 	opts := client.SnapOptions{
@@ -323,7 +527,7 @@ func (cs *clientSuite) TestClientOpInstallDangerous(c *check.C) {
 	_, err = cs.cli.InstallPath(snap, "", &opts)
 	c.Assert(err, check.IsNil)
 
-	body, err := ioutil.ReadAll(cs.req.Body)
+	body, err := io.ReadAll(cs.req.Body)
 	c.Assert(err, check.IsNil)
 
 	c.Assert(string(body), check.Matches, "(?s).*Content-Disposition: form-data; name=\"dangerous\"\r\n\r\ntrue\r\n.*")
@@ -332,11 +536,9 @@ func (cs *clientSuite) TestClientOpInstallDangerous(c *check.C) {
 	_, err = cs.cli.Install("foo", &opts)
 	c.Assert(err, check.Equals, client.ErrDangerousNotApplicable)
 
-	// nor does InstallMany (whether it fails because any option
-	// at all was provided, or because dangerous was provided, is
-	// unimportant)
+	// InstallMany just ignores it without error for the moment
 	_, err = cs.cli.InstallMany([]string{"foo"}, &opts)
-	c.Assert(err, check.NotNil)
+	c.Assert(err, check.IsNil)
 }
 
 func (cs *clientSuite) TestClientOpInstallUnaliased(c *check.C) {
@@ -349,7 +551,7 @@ func (cs *clientSuite) TestClientOpInstallUnaliased(c *check.C) {
 	bodyData := []byte("snap-data")
 
 	snap := filepath.Join(c.MkDir(), "foo.snap")
-	err := ioutil.WriteFile(snap, bodyData, 0644)
+	err := os.WriteFile(snap, bodyData, 0644)
 	c.Assert(err, check.IsNil)
 
 	opts := client.SnapOptions{
@@ -359,7 +561,7 @@ func (cs *clientSuite) TestClientOpInstallUnaliased(c *check.C) {
 	_, err = cs.cli.Install("foo", &opts)
 	c.Assert(err, check.IsNil)
 
-	body, err := ioutil.ReadAll(cs.req.Body)
+	body, err := io.ReadAll(cs.req.Body)
 	c.Assert(err, check.IsNil)
 	jsonBody := make(map[string]interface{})
 	err = json.Unmarshal(body, &jsonBody)
@@ -369,12 +571,86 @@ func (cs *clientSuite) TestClientOpInstallUnaliased(c *check.C) {
 	_, err = cs.cli.InstallPath(snap, "", &opts)
 	c.Assert(err, check.IsNil)
 
-	body, err = ioutil.ReadAll(cs.req.Body)
+	body, err = io.ReadAll(cs.req.Body)
 	c.Assert(err, check.IsNil)
 
 	c.Assert(string(body), check.Matches, "(?s).*Content-Disposition: form-data; name=\"unaliased\"\r\n\r\ntrue\r\n.*")
 }
 
+func (cs *clientSuite) TestClientOpInstallTransactional(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "66b3",
+		"status-code": 202,
+		"type": "async"
+	}`
+	bodyData := []byte("snap-data")
+
+	snap := filepath.Join(c.MkDir(), "foo.snap")
+	err := os.WriteFile(snap, bodyData, 0644)
+	c.Assert(err, check.IsNil)
+
+	opts := client.SnapOptions{
+		Transaction: client.TransactionAllSnaps,
+	}
+
+	_, err = cs.cli.InstallMany([]string{"foo", "bar"}, &opts)
+	c.Assert(err, check.IsNil)
+
+	body, err := io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+	jsonBody := make(map[string]interface{})
+	err = json.Unmarshal(body, &jsonBody)
+	c.Assert(err, check.IsNil, check.Commentf("body: %v", string(body)))
+	c.Check(jsonBody["transaction"], check.Equals, string(client.TransactionAllSnaps),
+		check.Commentf("body: %v", string(body)))
+
+	_, err = cs.cli.InstallPath(snap, "", &opts)
+	c.Assert(err, check.IsNil)
+
+	body, err = io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+
+	c.Assert(string(body), check.Matches,
+		"(?s).*Content-Disposition: form-data; name=\"transaction\"\r\n\r\nall-snaps\r\n.*")
+}
+
+func (cs *clientSuite) TestClientOpInstallPrefer(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "66b3",
+		"status-code": 202,
+		"type": "async"
+	}`
+	bodyData := []byte("snap-data")
+
+	snap := filepath.Join(c.MkDir(), "foo.snap")
+	err := os.WriteFile(snap, bodyData, 0644)
+	c.Assert(err, check.IsNil)
+
+	opts := client.SnapOptions{
+		Prefer: true,
+	}
+
+	_, err = cs.cli.Install("foo", &opts)
+	c.Assert(err, check.IsNil)
+
+	body, err := io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+	var jsonBody map[string]interface{}
+	err = json.Unmarshal(body, &jsonBody)
+	c.Assert(err, check.IsNil, check.Commentf("body: %v", string(body)))
+	c.Check(jsonBody["prefer"], check.Equals, true, check.Commentf("body: %v", string(body)))
+
+	_, err = cs.cli.InstallPath(snap, "", &opts)
+	c.Assert(err, check.IsNil)
+
+	body, err = io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+
+	c.Assert(string(body), check.Matches, "(?s).*Content-Disposition: form-data; name=\"prefer\"\r\n\r\ntrue\r\n.*")
+}
+
 func formToMap(c *check.C, mr *multipart.Reader) map[string]string {
 	formData := map[string]string{}
 	for {
@@ -383,7 +659,7 @@ func formToMap(c *check.C, mr *multipart.Reader) map[string]string {
 			break
 		}
 		c.Assert(err, check.IsNil)
-		slurp, err := ioutil.ReadAll(p)
+		slurp, err := io.ReadAll(p)
 		c.Assert(err, check.IsNil)
 		formData[p.FormName()] = string(slurp)
 	}
@@ -464,6 +740,7 @@ func (cs *clientSuite) TestSnapOptionsSerialises(c *check.C) {
 		`{"unaliased":true}`:         {Unaliased: true},
 		`{"purge":true}`:             {Purge: true},
 		`{"amend":true}`:             {Amend: true},
+		`{"prefer":true}`:            {Prefer: true},
 	}
 	for expected, opts := range tests {
 		buf, err := json.Marshal(&opts)
@@ -501,7 +778,7 @@ func (cs *clientSuite) TestClientOpDownload(c *check.C) {
 	// check we posted the right stuff
 	c.Assert(cs.req.Header.Get("Content-Type"), check.Equals, "application/json")
 	c.Assert(cs.req.Header.Get("range"), check.Equals, "")
-	body, err := ioutil.ReadAll(cs.req.Body)
+	body, err := io.ReadAll(cs.req.Body)
 	c.Assert(err, check.IsNil)
 	var jsonBody client.DownloadAction
 	err = json.Unmarshal(body, &jsonBody)
@@ -512,7 +789,7 @@ func (cs *clientSuite) TestClientOpDownload(c *check.C) {
 	c.Check(jsonBody.HeaderPeek, check.Equals, true)
 
 	// ensure we can read the response
-	content, err := ioutil.ReadAll(rc)
+	content, err := io.ReadAll(rc)
 	c.Assert(err, check.IsNil)
 	c.Check(string(content), check.Equals, cs.rsp)
 	// and we can close it
@@ -549,7 +826,7 @@ func (cs *clientSuite) TestClientOpDownloadResume(c *check.C) {
 	// check we posted the right stuff
 	c.Assert(cs.req.Header.Get("Content-Type"), check.Equals, "application/json")
 	c.Assert(cs.req.Header.Get("range"), check.Equals, "bytes: 64-")
-	body, err := ioutil.ReadAll(cs.req.Body)
+	body, err := io.ReadAll(cs.req.Body)
 	c.Assert(err, check.IsNil)
 	var jsonBody client.DownloadAction
 	err = json.Unmarshal(body, &jsonBody)
@@ -561,9 +838,79 @@ func (cs *clientSuite) TestClientOpDownloadResume(c *check.C) {
 	c.Check(jsonBody.ResumeToken, check.Equals, "some-token")
 
 	// ensure we can read the response
-	content, err := ioutil.ReadAll(rc)
+	content, err := io.ReadAll(rc)
 	c.Assert(err, check.IsNil)
 	c.Check(string(content), check.Equals, cs.rsp)
 	// and we can close it
 	c.Check(rc.Close(), check.IsNil)
 }
+
+func (cs *clientSuite) TestClientRefreshWithValidationSets(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "12",
+		"status-code": 202,
+		"type": "async"
+	}`
+
+	sets := []string{"foo/bar=2", "foo/baz"}
+	chgID, err := cs.cli.RefreshMany(nil, &client.SnapOptions{
+		ValidationSets: sets,
+	})
+	c.Assert(err, check.IsNil)
+	c.Check(chgID, check.Equals, "12")
+
+	type req struct {
+		ValidationSets []string `json:"validation-sets"`
+		Action         string   `json:"action"`
+	}
+	body, err := io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+
+	var decodedBody req
+	err = json.Unmarshal(body, &decodedBody)
+	c.Assert(err, check.IsNil)
+
+	c.Check(decodedBody, check.DeepEquals, req{
+		ValidationSets: sets,
+		Action:         "refresh",
+	})
+	c.Check(cs.req.Header["Content-Type"], check.DeepEquals, []string{"application/json"})
+}
+
+func (cs *clientSuite) TestClientHoldMany(c *check.C) {
+	cs.status = 202
+	cs.rsp = `{
+		"change": "12",
+		"status-code": 202,
+		"type": "async"
+	}`
+
+	chgID, err := cs.cli.HoldRefreshesMany([]string{"foo", "bar"}, &client.SnapOptions{
+		Time:      "forever",
+		HoldLevel: "general",
+	})
+	c.Assert(err, check.IsNil)
+	c.Check(chgID, check.Equals, "12")
+
+	type req struct {
+		Action    string   `json:"action"`
+		Snaps     []string `json:"snaps"`
+		Time      string   `json:"time"`
+		HoldLevel string   `json:"hold-level"`
+	}
+	body, err := io.ReadAll(cs.req.Body)
+	c.Assert(err, check.IsNil)
+
+	var decodedBody req
+	err = json.Unmarshal(body, &decodedBody)
+	c.Assert(err, check.IsNil)
+
+	c.Check(decodedBody, check.DeepEquals, req{
+		Action:    "hold",
+		Snaps:     []string{"foo", "bar"},
+		Time:      "forever",
+		HoldLevel: "general",
+	})
+	c.Check(cs.req.Header["Content-Type"], check.DeepEquals, []string{"application/json"})
+}
