@@ -20,6 +20,7 @@
 package main
 
 import (
+	"context"
 	"os"
 	"os/user"
 	"time"
@@ -29,13 +30,14 @@ import (
 	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/cmd/snaplock/runinhibit"
 	"github.com/snapcore/snapd/image"
+	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/sandbox/cgroup"
 	"github.com/snapcore/snapd/sandbox/selinux"
+	"github.com/snapcore/snapd/seed/seedwriter"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/store"
 	"github.com/snapcore/snapd/store/tooling"
 	"github.com/snapcore/snapd/testutil"
-	usersessionclient "github.com/snapcore/snapd/usersession/client"
 )
 
 var RunMain = run
@@ -130,28 +132,27 @@ func SetVerbose(iw *infoWriter, verbose bool) {
 }
 
 var (
-	ClientSnapFromPath                            = clientSnapFromPath
-	SetupDiskSnap                                 = (*infoWriter).setupDiskSnap
-	SetupSnap                                     = (*infoWriter).setupSnap
-	MaybePrintServices                            = (*infoWriter).maybePrintServices
-	MaybePrintCommands                            = (*infoWriter).maybePrintCommands
-	MaybePrintType                                = (*infoWriter).maybePrintType
-	PrintSummary                                  = (*infoWriter).printSummary
-	MaybePrintPublisher                           = (*infoWriter).maybePrintPublisher
-	MaybePrintNotes                               = (*infoWriter).maybePrintNotes
-	MaybePrintStandaloneVersion                   = (*infoWriter).maybePrintStandaloneVersion
-	MaybePrintBuildDate                           = (*infoWriter).maybePrintBuildDate
-	MaybePrintLinks                               = (*infoWriter).maybePrintLinks
-	MaybePrintBase                                = (*infoWriter).maybePrintBase
-	MaybePrintPath                                = (*infoWriter).maybePrintPath
-	MaybePrintSum                                 = (*infoWriter).maybePrintSum
-	MaybePrintCohortKey                           = (*infoWriter).maybePrintCohortKey
-	MaybePrintHealth                              = (*infoWriter).maybePrintHealth
-	MaybePrintRefreshInfo                         = (*infoWriter).maybePrintRefreshInfo
-	WaitInhibitUnlock                             = waitInhibitUnlock
-	WaitWhileInhibited                            = waitWhileInhibited
-	IsLocked                                      = isLocked
-	TryNotifyRefreshViaSnapDesktopIntegrationFlow = tryNotifyRefreshViaSnapDesktopIntegrationFlow
+	ClientSnapFromPath          = clientSnapFromPath
+	SetupDiskSnap               = (*infoWriter).setupDiskSnap
+	SetupSnap                   = (*infoWriter).setupSnap
+	MaybePrintServices          = (*infoWriter).maybePrintServices
+	MaybePrintCommands          = (*infoWriter).maybePrintCommands
+	MaybePrintType              = (*infoWriter).maybePrintType
+	PrintSummary                = (*infoWriter).printSummary
+	MaybePrintPublisher         = (*infoWriter).maybePrintPublisher
+	MaybePrintNotes             = (*infoWriter).maybePrintNotes
+	MaybePrintStandaloneVersion = (*infoWriter).maybePrintStandaloneVersion
+	MaybePrintBuildDate         = (*infoWriter).maybePrintBuildDate
+	MaybePrintLinks             = (*infoWriter).maybePrintLinks
+	MaybePrintBase              = (*infoWriter).maybePrintBase
+	MaybePrintPath              = (*infoWriter).maybePrintPath
+	MaybePrintSum               = (*infoWriter).maybePrintSum
+	MaybePrintCohortKey         = (*infoWriter).maybePrintCohortKey
+	MaybePrintHealth            = (*infoWriter).maybePrintHealth
+	MaybePrintRefreshInfo       = (*infoWriter).maybePrintRefreshInfo
+	WaitWhileInhibited          = waitWhileInhibited
+	NewInhibitionFlow           = newInhibitionFlow
+	ErrSnapRefreshConflict      = errSnapRefreshConflict
 )
 
 func MockPollTime(d time.Duration) (restore func()) {
@@ -355,6 +356,14 @@ func MockConfirmSystemdServiceTracking(fn func(securityTag string) error) (resto
 	}
 }
 
+func MockConfirmSystemdAppTracking(fn func(securityTag string) error) (restore func()) {
+	old := cgroupConfirmSystemdAppTracking
+	cgroupConfirmSystemdAppTracking = fn
+	return func() {
+		cgroupConfirmSystemdAppTracking = old
+	}
+}
+
 func MockApparmorSnapAppFromPid(f func(pid int) (string, string, string, error)) (restore func()) {
 	old := apparmorSnapAppFromPid
 	apparmorSnapAppFromPid = f
@@ -380,10 +389,10 @@ func MockSyscallUmount(f func(string, int) error) (restore func()) {
 }
 
 func MockIoutilTempDir(f func(string, string) (string, error)) (restore func()) {
-	old := ioutilTempDir
-	ioutilTempDir = f
+	old := osMkdirTemp
+	osMkdirTemp = f
 	return func() {
-		ioutilTempDir = old
+		osMkdirTemp = old
 	}
 }
 
@@ -419,53 +428,19 @@ func MockOsChmod(f func(string, os.FileMode) error) (restore func()) {
 	}
 }
 
-func MockWaitInhibitUnlock(f func(snapName string, waitFor runinhibit.Hint) (bool, error)) (restore func()) {
-	old := waitInhibitUnlock
-	waitInhibitUnlock = f
-	return func() {
-		waitInhibitUnlock = old
-	}
+func MockWaitWhileInhibited(f func(ctx context.Context, snapName string, notInhibited func(ctx context.Context) error, inhibited func(ctx context.Context, hint runinhibit.Hint, inhibitInfo *runinhibit.InhibitInfo) (cont bool, err error), interval time.Duration) (flock *osutil.FileLock, retErr error)) (restore func()) {
+	restore = testutil.Backup(&runinhibitWaitWhileInhibited)
+	runinhibitWaitWhileInhibited = f
+	return restore
 }
 
-func MockIsLocked(f func(snapName string) (runinhibit.Hint, error)) (restore func()) {
-	old := isLocked
-	isLocked = f
-	return func() {
-		isLocked = old
-	}
-}
-
-func MockIsGraphicalSession(graphical bool) (restore func()) {
-	old := isGraphicalSession
-	isGraphicalSession = func() bool {
-		return graphical
+func MockInhibitionFlow(flow inhibitionFlow) (restore func()) {
+	old := newInhibitionFlow
+	newInhibitionFlow = func(cli *client.Client, name string) inhibitionFlow {
+		return flow
 	}
 	return func() {
-		isGraphicalSession = old
-	}
-}
-
-func MockPendingRefreshNotification(f func(refreshInfo *usersessionclient.PendingSnapRefreshInfo) error) (restore func()) {
-	old := pendingRefreshNotification
-	pendingRefreshNotification = f
-	return func() {
-		pendingRefreshNotification = old
-	}
-}
-
-func MockFinishRefreshNotification(f func(refreshInfo *usersessionclient.FinishedSnapRefreshInfo) error) (restore func()) {
-	old := finishRefreshNotification
-	finishRefreshNotification = f
-	return func() {
-		finishRefreshNotification = old
-	}
-}
-
-func MockTryNotifyRefreshViaSnapDesktopIntegrationFlow(f func(snapName string) (bool, error)) (restore func()) {
-	old := tryNotifyRefreshViaSnapDesktopIntegrationFlow
-	tryNotifyRefreshViaSnapDesktopIntegrationFlow = f
-	return func() {
-		tryNotifyRefreshViaSnapDesktopIntegrationFlow = old
+		newInhibitionFlow = old
 	}
 }
 
@@ -490,8 +465,8 @@ func ParseQuotaValues(maxMemory, cpuMax, cpuSet, threadsMax, journalSizeMax, jou
 	return quotas.parseQuotas()
 }
 
-func MockImageReadSeedManifest(f func(manifestFile string) (map[string]snap.Revision, error)) (restore func()) {
-	restore = testutil.Backup(&imageReadSeedManifest)
-	imageReadSeedManifest = f
+func MockSeedWriterReadManifest(f func(manifestFile string) (*seedwriter.Manifest, error)) (restore func()) {
+	restore = testutil.Backup(&seedwriterReadManifest)
+	seedwriterReadManifest = f
 	return restore
 }
