@@ -21,14 +21,15 @@ package daemon_test
 
 import (
 	"bytes"
+	"context"
 	"encoding/json"
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"math"
 	"net/http"
 	"net/http/httptest"
+	"os/user"
 	"sort"
 	"strconv"
 	"strings"
@@ -36,6 +37,7 @@ import (
 	"gopkg.in/check.v1"
 
 	"github.com/snapcore/snapd/client"
+	"github.com/snapcore/snapd/client/clientutil"
 	"github.com/snapcore/snapd/daemon"
 	"github.com/snapcore/snapd/overlord/hookstate"
 	"github.com/snapcore/snapd/overlord/servicestate"
@@ -48,6 +50,12 @@ import (
 
 var _ = check.Suite(&appsSuite{})
 
+type appsSuiteDecoratorResult struct {
+	daemonType string
+	active     bool
+	enabled    bool
+}
+
 type appsSuite struct {
 	apiBaseSuite
 
@@ -55,8 +63,10 @@ type appsSuite struct {
 	jctlSvcses         [][]string
 	jctlNs             []int
 	jctlFollows        []bool
+	jctlNamespaces     []bool
 	jctlRCs            []io.ReadCloser
 	jctlErrs           []error
+	decoratorResults   map[string]appsSuiteDecoratorResult
 
 	serviceControlError error
 	serviceControlCalls []serviceControlArgs
@@ -64,10 +74,11 @@ type appsSuite struct {
 	infoA, infoB, infoC, infoD, infoE *snap.Info
 }
 
-func (s *appsSuite) journalctl(svcs []string, n int, follow bool) (rc io.ReadCloser, err error) {
+func (s *appsSuite) journalctl(svcs []string, n int, follow, namespaces bool) (rc io.ReadCloser, err error) {
 	s.jctlSvcses = append(s.jctlSvcses, svcs)
 	s.jctlNs = append(s.jctlNs, n)
 	s.jctlFollows = append(s.jctlFollows, follow)
+	s.jctlNamespaces = append(s.jctlNamespaces, namespaces)
 
 	if len(s.jctlErrs) > 0 {
 		err, s.jctlErrs = s.jctlErrs[0], s.jctlErrs[1:]
@@ -83,9 +94,11 @@ type serviceControlArgs struct {
 	action  string
 	options string
 	names   []string
+	scope   client.ScopeSelector
+	users   []string
 }
 
-func (s *appsSuite) fakeServiceControl(st *state.State, appInfos []*snap.AppInfo, inst *servicestate.Instruction, flags *servicestate.Flags, context *hookstate.Context) ([]*state.TaskSet, error) {
+func (s *appsSuite) fakeServiceControl(st *state.State, appInfos []*snap.AppInfo, inst *servicestate.Instruction, cu *user.User, flags *servicestate.Flags, context *hookstate.Context) ([]*state.TaskSet, error) {
 	if flags != nil {
 		panic("flags are not expected")
 	}
@@ -94,12 +107,17 @@ func (s *appsSuite) fakeServiceControl(st *state.State, appInfos []*snap.AppInfo
 		return nil, s.serviceControlError
 	}
 
-	serviceCommand := serviceControlArgs{action: inst.Action}
+	users, err := inst.Users.UserList(cu)
+	if err != nil {
+		return nil, err
+	}
+
+	serviceCommand := serviceControlArgs{action: inst.Action, scope: inst.Scope, users: users}
 	if inst.RestartOptions.Reload {
 		serviceCommand.options = "reload"
 	}
 	// only one flag should ever be set (depending on Action), but appending
-	// them below acts as an extra sanity check.
+	// them below acts as an extra validity check.
 	if inst.StartOptions.Enable {
 		serviceCommand.options += "enable"
 	}
@@ -111,7 +129,7 @@ func (s *appsSuite) fakeServiceControl(st *state.State, appInfos []*snap.AppInfo
 	}
 	s.serviceControlCalls = append(s.serviceControlCalls, serviceCommand)
 
-	t := st.NewTask("dummy", "")
+	t := st.NewTask("sample", "")
 	ts := state.NewTaskSet(t)
 	return []*state.TaskSet{ts}, nil
 }
@@ -122,7 +140,9 @@ func (s *appsSuite) SetUpSuite(c *check.C) {
 }
 
 func (s *appsSuite) TearDownSuite(c *check.C) {
-	s.journalctlRestorer()
+	if s.journalctlRestorer != nil {
+		s.journalctlRestorer()
+	}
 	s.apiBaseSuite.TearDownSuite(c)
 }
 
@@ -132,6 +152,7 @@ func (s *appsSuite) SetUpTest(c *check.C) {
 	s.jctlSvcses = nil
 	s.jctlNs = nil
 	s.jctlFollows = nil
+	s.jctlNamespaces = nil
 	s.jctlRCs = nil
 	s.jctlErrs = nil
 
@@ -145,6 +166,7 @@ func (s *appsSuite) SetUpTest(c *check.C) {
 	// turn off ensuring snap services which will call systemctl automatically
 	r := servicestate.MockEnsuredSnapServices(s.d.Overlord().ServiceManager(), true)
 	s.AddCleanup(r)
+	s.AddCleanup(snapstate.MockEnsuredMountsUpdated(d.Overlord().SnapManager(), true))
 
 	s.infoA = s.mkInstalledInState(c, s.d, "snap-a", "dev", "v1", snap.R(1), true, "apps: {svc1: {daemon: simple}, svc2: {daemon: simple, reload-command: x}}")
 	s.infoB = s.mkInstalledInState(c, s.d, "snap-b", "dev", "v1", snap.R(1), false, "apps: {svc3: {daemon: simple}, cmd1: {}}")
@@ -154,6 +176,13 @@ func (s *appsSuite) SetUpTest(c *check.C) {
 
 	d.Overlord().Loop()
 	s.AddCleanup(func() { d.Overlord().Stop() })
+	s.AddCleanup(systemd.MockSystemdVersion(237, nil))
+	s.expectAppsAccess()
+}
+
+func (s *appsSuite) expectAppsAccess() {
+	s.expectOpenAccess()
+	s.expectWriteAccess(daemon.AuthenticatedAccess{Polkit: "io.snapcraft.snapd.manage"})
 }
 
 func (s *appsSuite) TestSplitAppName(c *check.C) {
@@ -185,6 +214,7 @@ Names=snap.%[1]s.service
 Type=simple
 ActiveState=active
 UnitFileState=enabled
+NeedDaemonReload=no
 `[1:], name)))
 	}
 	// System services from inactive snaps
@@ -193,7 +223,7 @@ UnitFileState=enabled
 	svcNames = append(svcNames, "snap-e.svc4")
 	s.SysctlBufs = append(s.SysctlBufs, []byte("enabled\n"))
 
-	req, err := http.NewRequest("GET", "/v2/apps", nil)
+	req, err := http.NewRequest("GET", "/v2/apps?global=true", nil)
 	c.Assert(err, check.IsNil)
 
 	rsp := s.syncReq(c, req, nil)
@@ -265,22 +295,38 @@ func (s *appsSuite) TestGetAppsInfoNames(c *check.C) {
 }
 
 func (s *appsSuite) TestGetAppsInfoServices(c *check.C) {
+	r := daemon.MockNewStatusDecorator(func(ctx context.Context, isGlobal bool, uid string) clientutil.StatusDecorator {
+		c.Check(isGlobal, check.Equals, false)
+		c.Check(uid, check.Equals, "0")
+		return s
+	})
+	defer r()
+
+	// System and user services from active snaps
+	s.decoratorResults = map[string]appsSuiteDecoratorResult{
+		"snap-a.svc1": {
+			daemonType: "simple",
+			active:     true,
+			enabled:    true,
+		},
+		"snap-a.svc2": {
+			daemonType: "simple",
+			active:     true,
+			enabled:    true,
+		},
+		"snap-e.svc4": {
+			daemonType: "simple",
+			active:     false,
+			enabled:    true,
+		},
+	}
+
 	// System services from active snaps
 	svcNames := []string{"snap-a.svc1", "snap-a.svc2"}
-	for _, name := range svcNames {
-		s.SysctlBufs = append(s.SysctlBufs, []byte(fmt.Sprintf(`
-Id=snap.%s.service
-Names=snap.%[1]s.service
-Type=simple
-ActiveState=active
-UnitFileState=enabled
-`[1:], name)))
-	}
 	// System services from inactive snaps
 	svcNames = append(svcNames, "snap-b.svc3")
 	// User services from active snaps
 	svcNames = append(svcNames, "snap-e.svc4")
-	s.SysctlBufs = append(s.SysctlBufs, []byte("enabled\n"))
 
 	req, err := http.NewRequest("GET", "/v2/apps?select=service", nil)
 	c.Assert(err, check.IsNil)
@@ -319,6 +365,141 @@ UnitFileState=enabled
 	c.Check(sort.StringsAreSorted(appNames), check.Equals, true)
 }
 
+func (s *appsSuite) TestGetAppsInfoServicesWithGlobal(c *check.C) {
+	// System services from active snaps
+	svcNames := []string{"snap-a.svc1", "snap-a.svc2"}
+	for _, name := range svcNames {
+		s.SysctlBufs = append(s.SysctlBufs, []byte(fmt.Sprintf(`
+Id=snap.%s.service
+Names=snap.%[1]s.service
+Type=simple
+ActiveState=active
+UnitFileState=enabled
+NeedDaemonReload=no
+`[1:], name)))
+	}
+	// System services from inactive snaps
+	svcNames = append(svcNames, "snap-b.svc3")
+	// User services from active snaps
+	svcNames = append(svcNames, "snap-e.svc4")
+	s.SysctlBufs = append(s.SysctlBufs, []byte("enabled\n"))
+
+	req, err := http.NewRequest("GET", "/v2/apps?select=service&global=true", nil)
+	c.Assert(err, check.IsNil)
+	s.asUserAuth(c, req)
+
+	rsp := s.syncReq(c, req, nil)
+	c.Assert(rsp.Status, check.Equals, 200)
+	c.Assert(rsp.Result, check.FitsTypeOf, []client.AppInfo{})
+	svcs := rsp.Result.([]client.AppInfo)
+	c.Assert(svcs, check.HasLen, 4)
+
+	for _, name := range svcNames {
+		snapName, app := daemon.SplitAppName(name)
+		needle := client.AppInfo{
+			Snap:        snapName,
+			Name:        app,
+			Daemon:      "simple",
+			DaemonScope: snap.SystemDaemon,
+		}
+		if snapName != "snap-b" {
+			// snap-b is not active (all the others are)
+			needle.Active = true
+			needle.Enabled = true
+		}
+		if snapName == "snap-e" {
+			// snap-e contains user services
+			needle.DaemonScope = snap.UserDaemon
+			needle.Active = false
+		}
+		c.Check(svcs, testutil.DeepContains, needle)
+	}
+
+	appNames := make([]string, len(svcs))
+	for i, svc := range svcs {
+		appNames[i] = svc.Snap + "." + svc.Name
+	}
+	c.Check(sort.StringsAreSorted(appNames), check.Equals, true)
+}
+
+func (s *appsSuite) DecorateWithStatus(appInfo *client.AppInfo, snapApp *snap.AppInfo) error {
+	name := snapApp.Snap.RealName + "." + appInfo.Name
+	dec, ok := s.decoratorResults[name]
+	if !ok {
+		return fmt.Errorf("%s not found in expected test decorator results", name)
+	}
+	appInfo.Daemon = dec.daemonType
+	appInfo.Enabled = dec.enabled
+	appInfo.Active = dec.active
+	return nil
+}
+
+func (s *appsSuite) TestGetUserAppsInfoServices(c *check.C) {
+	r := daemon.MockNewStatusDecorator(func(ctx context.Context, isGlobal bool, uid string) clientutil.StatusDecorator {
+		c.Check(isGlobal, check.Equals, false)
+		c.Check(uid, check.Equals, "1337")
+		return s
+	})
+	defer r()
+
+	// System and user services from active snaps
+	s.decoratorResults = map[string]appsSuiteDecoratorResult{
+		"snap-a.svc1": {
+			daemonType: "simple",
+			active:     true,
+			enabled:    true,
+		},
+		"snap-a.svc2": {
+			daemonType: "simple",
+			active:     true,
+			enabled:    true,
+		},
+		"snap-e.svc4": {
+			daemonType: "simple",
+			active:     true,
+			enabled:    true,
+		},
+	}
+
+	// Perform the request as a non-root uid
+	req, err := http.NewRequest("GET", "/v2/apps?select=service", nil)
+	c.Assert(err, check.IsNil)
+	s.asUserAuth(c, req)
+
+	rsp := s.syncReq(c, req, nil)
+	c.Assert(rsp.Status, check.Equals, 200)
+	c.Assert(rsp.Result, check.FitsTypeOf, []client.AppInfo{})
+	svcs := rsp.Result.([]client.AppInfo)
+	c.Assert(svcs, check.HasLen, 4)
+
+	for name := range s.decoratorResults {
+		snapName, app := daemon.SplitAppName(name)
+		needle := client.AppInfo{
+			Snap:        snapName,
+			Name:        app,
+			Daemon:      "simple",
+			DaemonScope: snap.SystemDaemon,
+		}
+		if snapName != "snap-b" {
+			// snap-b is not active (all the others are)
+			needle.Active = true
+			needle.Enabled = true
+		}
+		if snapName == "snap-e" {
+			// snap-e contains user services
+			needle.DaemonScope = snap.UserDaemon
+			needle.Active = true
+		}
+		c.Check(svcs, testutil.DeepContains, needle)
+	}
+
+	appNames := make([]string, len(svcs))
+	for i, svc := range svcs {
+		appNames[i] = svc.Snap + "." + svc.Name
+	}
+	c.Check(sort.StringsAreSorted(appNames), check.Equals, true)
+}
+
 func (s *appsSuite) TestGetAppsInfoBadSelect(c *check.C) {
 	req, err := http.NewRequest("GET", "/v2/apps?select=potato", nil)
 	c.Assert(err, check.IsNil)
@@ -464,7 +645,41 @@ func (s *appsSuite) testPostApps(c *check.C, inst servicestate.Instruction, serv
 	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBuffer(postBody))
 	c.Assert(err, check.IsNil)
 
-	rsp := s.asyncReq(c, req, nil)
+	rsp := s.asyncReq(c, req, s.authUser)
+	c.Assert(rsp.Status, check.Equals, 202)
+	c.Check(rsp.Change, check.Matches, `[0-9]+`)
+
+	st := s.d.Overlord().State()
+	st.Lock()
+	defer st.Unlock()
+	chg := st.Change(rsp.Change)
+	c.Assert(chg, check.NotNil)
+	c.Check(chg.Tasks(), check.HasLen, len(servicecmds))
+
+	st.Unlock()
+	<-chg.Ready()
+	st.Lock()
+
+	c.Check(s.serviceControlCalls, check.DeepEquals, servicecmds)
+	return chg
+}
+
+func (s *appsSuite) testPostAppsUser(c *check.C, inst servicestate.Instruction, servicecmds []serviceControlArgs, expectedErr string) *state.Change {
+	postBody, err := json.Marshal(inst)
+	c.Assert(err, check.IsNil)
+
+	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBuffer(postBody))
+	c.Assert(err, check.IsNil)
+	s.asUserAuth(c, req)
+
+	if expectedErr != "" {
+		rspe := s.errorReq(c, req, s.authUser)
+		c.Check(rspe.Status, check.Equals, 400)
+		c.Check(rspe.Message, check.Matches, expectedErr)
+		return nil
+	}
+
+	rsp := s.asyncReq(c, req, s.authUser)
 	c.Assert(rsp.Status, check.Equals, 202)
 	c.Check(rsp.Change, check.Matches, `[0-9]+`)
 
@@ -486,7 +701,7 @@ func (s *appsSuite) testPostApps(c *check.C, inst servicestate.Instruction, serv
 func (s *appsSuite) TestPostAppsStartOne(c *check.C) {
 	inst := servicestate.Instruction{Action: "start", Names: []string{"snap-a.svc2"}}
 	expected := []serviceControlArgs{
-		{action: "start", names: []string{"snap-a.svc2"}},
+		{action: "start", names: []string{"snap-a.svc2"}, scope: client.ScopeSelector{"system", "user"}},
 	}
 	chg := s.testPostApps(c, inst, expected)
 	chg.State().Lock()
@@ -501,7 +716,7 @@ func (s *appsSuite) TestPostAppsStartOne(c *check.C) {
 func (s *appsSuite) TestPostAppsStartTwo(c *check.C) {
 	inst := servicestate.Instruction{Action: "start", Names: []string{"snap-a"}}
 	expected := []serviceControlArgs{
-		{action: "start", names: []string{"snap-a.svc1", "snap-a.svc2"}},
+		{action: "start", names: []string{"snap-a.svc1", "snap-a.svc2"}, scope: client.ScopeSelector{"system", "user"}},
 	}
 	chg := s.testPostApps(c, inst, expected)
 
@@ -517,7 +732,7 @@ func (s *appsSuite) TestPostAppsStartTwo(c *check.C) {
 func (s *appsSuite) TestPostAppsStartThree(c *check.C) {
 	inst := servicestate.Instruction{Action: "start", Names: []string{"snap-a", "snap-b"}}
 	expected := []serviceControlArgs{
-		{action: "start", names: []string{"snap-a.svc1", "snap-a.svc2", "snap-b.svc3"}},
+		{action: "start", names: []string{"snap-a.svc1", "snap-a.svc2", "snap-b.svc3"}, scope: client.ScopeSelector{"system", "user"}},
 	}
 	chg := s.testPostApps(c, inst, expected)
 	// check the summary expands the snap into actual apps
@@ -534,7 +749,7 @@ func (s *appsSuite) TestPostAppsStartThree(c *check.C) {
 func (s *appsSuite) TestPostAppsStop(c *check.C) {
 	inst := servicestate.Instruction{Action: "stop", Names: []string{"snap-a.svc2"}}
 	expected := []serviceControlArgs{
-		{action: "stop", names: []string{"snap-a.svc2"}},
+		{action: "stop", names: []string{"snap-a.svc2"}, scope: client.ScopeSelector{"system", "user"}},
 	}
 	s.testPostApps(c, inst, expected)
 }
@@ -542,7 +757,7 @@ func (s *appsSuite) TestPostAppsStop(c *check.C) {
 func (s *appsSuite) TestPostAppsRestart(c *check.C) {
 	inst := servicestate.Instruction{Action: "restart", Names: []string{"snap-a.svc2"}}
 	expected := []serviceControlArgs{
-		{action: "restart", names: []string{"snap-a.svc2"}},
+		{action: "restart", names: []string{"snap-a.svc2"}, scope: client.ScopeSelector{"system", "user"}},
 	}
 	s.testPostApps(c, inst, expected)
 }
@@ -551,7 +766,7 @@ func (s *appsSuite) TestPostAppsReload(c *check.C) {
 	inst := servicestate.Instruction{Action: "restart", Names: []string{"snap-a.svc2"}}
 	inst.Reload = true
 	expected := []serviceControlArgs{
-		{action: "restart", options: "reload", names: []string{"snap-a.svc2"}},
+		{action: "restart", options: "reload", names: []string{"snap-a.svc2"}, scope: client.ScopeSelector{"system", "user"}},
 	}
 	s.testPostApps(c, inst, expected)
 }
@@ -560,7 +775,7 @@ func (s *appsSuite) TestPostAppsEnableNow(c *check.C) {
 	inst := servicestate.Instruction{Action: "start", Names: []string{"snap-a.svc2"}}
 	inst.Enable = true
 	expected := []serviceControlArgs{
-		{action: "start", options: "enable", names: []string{"snap-a.svc2"}},
+		{action: "start", options: "enable", names: []string{"snap-a.svc2"}, scope: client.ScopeSelector{"system", "user"}},
 	}
 	s.testPostApps(c, inst, expected)
 }
@@ -569,11 +784,136 @@ func (s *appsSuite) TestPostAppsDisableNow(c *check.C) {
 	inst := servicestate.Instruction{Action: "stop", Names: []string{"snap-a.svc2"}}
 	inst.Disable = true
 	expected := []serviceControlArgs{
-		{action: "stop", options: "disable", names: []string{"snap-a.svc2"}},
+		{action: "stop", options: "disable", names: []string{"snap-a.svc2"}, scope: client.ScopeSelector{"system", "user"}},
 	}
 	s.testPostApps(c, inst, expected)
 }
 
+func (s *appsSuite) TestPostAppsFailedToGetUser(c *check.C) {
+	r := daemon.MockSystemUserFromRequest(func(r *http.Request) (*user.User, error) {
+		return nil, fmt.Errorf("failed")
+	})
+	defer r()
+
+	req, err := http.NewRequest("POST", "/v2/apps", nil)
+	c.Assert(err, check.IsNil)
+	rspe := s.errorReq(c, req, nil)
+	c.Check(rspe.Status, check.Equals, 400)
+	c.Check(rspe.Message, check.Matches, "cannot perform operation on services: failed")
+}
+
+func (s *appsSuite) TestPostAppsScopesSelfAsRootNotAllowed(c *check.C) {
+	inst := servicestate.Instruction{
+		Action: "start",
+		Names:  []string{"snap-a.svc1"},
+		Scope:  client.ScopeSelector{"user"},
+		Users: client.UserSelector{
+			Selector: client.UserSelectionSelf,
+		},
+	}
+	postBody, err := json.Marshal(inst)
+	c.Assert(err, check.IsNil)
+
+	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBuffer(postBody))
+	c.Assert(err, check.IsNil)
+
+	rspe := s.errorReq(c, req, s.authUser)
+	c.Check(rspe.Status, check.Equals, 400)
+	c.Check(rspe.Message, check.Matches, `cannot use "self" for root user`)
+}
+
+func (s *appsSuite) TestPostAppsAllUsersAsRootHappy(c *check.C) {
+	inst := servicestate.Instruction{
+		Action: "start",
+		Names:  []string{"snap-a.svc1"},
+		Scope:  client.ScopeSelector{"user"},
+		Users: client.UserSelector{
+			Selector: client.UserSelectionAll,
+		},
+	}
+	expected := []serviceControlArgs{
+		// Expect no user to appear as we are not logged in
+		{action: "start", names: []string{"snap-a.svc1"}, scope: client.ScopeSelector{"user"}},
+	}
+	s.testPostApps(c, inst, expected)
+}
+
+func (s *appsSuite) TestPostAppsScopesNotSpecifiedForRoot(c *check.C) {
+	inst := servicestate.Instruction{Action: "start", Names: []string{"snap-e.svc4"}}
+	expected := []serviceControlArgs{
+		{action: "start", names: []string{"snap-e.svc4"}, scope: client.ScopeSelector{"system", "user"}},
+	}
+	s.testPostApps(c, inst, expected)
+}
+
+func (s *appsSuite) TestPostAppsUsersAsUserHappy(c *check.C) {
+	inst := servicestate.Instruction{
+		Action: "start",
+		Names:  []string{"snap-a.svc1"},
+		Scope:  client.ScopeSelector{"user"},
+		Users: client.UserSelector{
+			Selector: client.UserSelectionAll,
+		},
+	}
+	expected := []serviceControlArgs{
+		{action: "start", names: []string{"snap-a.svc1"}, scope: client.ScopeSelector{"user"}},
+	}
+	s.testPostAppsUser(c, inst, expected, "")
+}
+
+func (s *appsSuite) TestPostAppsScopesNotSpecifiedForUser(c *check.C) {
+	inst := servicestate.Instruction{
+		Action: "start",
+		Names:  []string{"snap-e.svc4"},
+		Users: client.UserSelector{
+			Selector: client.UserSelectionSelf,
+		},
+	}
+	s.testPostAppsUser(c, inst, nil, "cannot perform operation on services: non-root users must specify service scope when targeting user services")
+}
+
+func (s *appsSuite) TestPostAppsUsersUser(c *check.C) {
+	inst := servicestate.Instruction{
+		Action: "start",
+		Names:  []string{"snap-a.svc1"},
+		Users: client.UserSelector{
+			Selector: client.UserSelectionSelf,
+		},
+	}
+	expected := []serviceControlArgs{
+		{action: "start", names: []string{"snap-a.svc1"}, scope: client.ScopeSelector{"system"}, users: []string{"username"}},
+	}
+	s.testPostAppsUser(c, inst, expected, "")
+}
+
+func (s *appsSuite) TestPostAppsUsersWithUsernames(c *check.C) {
+	inst := servicestate.Instruction{
+		Action: "start",
+		Names:  []string{"snap-a.svc1"},
+		Users: client.UserSelector{
+			Names: []string{"my-user", "other-user"},
+		},
+	}
+	expected := []serviceControlArgs{
+		// Expect no user to appear as we are not logged in
+		{action: "start", names: []string{"snap-a.svc1"}, scope: client.ScopeSelector{"system", "user"}, users: []string{"my-user", "other-user"}},
+	}
+	s.testPostApps(c, inst, expected)
+}
+
+func (s *appsSuite) TestPostAppsUserNotSpecifiedForRoot(c *check.C) {
+	inst := servicestate.Instruction{Action: "start", Names: []string{"snap-e.svc4"}, Scope: client.ScopeSelector{"system"}}
+	expected := []serviceControlArgs{
+		{action: "start", names: []string{"snap-e.svc4"}, scope: client.ScopeSelector{"system"}},
+	}
+	s.testPostApps(c, inst, expected)
+}
+
+func (s *appsSuite) TestPostAppsUserNotSpecifiedForUser(c *check.C) {
+	inst := servicestate.Instruction{Action: "start", Names: []string{"snap-e.svc4"}, Scope: client.ScopeSelector{"user"}}
+	s.testPostAppsUser(c, inst, nil, "cannot perform operation on services: non-root users must specify users when targeting user services")
+}
+
 func (s *appsSuite) TestPostAppsBadJSON(c *check.C) {
 	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBufferString(`'junk`))
 	c.Assert(err, check.IsNil)
@@ -631,7 +971,7 @@ func (s *appsSuite) expectLogsAccess() {
 func (s *appsSuite) TestLogs(c *check.C) {
 	s.expectLogsAccess()
 
-	s.jctlRCs = []io.ReadCloser{ioutil.NopCloser(strings.NewReader(`
+	s.jctlRCs = []io.ReadCloser{io.NopCloser(strings.NewReader(`
 {"MESSAGE": "hello1", "SYSLOG_IDENTIFIER": "xyzzy", "_PID": "42", "__REALTIME_TIMESTAMP": "42"}
 {"MESSAGE": "hello2", "SYSLOG_IDENTIFIER": "xyzzy", "_PID": "42", "__REALTIME_TIMESTAMP": "44"}
 {"MESSAGE": "hello3", "SYSLOG_IDENTIFIER": "xyzzy", "_PID": "42", "__REALTIME_TIMESTAMP": "46"}
@@ -648,9 +988,10 @@ func (s *appsSuite) TestLogs(c *check.C) {
 	c.Check(s.jctlSvcses, check.DeepEquals, [][]string{{"snap.snap-a.svc2.service"}})
 	c.Check(s.jctlNs, check.DeepEquals, []int{42})
 	c.Check(s.jctlFollows, check.DeepEquals, []bool{false})
+	c.Check(s.jctlNamespaces, check.DeepEquals, []bool{false})
 
 	c.Check(rec.Code, check.Equals, 200)
-	c.Check(rec.HeaderMap.Get("Content-Type"), check.Equals, "application/json-seq")
+	c.Check(rec.Header().Get("Content-Type"), check.Equals, "application/json-seq")
 	c.Check(rec.Body.String(), check.Equals, `
 {"timestamp":"1970-01-01T00:00:00.000042Z","message":"hello1","sid":"xyzzy","pid":"42"}
 {"timestamp":"1970-01-01T00:00:00.000044Z","message":"hello2","sid":"xyzzy","pid":"42"}
@@ -660,6 +1001,54 @@ func (s *appsSuite) TestLogs(c *check.C) {
 `[1:])
 }
 
+func (s *appsSuite) TestLogsNoNamespaceOption(c *check.C) {
+	restore := systemd.MockSystemdVersion(237, nil)
+	defer restore()
+
+	s.expectLogsAccess()
+
+	s.jctlRCs = []io.ReadCloser{io.NopCloser(strings.NewReader(""))}
+
+	req, err := http.NewRequest("GET", "/v2/logs?names=snap-a.svc2&n=42&follow=false", nil)
+	c.Assert(err, check.IsNil)
+
+	rec := httptest.NewRecorder()
+	s.req(c, req, nil).ServeHTTP(rec, req)
+
+	c.Check(s.jctlSvcses, check.DeepEquals, [][]string{{"snap.snap-a.svc2.service"}})
+	c.Check(s.jctlNs, check.DeepEquals, []int{42})
+	c.Check(s.jctlFollows, check.DeepEquals, []bool{false})
+	c.Check(s.jctlNamespaces, check.DeepEquals, []bool{false})
+
+	c.Check(rec.Code, check.Equals, 200)
+	c.Check(rec.Header().Get("Content-Type"), check.Equals, "application/json-seq")
+	c.Check(rec.Body.String(), check.Equals, "")
+}
+
+func (s *appsSuite) TestLogsWithNamespaceOption(c *check.C) {
+	restore := systemd.MockSystemdVersion(245, nil)
+	defer restore()
+
+	s.expectLogsAccess()
+
+	s.jctlRCs = []io.ReadCloser{io.NopCloser(strings.NewReader(""))}
+
+	req, err := http.NewRequest("GET", "/v2/logs?names=snap-a.svc2&n=42&follow=false", nil)
+	c.Assert(err, check.IsNil)
+
+	rec := httptest.NewRecorder()
+	s.req(c, req, nil).ServeHTTP(rec, req)
+
+	c.Check(s.jctlSvcses, check.DeepEquals, [][]string{{"snap.snap-a.svc2.service"}})
+	c.Check(s.jctlNs, check.DeepEquals, []int{42})
+	c.Check(s.jctlFollows, check.DeepEquals, []bool{false})
+	c.Check(s.jctlNamespaces, check.DeepEquals, []bool{true})
+
+	c.Check(rec.Code, check.Equals, 200)
+	c.Check(rec.Header().Get("Content-Type"), check.Equals, "application/json-seq")
+	c.Check(rec.Body.String(), check.Equals, "")
+}
+
 func (s *appsSuite) TestLogsN(c *check.C) {
 	s.expectLogsAccess()
 
@@ -676,7 +1065,7 @@ func (s *appsSuite) TestLogsN(c *check.C) {
 		{in: strconv.Itoa(math.MaxInt32), out: math.MaxInt32},
 	} {
 
-		s.jctlRCs = []io.ReadCloser{ioutil.NopCloser(strings.NewReader(""))}
+		s.jctlRCs = []io.ReadCloser{io.NopCloser(strings.NewReader(""))}
 		s.jctlNs = nil
 
 		req, err := http.NewRequest("GET", "/v2/logs?n="+t.in, nil)
@@ -703,9 +1092,9 @@ func (s *appsSuite) TestLogsFollow(c *check.C) {
 	s.expectLogsAccess()
 
 	s.jctlRCs = []io.ReadCloser{
-		ioutil.NopCloser(strings.NewReader("")),
-		ioutil.NopCloser(strings.NewReader("")),
-		ioutil.NopCloser(strings.NewReader("")),
+		io.NopCloser(strings.NewReader("")),
+		io.NopCloser(strings.NewReader("")),
+		io.NopCloser(strings.NewReader("")),
 	}
 
 	reqT, err := http.NewRequest("GET", "/v2/logs?follow=true", nil)
