@@ -24,11 +24,13 @@ import (
 	"encoding/json"
 	"errors"
 	"fmt"
+	"mime/multipart"
 	"net/http"
 	"net/http/httptest"
 	"os"
 	"path"
 	"path/filepath"
+	"strconv"
 	"strings"
 	"time"
 
@@ -45,9 +47,12 @@ import (
 	"github.com/snapcore/snapd/gadget"
 	"github.com/snapcore/snapd/gadget/quantity"
 	"github.com/snapcore/snapd/overlord/assertstate/assertstatetest"
+	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/devicestate"
 	"github.com/snapcore/snapd/overlord/hookstate"
+	"github.com/snapcore/snapd/overlord/install"
 	"github.com/snapcore/snapd/overlord/restart"
+	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/release"
 	"github.com/snapcore/snapd/secboot"
@@ -202,6 +207,12 @@ func (s *systemsSuite) TestSystemsGetSome(c *check.C) {
 		"revision": 2, "timestamp": "2009-11-10T23:00:00Z",
 		"seed-time": "2009-11-10T23:00:00Z",
 	}})
+	st.Set("default-recovery-system", devicestate.DefaultRecoverySystem{
+		System:   "20200318",
+		Model:    "my-model-2",
+		BrandID:  "my-brand",
+		Revision: 2,
+	})
 	st.Unlock()
 
 	s.expectAuthenticatedAccess()
@@ -234,10 +245,13 @@ func (s *systemsSuite) TestSystemsGetSome(c *check.C) {
 				},
 				Actions: []client.SystemAction{
 					{Title: "Install", Mode: "install"},
+					{Title: "Recover", Mode: "recover"},
+					{Title: "Factory reset", Mode: "factory-reset"},
 				},
 			}, {
-				Current: true,
-				Label:   "20200318",
+				Current:               true,
+				DefaultRecoverySystem: true,
+				Label:                 "20200318",
 				Model: client.SystemModelData{
 					Model:       "my-model-2",
 					BrandID:     "my-brand",
@@ -836,7 +850,7 @@ func (s *systemsSuite) TestSystemsGetSystemDetailsForLabel(c *check.C) {
 			client.StorageEncryptionSupportAvailable, "encrypted", "",
 		},
 	} {
-		mockEncryptionSupportInfo := &devicestate.EncryptionSupportInfo{
+		mockEncryptionSupportInfo := &install.EncryptionSupportInfo{
 			Available:          tc.available,
 			Disabled:           tc.disabled,
 			StorageSafety:      tc.storageSafety,
@@ -844,7 +858,7 @@ func (s *systemsSuite) TestSystemsGetSystemDetailsForLabel(c *check.C) {
 			UnavailableWarning: tc.unavailableWarning,
 		}
 
-		r := daemon.MockDeviceManagerSystemAndGadgetAndEncryptionInfo(func(mgr *devicestate.DeviceManager, label string) (*devicestate.System, *gadget.Info, *devicestate.EncryptionSupportInfo, error) {
+		r := daemon.MockDeviceManagerSystemAndGadgetAndEncryptionInfo(func(mgr *devicestate.DeviceManager, label string) (*devicestate.System, *gadget.Info, *install.EncryptionSupportInfo, error) {
 			c.Check(label, check.Equals, "20191119")
 			sys := &devicestate.System{
 				Model: s.seedModelForLabel20191119,
@@ -884,7 +898,7 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelError(c *check.C) {
 	s.daemon(c)
 	s.expectRootAccess()
 
-	r := daemon.MockDeviceManagerSystemAndGadgetAndEncryptionInfo(func(mgr *devicestate.DeviceManager, label string) (*devicestate.System, *gadget.Info, *devicestate.EncryptionSupportInfo, error) {
+	r := daemon.MockDeviceManagerSystemAndGadgetAndEncryptionInfo(func(mgr *devicestate.DeviceManager, label string) (*devicestate.System, *gadget.Info, *install.EncryptionSupportInfo, error) {
 		return nil, nil, nil, fmt.Errorf("boom")
 	})
 	defer r()
@@ -922,7 +936,7 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelIntegration(c *check.C) {
 	restore = s.mockSystemSeeds(c)
 	defer restore()
 
-	r := daemon.MockDeviceManagerSystemAndGadgetAndEncryptionInfo(func(mgr *devicestate.DeviceManager, label string) (*devicestate.System, *gadget.Info, *devicestate.EncryptionSupportInfo, error) {
+	r := daemon.MockDeviceManagerSystemAndGadgetAndEncryptionInfo(func(mgr *devicestate.DeviceManager, label string) (*devicestate.System, *gadget.Info, *install.EncryptionSupportInfo, error) {
 		// mockSystemSeed will ensure everything here is coming from
 		// the mocked seed except the encryptionInfo
 		sys, gadgetInfo, encInfo, err := deviceMgr.SystemAndGadgetAndEncryptionInfo(label)
@@ -942,11 +956,13 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelIntegration(c *check.C) {
 	c.Assert(rsp.Status, check.Equals, 200)
 	sys := rsp.Result.(client.SystemDetails)
 
-	c.Assert(sys, check.DeepEquals, client.SystemDetails{
+	sd := client.SystemDetails{
 		Label: "20191119",
 		Model: s.seedModelForLabel20191119.Headers(),
 		Actions: []client.SystemAction{
 			{Title: "Install", Mode: "install"},
+			{Title: "Recover", Mode: "recover"},
+			{Title: "Factory reset", Mode: "factory-reset"},
 		},
 
 		Brand: snap.StoreAccount{
@@ -972,17 +988,20 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelIntegration(c *check.C) {
 						Type:       "mbr",
 						Role:       "mbr",
 						Offset:     asOffsetPtr(0),
+						MinSize:    440,
 						Size:       440,
 						Content: []gadget.VolumeContent{
 							{
 								Image: "pc-boot.img",
 							},
 						},
+						YamlIndex: 0,
 					},
 					{
 						Name:       "BIOS Boot",
 						VolumeName: "pc",
 						Type:       "DA,21686148-6449-6E6F-744E-656564454649",
+						MinSize:    1 * quantity.SizeMiB,
 						Size:       1 * quantity.SizeMiB,
 						Offset:     asOffsetPtr(1 * quantity.OffsetMiB),
 						OffsetWrite: &gadget.RelativeOffset{
@@ -994,6 +1013,7 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelIntegration(c *check.C) {
 								Image: "pc-core.img",
 							},
 						},
+						YamlIndex: 1,
 					},
 					{
 						Name:       "ubuntu-seed",
@@ -1002,6 +1022,7 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelIntegration(c *check.C) {
 						VolumeName: "pc",
 						Type:       "EF,C12A7328-F81F-11D2-BA4B-00A0C93EC93B",
 						Offset:     asOffsetPtr(2 * quantity.OffsetMiB),
+						MinSize:    1200 * quantity.SizeMiB,
 						Size:       1200 * quantity.SizeMiB,
 						Filesystem: "vfat",
 						Content: []gadget.VolumeContent{
@@ -1014,6 +1035,7 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelIntegration(c *check.C) {
 								Target:           "EFI/boot/bootx64.efi",
 							},
 						},
+						YamlIndex: 2,
 					},
 					{
 						Name:       "ubuntu-boot",
@@ -1022,8 +1044,10 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelIntegration(c *check.C) {
 						VolumeName: "pc",
 						Type:       "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
 						Offset:     asOffsetPtr(1202 * quantity.OffsetMiB),
+						MinSize:    750 * quantity.SizeMiB,
 						Size:       750 * quantity.SizeMiB,
 						Filesystem: "ext4",
+						YamlIndex:  3,
 					},
 					{
 						Name:       "ubuntu-save",
@@ -1032,8 +1056,10 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelIntegration(c *check.C) {
 						VolumeName: "pc",
 						Type:       "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
 						Offset:     asOffsetPtr(1952 * quantity.OffsetMiB),
+						MinSize:    16 * quantity.SizeMiB,
 						Size:       16 * quantity.SizeMiB,
 						Filesystem: "ext4",
+						YamlIndex:  4,
 					},
 					{
 						Name:       "ubuntu-data",
@@ -1042,13 +1068,17 @@ func (s *systemsSuite) TestSystemsGetSpecificLabelIntegration(c *check.C) {
 						VolumeName: "pc",
 						Type:       "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
 						Offset:     asOffsetPtr(1968 * quantity.OffsetMiB),
+						MinSize:    1 * quantity.SizeGiB,
 						Size:       1 * quantity.SizeGiB,
 						Filesystem: "ext4",
+						YamlIndex:  5,
 					},
 				},
 			},
 		},
-	})
+	}
+	gadget.SetEnclosingVolumeInStructs(sd.Volumes)
+	c.Assert(sys, check.DeepEquals, sd)
 }
 
 func (s *systemsSuite) TestSystemInstallActionSetupStorageEncryptionCallsDevicestate(c *check.C) {
@@ -1202,3 +1232,693 @@ func (s *systemsSuite) TestSystemInstallActionError(c *check.C) {
 	rspe := s.errorReq(c, req, nil)
 	c.Check(rspe.Error(), check.Equals, `unsupported install step "unknown-install-step" (api)`)
 }
+
+var _ = check.Suite(&systemsCreateSuite{})
+
+type systemsCreateSuite struct {
+	apiBaseSuite
+
+	storeSigning              *assertstest.StoreStack
+	dev1Signing               *assertstest.SigningDB
+	dev1acct                  *asserts.Account
+	acct1Key                  *asserts.AccountKey
+	mockSeqFormingAssertionFn func(assertType *asserts.AssertionType, sequenceKey []string, sequence int, user *auth.UserState) (asserts.Assertion, error)
+	mockAssertionFn           func(at *asserts.AssertionType, headers []string, user *auth.UserState) (asserts.Assertion, error)
+}
+
+func (s *systemsCreateSuite) mockDevAssertion(c *check.C, t *asserts.AssertionType, extras map[string]interface{}) asserts.Assertion {
+	headers := map[string]interface{}{
+		"type":         t.Name,
+		"authority-id": s.dev1acct.AccountID(),
+		"account-id":   s.dev1acct.AccountID(),
+		"series":       "16",
+		"revision":     "5",
+		"timestamp":    "2030-11-06T09:16:26Z",
+	}
+
+	for k, v := range extras {
+		headers[k] = v
+	}
+
+	vs, err := s.dev1Signing.Sign(t, headers, nil, "")
+	c.Assert(err, check.IsNil)
+	return vs
+}
+
+func (s *systemsCreateSuite) mockStoreAssertion(c *check.C, t *asserts.AssertionType, extras map[string]interface{}) asserts.Assertion {
+	headers := map[string]interface{}{
+		"type":         t.Name,
+		"authority-id": s.storeSigning.AuthorityID,
+		"account-id":   s.dev1acct.AccountID(),
+		"series":       "16",
+		"revision":     "5",
+		"timestamp":    "2030-11-06T09:16:26Z",
+	}
+
+	for k, v := range extras {
+		headers[k] = v
+	}
+
+	vs, err := s.storeSigning.Sign(t, headers, nil, "")
+	c.Assert(err, check.IsNil)
+	return vs
+}
+
+func (s *systemsCreateSuite) SetUpTest(c *check.C) {
+	s.apiBaseSuite.SetUpTest(c)
+	d := s.daemon(c)
+
+	s.expectRootAccess()
+
+	restore := asserts.MockMaxSupportedFormat(asserts.ValidationSetType, 1)
+	s.AddCleanup(restore)
+
+	s.mockSeqFormingAssertionFn = nil
+	s.mockAssertionFn = nil
+
+	s.storeSigning = assertstest.NewStoreStack("can0nical", nil)
+
+	st := d.Overlord().State()
+	st.Lock()
+	snapstate.ReplaceStore(st, s)
+	assertstatetest.AddMany(st, s.storeSigning.StoreAccountKey(""))
+	st.Unlock()
+
+	s.dev1acct = assertstest.NewAccount(s.storeSigning, "developer1", nil, "")
+	c.Assert(s.storeSigning.Add(s.dev1acct), check.IsNil)
+
+	dev1PrivKey, _ := assertstest.GenerateKey(752)
+	s.acct1Key = assertstest.NewAccountKey(s.storeSigning, s.dev1acct, nil, dev1PrivKey.PublicKey(), "")
+
+	s.dev1Signing = assertstest.NewSigningDB(s.dev1acct.AccountID(), dev1PrivKey)
+	c.Assert(s.storeSigning.Add(s.acct1Key), check.IsNil)
+
+	d.Overlord().Loop()
+	s.AddCleanup(func() { d.Overlord().Stop() })
+}
+
+func (s *systemsCreateSuite) Assertion(at *asserts.AssertionType, headers []string, user *auth.UserState) (asserts.Assertion, error) {
+	s.pokeStateLock()
+	return s.mockAssertionFn(at, headers, user)
+}
+
+func (s *systemsCreateSuite) SeqFormingAssertion(assertType *asserts.AssertionType, sequenceKey []string, sequence int, user *auth.UserState) (asserts.Assertion, error) {
+	s.pokeStateLock()
+	return s.mockSeqFormingAssertionFn(assertType, sequenceKey, sequence, user)
+}
+
+func (s *systemsCreateSuite) TestCreateSystemActionBadRequests(c *check.C) {
+	type test struct {
+		body       map[string]interface{}
+		routeLabel string
+		result     string
+	}
+
+	tests := []test{
+		{
+			body: map[string]interface{}{
+				"action": "create",
+			},
+			routeLabel: "label",
+			result:     `label should not be provided in route when creating a system \(api\)`,
+		},
+		{
+			body: map[string]interface{}{
+				"action": "create",
+				"label":  "",
+			},
+			result: `label must be provided in request body for action "create" \(api\)`,
+		},
+		{
+			body: map[string]interface{}{
+				"action": "create",
+				"label":  "label",
+				"validation-sets": []string{
+					"not-a-validation-set",
+				},
+			},
+			result: `cannot parse validation sets: cannot parse validation set "not-a-validation-set": expected a single account/name \(api\)`,
+		},
+		{
+			body: map[string]interface{}{
+				"action": "create",
+				"label":  "label",
+				"validation-sets": []string{
+					"account/name",
+				},
+			},
+			result: `cannot fetch validation sets: validation-set assertion not found \(api\)`,
+		},
+	}
+
+	s.mockSeqFormingAssertionFn = func(assertType *asserts.AssertionType, sequenceKey []string, sequence int, user *auth.UserState) (asserts.Assertion, error) {
+		return nil, &asserts.NotFoundError{
+			Type: assertType,
+		}
+	}
+
+	for _, tc := range tests {
+		b, err := json.Marshal(tc.body)
+		c.Assert(err, check.IsNil)
+
+		url := "/v2/systems"
+		if tc.routeLabel != "" {
+			url += "/" + tc.routeLabel
+		}
+
+		req, err := http.NewRequest("POST", url, bytes.NewBuffer(b))
+		c.Assert(err, check.IsNil)
+
+		rspe := s.errorReq(c, req, nil)
+		c.Check(rspe.Status, check.Equals, 400)
+		c.Check(rspe, check.ErrorMatches, tc.result, check.Commentf("%+v", tc))
+	}
+}
+
+func (s *systemsCreateSuite) TestCreateSystemActionValidationSet(c *check.C) {
+	const valSetSequence = 0
+	s.testCreateSystemAction(c, valSetSequence)
+}
+
+func (s *systemsCreateSuite) TestCreateSystemActionSpecificValdationSet(c *check.C) {
+	const valSetSequence = 1
+	s.testCreateSystemAction(c, valSetSequence)
+}
+
+func (s *systemsCreateSuite) testCreateSystemAction(c *check.C, requestedValSetSequence int) {
+	snaps := []interface{}{
+		map[string]interface{}{
+			"name":     "pc-kernel",
+			"id":       snaptest.AssertedSnapID("pc-kernel"),
+			"revision": "10",
+			"presence": "required",
+		},
+		map[string]interface{}{
+			"name":     "pc",
+			"id":       snaptest.AssertedSnapID("pc"),
+			"revision": "10",
+			"presence": "required",
+		},
+		map[string]interface{}{
+			"name":     "core20",
+			"id":       snaptest.AssertedSnapID("core20"),
+			"revision": "10",
+			"presence": "required",
+		},
+	}
+
+	accountID := s.dev1acct.AccountID()
+
+	const validationSet = "validation-set-1"
+
+	vsetAssert := s.mockDevAssertion(c, asserts.ValidationSetType, map[string]interface{}{
+		"name":     validationSet,
+		"sequence": "1",
+		"snaps":    snaps,
+	})
+
+	s.mockAssertionFn = func(at *asserts.AssertionType, key []string, user *auth.UserState) (asserts.Assertion, error) {
+		headers, err := asserts.HeadersFromPrimaryKey(at, key)
+		if err != nil {
+			return nil, err
+		}
+
+		return s.storeSigning.Find(at, headers)
+	}
+
+	s.mockSeqFormingAssertionFn = func(assertType *asserts.AssertionType, sequenceKey []string, sequence int, user *auth.UserState) (asserts.Assertion, error) {
+		if assertType != asserts.ValidationSetType {
+			return nil, &asserts.NotFoundError{
+				Type: assertType,
+			}
+		}
+
+		c.Check(sequence, check.Equals, requestedValSetSequence)
+
+		return vsetAssert, nil
+	}
+
+	const (
+		markDefault   = true
+		testSystem    = true
+		expectedLabel = "1234"
+	)
+
+	daemon.MockDevicestateCreateRecoverySystem(func(st *state.State, label string, opts devicestate.CreateRecoverySystemOptions) (*state.Change, error) {
+		c.Check(expectedLabel, check.Equals, label)
+		c.Check(markDefault, check.Equals, opts.MarkDefault)
+		c.Check(testSystem, check.Equals, opts.TestSystem)
+
+		c.Check(opts.ValidationSets, check.HasLen, 1)
+
+		for _, vs := range opts.ValidationSets {
+			c.Check(vs.AccountID(), check.Equals, accountID)
+		}
+
+		return st.NewChange("change", "..."), nil
+	})
+
+	valSetString := accountID + "/" + validationSet
+	if requestedValSetSequence > 0 {
+		valSetString += "=" + strconv.Itoa(requestedValSetSequence)
+	}
+
+	body := map[string]interface{}{
+		"action":          "create",
+		"label":           expectedLabel,
+		"validation-sets": []string{valSetString},
+		"mark-default":    markDefault,
+		"test-system":     testSystem,
+	}
+
+	b, err := json.Marshal(body)
+	c.Assert(err, check.IsNil)
+
+	req, err := http.NewRequest("POST", "/v2/systems", bytes.NewBuffer(b))
+	c.Assert(err, check.IsNil)
+
+	res := s.asyncReq(c, req, nil)
+
+	st := s.d.Overlord().State()
+	st.Lock()
+	defer st.Unlock()
+
+	c.Check(st.Change(res.Change), check.NotNil)
+}
+
+func createFormData(c *check.C, fields map[string][]string, snaps map[string]string) (bytes.Buffer, string) {
+	var b bytes.Buffer
+	w := multipart.NewWriter(&b)
+
+	for k, vs := range fields {
+		for _, v := range vs {
+			err := w.WriteField(k, v)
+			c.Assert(err, check.IsNil)
+		}
+	}
+
+	for name, content := range snaps {
+		part, err := w.CreateFormFile("snap", name)
+		c.Assert(err, check.IsNil)
+
+		_, err = part.Write([]byte(content))
+		c.Assert(err, check.IsNil)
+	}
+
+	err := w.Close()
+	c.Assert(err, check.IsNil)
+
+	return b, w.Boundary()
+}
+
+func (s *systemsCreateSuite) TestRemoveSystemAction(c *check.C) {
+	const expectedLabel = "1234"
+
+	daemon.MockDevicestateRemoveRecoverySystem(func(st *state.State, label string) (*state.Change, error) {
+		c.Check(expectedLabel, check.Equals, label)
+
+		return st.NewChange("change", "..."), nil
+	})
+
+	body := map[string]interface{}{
+		"action": "remove",
+	}
+
+	b, err := json.Marshal(body)
+	c.Assert(err, check.IsNil)
+
+	req, err := http.NewRequest("POST", "/v2/systems/"+expectedLabel, bytes.NewBuffer(b))
+	c.Assert(err, check.IsNil)
+
+	res := s.asyncReq(c, req, nil)
+
+	st := s.d.Overlord().State()
+	st.Lock()
+	defer st.Unlock()
+
+	c.Check(st.Change(res.Change), check.NotNil)
+}
+
+func (s *systemsCreateSuite) TestRemoveSystemActionNotFound(c *check.C) {
+	const expectedLabel = "1234"
+
+	daemon.MockDevicestateRemoveRecoverySystem(func(st *state.State, label string) (*state.Change, error) {
+		c.Check(expectedLabel, check.Equals, label)
+		return nil, devicestate.ErrNoRecoverySystem
+	})
+
+	body := map[string]interface{}{
+		"action": "remove",
+	}
+
+	b, err := json.Marshal(body)
+	c.Assert(err, check.IsNil)
+
+	req, err := http.NewRequest("POST", "/v2/systems/"+expectedLabel, bytes.NewBuffer(b))
+	c.Assert(err, check.IsNil)
+
+	res := s.errorReq(c, req, nil)
+	c.Check(res.Status, check.Equals, 404)
+	c.Check(res.Message, check.Equals, "recovery system does not exist")
+}
+
+func (s *systemsCreateSuite) TestCreateSystemActionOfflineBadRequests(c *check.C) {
+	type test struct {
+		fields map[string][]string
+		result string
+	}
+
+	tests := []test{
+		{
+			fields: map[string][]string{
+				"action": {"create"},
+				"label":  {"1", "2"},
+			},
+			result: `expected exactly one "label" value in form \(api\)`,
+		},
+		{
+			fields: map[string][]string{
+				"action":      {"create"},
+				"label":       {"1"},
+				"test-system": {"false", "true"},
+			},
+			result: `expected at most one "test-system" value in form \(api\)`,
+		},
+		{
+			fields: map[string][]string{
+				"action":       {"create"},
+				"label":        {"1"},
+				"mark-default": {"false", "true"},
+			},
+			result: `expected at most one "mark-default" value in form \(api\)`,
+		},
+		{
+			fields: map[string][]string{
+				"action":          {"create"},
+				"label":           {"1"},
+				"validation-sets": {"id/set-1", "id/set-2"},
+			},
+			result: `expected at most one "validation-sets" value in form \(api\)`,
+		},
+		{
+			fields: map[string][]string{
+				"action":      {"create"},
+				"label":       {"1"},
+				"test-system": {"not-valid"},
+			},
+			result: `cannot parse "test-system" value as boolean: not-valid \(api\)`,
+		},
+		{
+			fields: map[string][]string{
+				"action":       {"create"},
+				"label":        {"1"},
+				"mark-default": {"not-valid"},
+			},
+			result: `cannot parse "mark-default" value as boolean: not-valid \(api\)`,
+		},
+		{
+			fields: map[string][]string{
+				"action":          {"create"},
+				"label":           {"1"},
+				"validation-sets": {"invalid-set-name"},
+			},
+			result: `cannot parse validation sets: cannot parse validation set "invalid-set-name": expected a single account/name \(api\)`,
+		},
+	}
+
+	snaps := map[string]string{
+		"snap-1": "snap-1 contents",
+		"snap-2": "snap-2 contents",
+	}
+
+	for _, tc := range tests {
+		form, boundary := createFormData(c, tc.fields, snaps)
+
+		req, err := http.NewRequest("POST", "/v2/systems", &form)
+		c.Assert(err, check.IsNil)
+		req.Header.Set("Content-Type", "multipart/form-data; boundary="+boundary)
+		req.Header.Set("Content-Length", strconv.Itoa(form.Len()))
+
+		rspe := s.errorReq(c, req, nil)
+		c.Check(rspe.Status, check.Equals, 400)
+		c.Check(rspe, check.ErrorMatches, tc.result, check.Commentf("%+v", tc))
+
+		// make sure that form files we uploaded get removed on failure
+		files, err := filepath.Glob(filepath.Join(dirs.SnapBlobDir, dirs.LocalInstallBlobTempPrefix+"*"))
+		c.Assert(err, check.IsNil)
+		c.Check(files, check.HasLen, 0)
+	}
+}
+
+func (s *systemsCreateSuite) TestCreateSystemActionOffline(c *check.C) {
+	snaps := []interface{}{
+		map[string]interface{}{
+			"name":     "pc-kernel",
+			"id":       snaptest.AssertedSnapID("pc-kernel"),
+			"revision": "10",
+			"presence": "required",
+		},
+		map[string]interface{}{
+			"name":     "pc",
+			"id":       snaptest.AssertedSnapID("pc"),
+			"revision": "10",
+			"presence": "required",
+		},
+		map[string]interface{}{
+			"name":     "core20",
+			"id":       snaptest.AssertedSnapID("core20"),
+			"revision": "10",
+			"presence": "required",
+		},
+	}
+
+	accountID := s.dev1acct.AccountID()
+
+	const (
+		validationSet = "validation-set-1"
+		expectedLabel = "1234"
+	)
+
+	vsetAssert := s.mockDevAssertion(c, asserts.ValidationSetType, map[string]interface{}{
+		"name":     validationSet,
+		"sequence": "1",
+		"snaps":    snaps,
+	})
+
+	assertions := []string{
+		string(asserts.Encode(vsetAssert)),
+		string(asserts.Encode(s.acct1Key)),
+		string(asserts.Encode(s.dev1acct)),
+	}
+
+	snapFormData := make(map[string]string)
+	for _, name := range []string{"pc-kernel", "pc", "core20"} {
+		f := snaptest.MakeTestSnapWithFiles(c, fmt.Sprintf("name: %s\nversion: 1", name), nil)
+		digest, size, err := asserts.SnapFileSHA3_384(f)
+		c.Assert(err, check.IsNil)
+
+		rev := s.mockStoreAssertion(c, asserts.SnapRevisionType, map[string]interface{}{
+			"snap-id":       snaptest.AssertedSnapID(name),
+			"snap-sha3-384": digest,
+			"developer-id":  s.dev1acct.AccountID(),
+			"snap-size":     strconv.Itoa(int(size)),
+			"snap-revision": "10",
+		})
+
+		// this is required right now. should it be?
+		decl := s.mockStoreAssertion(c, asserts.SnapDeclarationType, map[string]interface{}{
+			"series":       "16",
+			"snap-id":      snaptest.AssertedSnapID(name),
+			"snap-name":    name,
+			"publisher-id": s.dev1acct.AccountID(),
+			"timestamp":    time.Now().Format(time.RFC3339),
+		})
+
+		assertions = append(assertions, string(asserts.Encode(rev)), string(asserts.Encode(decl)))
+
+		content, err := os.ReadFile(f)
+		c.Assert(err, check.IsNil)
+
+		snapFormData[name] = string(content)
+	}
+
+	valSetString := accountID + "/" + validationSet
+	fields := map[string][]string{
+		"action":          {"create"},
+		"assertion":       assertions,
+		"label":           {expectedLabel},
+		"validation-sets": {valSetString},
+	}
+
+	form, boundary := createFormData(c, fields, snapFormData)
+
+	daemon.MockDevicestateCreateRecoverySystem(func(st *state.State, label string, opts devicestate.CreateRecoverySystemOptions) (*state.Change, error) {
+		c.Check(expectedLabel, check.Equals, label)
+		c.Check(opts.ValidationSets, check.HasLen, 1)
+		c.Check(opts.ValidationSets[0].Body(), check.DeepEquals, vsetAssert.Body())
+
+		c.Check(opts.LocalSnaps, check.HasLen, 3)
+
+		for _, vs := range opts.ValidationSets {
+			c.Check(vs.AccountID(), check.Equals, accountID)
+		}
+
+		return st.NewChange("change", "..."), nil
+	})
+
+	req, err := http.NewRequest("POST", "/v2/systems", &form)
+	c.Assert(err, check.IsNil)
+	req.Header.Set("Content-Type", "multipart/form-data; boundary="+boundary)
+	req.Header.Set("Content-Length", strconv.Itoa(form.Len()))
+
+	res := s.asyncReq(c, req, nil)
+
+	st := s.d.Overlord().State()
+	st.Lock()
+	defer st.Unlock()
+
+	c.Check(st.Change(res.Change), check.NotNil)
+}
+
+func (s *systemsCreateSuite) TestCreateSystemActionOfflinePreinstalledJSON(c *check.C) {
+	const (
+		expectedLabel = "1234"
+	)
+
+	daemon.MockDevicestateCreateRecoverySystem(func(st *state.State, label string, opts devicestate.CreateRecoverySystemOptions) (*state.Change, error) {
+		c.Check(expectedLabel, check.Equals, label)
+		c.Check(opts.ValidationSets, check.HasLen, 0)
+		c.Check(opts.LocalSnaps, check.HasLen, 0)
+		c.Check(opts.Offline, check.Equals, true)
+
+		return st.NewChange("change", "..."), nil
+	})
+
+	body := map[string]interface{}{
+		"action":  "create",
+		"label":   expectedLabel,
+		"offline": true,
+	}
+
+	b, err := json.Marshal(body)
+	c.Assert(err, check.IsNil)
+
+	req, err := http.NewRequest("POST", "/v2/systems", bytes.NewBuffer(b))
+	c.Assert(err, check.IsNil)
+
+	res := s.asyncReq(c, req, nil)
+
+	st := s.d.Overlord().State()
+	st.Lock()
+	defer st.Unlock()
+
+	c.Check(st.Change(res.Change), check.NotNil)
+}
+
+func (s *systemsCreateSuite) TestCreateSystemActionOfflinePreinstalledForm(c *check.C) {
+	const (
+		expectedLabel = "1234"
+	)
+
+	fields := map[string][]string{
+		"action": {"create"},
+		"label":  {expectedLabel},
+	}
+
+	form, boundary := createFormData(c, fields, nil)
+
+	daemon.MockDevicestateCreateRecoverySystem(func(st *state.State, label string, opts devicestate.CreateRecoverySystemOptions) (*state.Change, error) {
+		c.Check(expectedLabel, check.Equals, label)
+		c.Check(opts.ValidationSets, check.HasLen, 0)
+		c.Check(opts.LocalSnaps, check.HasLen, 0)
+		c.Check(opts.Offline, check.Equals, true)
+
+		return st.NewChange("change", "..."), nil
+	})
+
+	req, err := http.NewRequest("POST", "/v2/systems", &form)
+	c.Assert(err, check.IsNil)
+	req.Header.Set("Content-Type", "multipart/form-data; boundary="+boundary)
+	req.Header.Set("Content-Length", strconv.Itoa(form.Len()))
+
+	res := s.asyncReq(c, req, nil)
+
+	st := s.d.Overlord().State()
+	st.Lock()
+	defer st.Unlock()
+
+	c.Check(st.Change(res.Change), check.NotNil)
+}
+
+func (s *systemsCreateSuite) TestCreateSystemActionOfflineJustValidationSets(c *check.C) {
+	accountID := s.dev1acct.AccountID()
+
+	const (
+		validationSet = "validation-set-1"
+		expectedLabel = "1234"
+	)
+
+	vsetAssert := s.mockDevAssertion(c, asserts.ValidationSetType, map[string]interface{}{
+		"name":     validationSet,
+		"sequence": "1",
+		"snaps": []interface{}{
+			map[string]interface{}{
+				"name":     "pc-kernel",
+				"id":       snaptest.AssertedSnapID("pc-kernel"),
+				"revision": "10",
+				"presence": "required",
+			},
+			map[string]interface{}{
+				"name":     "pc",
+				"id":       snaptest.AssertedSnapID("pc"),
+				"revision": "10",
+				"presence": "required",
+			},
+			map[string]interface{}{
+				"name":     "core20",
+				"id":       snaptest.AssertedSnapID("core20"),
+				"revision": "10",
+				"presence": "required",
+			},
+		},
+	})
+
+	assertions := []string{
+		string(asserts.Encode(vsetAssert)),
+		string(asserts.Encode(s.acct1Key)),
+		string(asserts.Encode(s.dev1acct)),
+	}
+
+	valSetString := accountID + "/" + validationSet
+	fields := map[string][]string{
+		"action":          {"create"},
+		"assertion":       assertions,
+		"label":           {expectedLabel},
+		"validation-sets": {valSetString},
+	}
+
+	form, boundary := createFormData(c, fields, nil)
+
+	daemon.MockDevicestateCreateRecoverySystem(func(st *state.State, label string, opts devicestate.CreateRecoverySystemOptions) (*state.Change, error) {
+		c.Check(expectedLabel, check.Equals, label)
+		c.Check(opts.ValidationSets, check.HasLen, 1)
+		c.Check(opts.ValidationSets[0].Body(), check.DeepEquals, vsetAssert.Body())
+		c.Check(opts.LocalSnaps, check.HasLen, 0)
+		c.Check(opts.Offline, check.Equals, true)
+
+		return st.NewChange("change", "..."), nil
+	})
+
+	req, err := http.NewRequest("POST", "/v2/systems", &form)
+	c.Assert(err, check.IsNil)
+	req.Header.Set("Content-Type", "multipart/form-data; boundary="+boundary)
+	req.Header.Set("Content-Length", strconv.Itoa(form.Len()))
+
+	res := s.asyncReq(c, req, nil)
+
+	st := s.d.Overlord().State()
+	st.Lock()
+	defer st.Unlock()
+
+	c.Check(st.Change(res.Change), check.NotNil)
+}
