@@ -21,7 +21,7 @@ package install_test
 
 import (
 	"fmt"
-	"io/ioutil"
+	"os"
 	"path/filepath"
 	"testing"
 	"time"
@@ -68,7 +68,8 @@ func (s *partitionTestSuite) SetUpTest(c *C) {
 }
 
 const (
-	scriptPartitionsBios = iota
+	scriptPartitionsNone = iota
+	scriptPartitionsBios
 	scriptPartitionsBiosSeed
 	scriptPartitionsBiosSeedData
 )
@@ -95,7 +96,7 @@ func makeMockDiskMappingIncludingPartitions(num int) *disks.MockDiskMapping {
 			PartitionLabel:   "BIOS Boot",
 			Major:            42,
 			Minor:            1,
-			StructureIndex:   1,
+			DiskIndex:        1,
 		})
 	}
 
@@ -109,7 +110,7 @@ func makeMockDiskMappingIncludingPartitions(num int) *disks.MockDiskMapping {
 			PartitionLabel:   "Recovery",
 			Major:            42,
 			Minor:            2,
-			StructureIndex:   2,
+			DiskIndex:        2,
 			FilesystemType:   "vfat",
 			FilesystemUUID:   "A644-B807",
 			FilesystemLabel:  "ubuntu-seed",
@@ -126,7 +127,7 @@ func makeMockDiskMappingIncludingPartitions(num int) *disks.MockDiskMapping {
 			PartitionLabel:   "Writable",
 			Major:            42,
 			Minor:            3,
-			StructureIndex:   3,
+			DiskIndex:        3,
 			FilesystemType:   "ext4",
 			FilesystemUUID:   "8781-433a",
 			FilesystemLabel:  "ubuntu-data",
@@ -137,59 +138,51 @@ func makeMockDiskMappingIncludingPartitions(num int) *disks.MockDiskMapping {
 }
 
 var mockOnDiskStructureWritable = gadget.OnDiskStructure{
-	Node: "/dev/node3",
-	LaidOutStructure: gadget.LaidOutStructure{
-		VolumeStructure: &gadget.VolumeStructure{
-			VolumeName: "pc",
-			Name:       "Writable",
-			Size:       1258291200,
-			Type:       "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
-			Role:       "system-data",
-			Label:      "ubuntu-data",
-			Filesystem: "ext4",
-		},
-		StartOffset: 1260388352,
-		Index:       3,
-	},
+	Node:             "/dev/node3",
+	Name:             "Writable",
+	Type:             "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
+	PartitionFSLabel: "ubuntu-data",
+	PartitionFSType:  "ext4",
+	StartOffset:      1260388352,
+	// Note the DiskIndex appears to be the same as the YamlIndex, but this is
+	// because YamlIndex starts at 0 and DiskIndex starts at 1, and there is a
+	// yaml structure (the MBR) that does not appear on disk
+	DiskIndex: 3,
 	// expanded to fill the disk
 	Size: 2*quantity.SizeGiB + 845*quantity.SizeMiB + 1031680,
 }
 
-var mockOnDiskStructureSave = gadget.OnDiskStructure{
-	Node: "/dev/node3",
-	LaidOutStructure: gadget.LaidOutStructure{
-		VolumeStructure: &gadget.VolumeStructure{
-			VolumeName: "pc",
-			Name:       "Save",
-			Size:       128 * quantity.SizeMiB,
-			Type:       "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
-			Role:       "system-save",
-			Label:      "ubuntu-save",
-			Filesystem: "ext4",
-		},
-		StartOffset: 1260388352,
-		Index:       3,
-	},
-	Size: 128 * quantity.SizeMiB,
+func createOnDiskStructureSave(enclosing *gadget.Volume) *gadget.OnDiskStructure {
+	return &gadget.OnDiskStructure{
+		Node:             "/dev/node3",
+		Name:             "Save",
+		Size:             128 * quantity.SizeMiB,
+		Type:             "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
+		PartitionFSLabel: "ubuntu-save",
+		PartitionFSType:  "ext4",
+		StartOffset:      1260388352,
+		// Note the DiskIndex appears to be the same as the YamlIndex, but this is
+		// because YamlIndex starts at 0 and DiskIndex starts at 1, and there is a
+		// yaml structure (the MBR) that does not appear on disk
+		DiskIndex: 3,
+	}
 }
 
-var mockOnDiskStructureWritableAfterSave = gadget.OnDiskStructure{
-	Node: "/dev/node4",
-	LaidOutStructure: gadget.LaidOutStructure{
-		VolumeStructure: &gadget.VolumeStructure{
-			VolumeName: "pc",
-			Name:       "Writable",
-			Size:       1200 * quantity.SizeMiB,
-			Type:       "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
-			Role:       "system-data",
-			Label:      "ubuntu-data",
-			Filesystem: "ext4",
-		},
-		StartOffset: 1394606080,
-		Index:       4,
-	},
-	// expanded to fill the disk
-	Size: 2*quantity.SizeGiB + 717*quantity.SizeMiB + 1031680,
+func createOnDiskStructureWritableAfterSave(enclosing *gadget.Volume) *gadget.OnDiskStructure {
+	return &gadget.OnDiskStructure{
+		Node: "/dev/node4",
+		Name: "Writable",
+		// expanded to fill the disk
+		Size:             2*quantity.SizeGiB + 717*quantity.SizeMiB + 1031680,
+		Type:             "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
+		PartitionFSLabel: "ubuntu-data",
+		PartitionFSType:  "ext4",
+		StartOffset:      1394606080,
+		// Note the DiskIndex appears to be the same as the YamlIndex, but this is
+		// because YamlIndex starts at 0 and DiskIndex starts at 1, and there is a
+		// yaml structure (the MBR) that does not appear on disk
+		DiskIndex: 4,
+	}
 }
 
 type uc20Model struct{}
@@ -207,7 +200,7 @@ func (s *partitionTestSuite) TestBuildPartitionList(c *C) {
 	restore := disks.MockDeviceNameToDiskMapping(m)
 	defer restore()
 
-	err := makeMockGadget(s.gadgetRoot, gptGadgetContentWithSave)
+	err := gadgettest.MakeMockGadget(s.gadgetRoot, gptGadgetContentWithSave)
 	c.Assert(err, IsNil)
 	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
 	c.Assert(err, IsNil)
@@ -217,14 +210,63 @@ func (s *partitionTestSuite) TestBuildPartitionList(c *C) {
 
 	// the expected expanded writable partition size is:
 	// start offset = (2M + 1200M), expanded size in sectors = (8388575*512 - start offset)/512
-	sfdiskInput, create, err := install.BuildPartitionList(dl, pv)
+	sfdiskInput, create, err := install.BuildPartitionList(dl, pv.Volume, nil)
 	c.Assert(err, IsNil)
 	c.Assert(sfdiskInput.String(), Equals,
 		`/dev/node3 : start=     2461696, size=      262144, type=0FC63DAF-8483-4772-8E79-3D69D8477DE4, name="Save"
 /dev/node4 : start=     2723840, size=     5664735, type=0FC63DAF-8483-4772-8E79-3D69D8477DE4, name="Writable"
 `)
 	c.Check(create, NotNil)
-	c.Assert(create, DeepEquals, []gadget.OnDiskStructure{mockOnDiskStructureSave, mockOnDiskStructureWritableAfterSave})
+	c.Assert(create, DeepEquals, []*gadget.OnDiskAndGadgetStructurePair{
+		{
+			DiskStructure:   createOnDiskStructureSave(pv.Volume),
+			GadgetStructure: &pv.Volume.Structure[3],
+		},
+		{
+			DiskStructure:   createOnDiskStructureWritableAfterSave(pv.Volume),
+			GadgetStructure: &pv.Volume.Structure[4],
+		},
+	})
+}
+
+func (s *partitionTestSuite) TestBuildPartitionListPartsNotInGadget(c *C) {
+	m := map[string]*disks.MockDiskMapping{
+		"/dev/node": makeMockDiskMappingIncludingPartitions(scriptPartitionsBiosSeed),
+	}
+
+	restore := disks.MockDeviceNameToDiskMapping(m)
+	defer restore()
+
+	// This gadget does not specify the bios partition, but it is on the disk
+	err := gadgettest.MakeMockGadget(s.gadgetRoot, gptGadgetContentWithGap)
+	c.Assert(err, IsNil)
+	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
+	c.Assert(err, IsNil)
+
+	dl, err := gadget.OnDiskVolumeFromDevice("/dev/node")
+	c.Assert(err, IsNil)
+
+	// the expected expanded writable partition size is: start
+	// offset = (2M + 1200M), expanded size in sectors =
+	// (8388575*512 - start offset)/512
+	sfdiskInput, create, err := install.BuildPartitionList(dl, pv.Volume,
+		&install.CreateOptions{})
+	c.Assert(err, IsNil)
+	c.Assert(sfdiskInput.String(), Equals,
+		`/dev/node3 : start=     2461696, size=      262144, type=0FC63DAF-8483-4772-8E79-3D69D8477DE4, name="Save"
+/dev/node4 : start=     2723840, size=     5664735, type=0FC63DAF-8483-4772-8E79-3D69D8477DE4, name="Writable"
+`)
+	c.Check(create, NotNil)
+	c.Assert(create, DeepEquals, []*gadget.OnDiskAndGadgetStructurePair{
+		{
+			DiskStructure:   createOnDiskStructureSave(pv.Volume),
+			GadgetStructure: &pv.Volume.Structure[1],
+		},
+		{
+			DiskStructure:   createOnDiskStructureWritableAfterSave(pv.Volume),
+			GadgetStructure: &pv.Volume.Structure[2],
+		},
+	})
 }
 
 func (s *partitionTestSuite) TestBuildPartitionListOnlyCreatablePartitions(c *C) {
@@ -233,7 +275,7 @@ func (s *partitionTestSuite) TestBuildPartitionListOnlyCreatablePartitions(c *C)
 	// partition just vanished from the disk)
 	mockDisk := makeMockDiskMappingIncludingPartitions(scriptPartitionsBiosSeed)
 	mockDisk.Structure = mockDisk.Structure[1:]
-	mockDisk.Structure[0].StructureIndex = 1
+	mockDisk.Structure[0].DiskIndex = 1
 	m := map[string]*disks.MockDiskMapping{
 		"/dev/node": mockDisk,
 	}
@@ -241,7 +283,7 @@ func (s *partitionTestSuite) TestBuildPartitionListOnlyCreatablePartitions(c *C)
 	restore := disks.MockDeviceNameToDiskMapping(m)
 	defer restore()
 
-	err := makeMockGadget(s.gadgetRoot, gptGadgetContentWithSave)
+	err := gadgettest.MakeMockGadget(s.gadgetRoot, gptGadgetContentWithSave)
 	c.Assert(err, IsNil)
 	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
 	c.Assert(err, IsNil)
@@ -249,8 +291,49 @@ func (s *partitionTestSuite) TestBuildPartitionListOnlyCreatablePartitions(c *C)
 	dl, err := gadget.OnDiskVolumeFromDevice("/dev/node")
 	c.Assert(err, IsNil)
 
-	_, _, err = install.BuildPartitionList(dl, pv)
-	c.Assert(err, ErrorMatches, `cannot create partition #1 \(\"BIOS Boot\"\)`)
+	_, _, err = install.BuildPartitionList(dl, pv.Volume, nil)
+	c.Assert(err, ErrorMatches, `gadget and boot device /dev/node partition table not compatible: cannot find gadget structure "BIOS Boot" on disk`)
+}
+
+func (s *partitionTestSuite) TestBuildPartitionListExistingPartsInSizeRange(c *C) {
+	m := map[string]*disks.MockDiskMapping{
+		"/dev/node": makeMockDiskMappingIncludingPartitions(scriptPartitionsBiosSeed),
+	}
+
+	restore := disks.MockDeviceNameToDiskMapping(m)
+	defer restore()
+
+	// The gadget has size rage of [1000, 1400]MiB for the seed partition,
+	// and the actual size on disk is 1200MiB. The partition on disk should
+	// match the one in the gadget and we will created save and data
+	// partitions right after it.
+	err := gadgettest.MakeMockGadget(s.gadgetRoot, gptGadgetContentWithRangeForSeed)
+	c.Assert(err, IsNil)
+	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
+	c.Assert(err, IsNil)
+
+	dl, err := gadget.OnDiskVolumeFromDevice("/dev/node")
+	c.Assert(err, IsNil)
+
+	// the expected expanded writable partition size is:
+	// start offset = (2M + 1200M), expanded size in sectors = (8388575*512 - start offset)/512
+	sfdiskInput, create, err := install.BuildPartitionList(dl, pv.Volume, nil)
+	c.Assert(err, IsNil)
+	c.Assert(sfdiskInput.String(), Equals,
+		`/dev/node3 : start=     2461696, size=      262144, type=0FC63DAF-8483-4772-8E79-3D69D8477DE4, name="Save"
+/dev/node4 : start=     2723840, size=     5664735, type=0FC63DAF-8483-4772-8E79-3D69D8477DE4, name="Writable"
+`)
+	c.Check(create, NotNil)
+	c.Assert(create, DeepEquals, []*gadget.OnDiskAndGadgetStructurePair{
+		{
+			DiskStructure:   createOnDiskStructureSave(pv.Volume),
+			GadgetStructure: &pv.Volume.Structure[3],
+		},
+		{
+			DiskStructure:   createOnDiskStructureWritableAfterSave(pv.Volume),
+			GadgetStructure: &pv.Volume.Structure[4],
+		},
+	})
 }
 
 func (s *partitionTestSuite) TestCreatePartitions(c *C) {
@@ -264,25 +347,37 @@ func (s *partitionTestSuite) TestCreatePartitions(c *C) {
 	restore := disks.MockDeviceNameToDiskMapping(m)
 	defer restore()
 
+	cmdUdevadm := testutil.MockCommand(c, "udevadm", "")
+	defer cmdUdevadm.Restore()
+
 	calls := 0
-	restore = install.MockEnsureNodesExist(func(ds []gadget.OnDiskStructure, timeout time.Duration) error {
+	restore = install.MockEnsureNodesExist(func(nodes []string, timeout time.Duration) error {
 		calls++
-		c.Assert(ds, HasLen, 1)
-		c.Assert(ds[0].Node, Equals, "/dev/node3")
+		c.Assert(nodes, HasLen, 1)
+		c.Assert(nodes[0], Equals, "/dev/node3")
 		return nil
 	})
 	defer restore()
 
-	err := makeMockGadget(s.gadgetRoot, gadgetContent)
+	err := gadgettest.MakeMockGadget(s.gadgetRoot, gadgetContent)
 	c.Assert(err, IsNil)
 	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
 	c.Assert(err, IsNil)
 
 	dl, err := gadget.OnDiskVolumeFromDevice("/dev/node")
 	c.Assert(err, IsNil)
-	created, err := install.CreateMissingPartitions(dl, pv)
+	opts := &install.CreateOptions{
+		GadgetRootDir: s.gadgetRoot,
+	}
+	created, err := install.TestCreateMissingPartitions(dl, pv.Volume, opts)
 	c.Assert(err, IsNil)
-	c.Assert(created, DeepEquals, []gadget.OnDiskStructure{mockOnDiskStructureWritable})
+	c.Assert(created, DeepEquals, []*gadget.OnDiskAndGadgetStructurePair{
+		{
+			DiskStructure:   &mockOnDiskStructureWritable,
+			GadgetStructure: &pv.Volume.Structure[3],
+		},
+	})
+
 	c.Assert(calls, Equals, 1)
 
 	// Check partition table write
@@ -294,6 +389,51 @@ func (s *partitionTestSuite) TestCreatePartitions(c *C) {
 	c.Assert(s.cmdPartx.Calls(), DeepEquals, [][]string{
 		{"partx", "-u", "/dev/node"},
 	})
+
+	c.Assert(cmdUdevadm.Calls(), DeepEquals, [][]string{
+		{"udevadm", "settle", "--timeout=180"},
+	})
+}
+
+func (s *partitionTestSuite) TestCreatePartitionsNonRolePartitions(c *C) {
+	cmdSfdisk := testutil.MockCommand(c, "sfdisk", "")
+	defer cmdSfdisk.Restore()
+
+	m := map[string]*disks.MockDiskMapping{
+		"/dev/node": makeMockDiskMappingIncludingPartitions(scriptPartitionsNone),
+	}
+	restore := disks.MockDeviceNameToDiskMapping(m)
+	defer restore()
+
+	cmdUdevadm := testutil.MockCommand(c, "udevadm", "")
+	defer cmdUdevadm.Restore()
+
+	calls := 0
+	restore = install.MockEnsureNodesExist(func(nodes []string, timeout time.Duration) error {
+		calls++
+		c.Assert(nodes, HasLen, 3)
+		c.Assert(nodes[0], Equals, "/dev/node1")
+		c.Assert(nodes[1], Equals, "/dev/node2")
+		c.Assert(nodes[2], Equals, "/dev/node3")
+		return nil
+	})
+	defer restore()
+
+	err := gadgettest.MakeMockGadget(s.gadgetRoot, gadgetContent)
+	c.Assert(err, IsNil)
+	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
+	c.Assert(err, IsNil)
+
+	dl, err := gadget.OnDiskVolumeFromDevice("/dev/node")
+	c.Assert(err, IsNil)
+	opts := &install.CreateOptions{
+		GadgetRootDir:              s.gadgetRoot,
+		CreateAllMissingPartitions: true,
+	}
+	created, err := install.TestCreateMissingPartitions(dl, pv.Volume, opts)
+	c.Assert(err, IsNil)
+	c.Assert(created, HasLen, 3)
+	c.Assert(calls, Equals, 1)
 }
 
 func (s *partitionTestSuite) TestRemovePartitionsTrivial(c *C) {
@@ -305,15 +445,15 @@ func (s *partitionTestSuite) TestRemovePartitionsTrivial(c *C) {
 	restore := disks.MockDeviceNameToDiskMapping(m)
 	defer restore()
 
-	err := makeMockGadget(s.gadgetRoot, gadgetContent)
+	err := gadgettest.MakeMockGadget(s.gadgetRoot, gadgetContent)
 	c.Assert(err, IsNil)
-	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
+	gInfo, err := gadget.ReadInfoAndValidate(s.gadgetRoot, uc20Mod, nil)
 	c.Assert(err, IsNil)
 
 	dl, err := gadget.OnDiskVolumeFromDevice("/dev/node")
 	c.Assert(err, IsNil)
 
-	err = install.RemoveCreatedPartitions(pv, dl)
+	err = install.RemoveCreatedPartitions(s.gadgetRoot, gInfo.Volumes["pc"], dl)
 	c.Assert(err, IsNil)
 }
 
@@ -339,7 +479,7 @@ func (s *partitionTestSuite) TestRemovePartitions(c *C) {
 					PartitionLabel:   "BIOS Boot",
 					Major:            42,
 					Minor:            1,
-					StructureIndex:   1,
+					DiskIndex:        1,
 				},
 				{
 					KernelDeviceNode: "/dev/node2",
@@ -350,7 +490,7 @@ func (s *partitionTestSuite) TestRemovePartitions(c *C) {
 					PartitionLabel:   "Recovery",
 					Major:            42,
 					Minor:            2,
-					StructureIndex:   2,
+					DiskIndex:        2,
 					FilesystemType:   "vfat",
 					FilesystemUUID:   "A644-B807",
 					FilesystemLabel:  "ubuntu-seed",
@@ -364,7 +504,7 @@ func (s *partitionTestSuite) TestRemovePartitions(c *C) {
 					PartitionLabel:   "Writable",
 					Major:            42,
 					Minor:            3,
-					StructureIndex:   3,
+					DiskIndex:        3,
 					FilesystemType:   "ext4",
 					FilesystemUUID:   "8781-433a",
 					FilesystemLabel:  "ubuntu-data",
@@ -385,51 +525,165 @@ func (s *partitionTestSuite) TestRemovePartitions(c *C) {
 	dl, err := gadget.OnDiskVolumeFromDevice("/dev/node")
 	c.Assert(err, IsNil)
 
-	err = makeMockGadget(s.gadgetRoot, gadgetContent)
+	err = gadgettest.MakeMockGadget(s.gadgetRoot, gadgetContent)
 	c.Assert(err, IsNil)
-	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
+	gInfo, err := gadget.ReadInfoAndValidate(s.gadgetRoot, uc20Mod, nil)
 	c.Assert(err, IsNil)
 
-	err = install.RemoveCreatedPartitions(pv, dl)
+	err = install.RemoveCreatedPartitions(s.gadgetRoot, gInfo.Volumes["pc"], dl)
 	c.Assert(err, IsNil)
 
 	c.Assert(cmdSfdisk.Calls(), DeepEquals, [][]string{
 		{"sfdisk", "--no-reread", "--delete", "/dev/node", "3"},
 	})
 
+	c.Assert(s.cmdPartx.Calls(), DeepEquals, [][]string{
+		{"partx", "-u", "/dev/node"},
+	})
+
 	// check that the OnDiskVolume was updated as expected
 	c.Assert(dl.Structure, DeepEquals, []gadget.OnDiskStructure{
 		{
-			LaidOutStructure: gadget.LaidOutStructure{
-				VolumeStructure: &gadget.VolumeStructure{
-					Name: "BIOS Boot",
-					Size: 1024 * 1024,
-					Type: "21686148-6449-6E6F-744E-656564454649",
-					ID:   "2E59D969-52AB-430B-88AC-F83873519F6F",
-				},
-				StartOffset: 1024 * 1024,
-				Index:       1,
-			},
-			Node: "/dev/node1",
-			Size: 1024 * 1024,
+			Name:        "BIOS Boot",
+			Size:        1024 * 1024,
+			Type:        "21686148-6449-6E6F-744E-656564454649",
+			StartOffset: 1024 * 1024,
+			DiskIndex:   1,
+			Node:        "/dev/node1",
 		},
 		{
-			LaidOutStructure: gadget.LaidOutStructure{
-				VolumeStructure: &gadget.VolumeStructure{
-					Label:      "ubuntu-seed",
-					Name:       "Recovery",
-					Size:       2457600 * 512,
-					Type:       "C12A7328-F81F-11D2-BA4B-00A0C93EC93B",
-					ID:         "44C3D5C3-CAE1-4306-83E8-DF437ACDB32F",
-					Filesystem: "vfat",
+			PartitionFSLabel: "ubuntu-seed",
+			Name:             "Recovery",
+			Type:             "C12A7328-F81F-11D2-BA4B-00A0C93EC93B",
+			PartitionFSType:  "vfat",
+			StartOffset:      1024*1024 + 1024*1024,
+			DiskIndex:        2,
+			Node:             "/dev/node2",
+			Size:             2457600 * 512,
+		},
+	})
+}
+
+func (s *partitionTestSuite) TestRemovePartitionsWithDeviceRescan(c *C) {
+	devPath := filepath.Join(s.dir, "/sys/foo/")
+	m := map[string]*disks.MockDiskMapping{
+		"/dev/node": {
+			DevNum:  "42:0",
+			DevNode: "/dev/node",
+			DevPath: devPath,
+			// assume GPT backup header section is 34 sectors long
+			DiskSizeInBytes:     (8388574 + 34) * 512,
+			DiskUsableSectorEnd: 8388574 + 1,
+			DiskSchema:          "gpt",
+			ID:                  "9151F25B-CDF0-48F1-9EDE-68CBD616E2CA",
+			SectorSizeBytes:     512,
+			Structure: []disks.Partition{
+				// all 3 partitions present
+				{
+					KernelDeviceNode: "/dev/node1",
+					StartInBytes:     2048 * 512,
+					SizeInBytes:      2048 * 512,
+					PartitionType:    "21686148-6449-6E6F-744E-656564454649",
+					PartitionUUID:    "2E59D969-52AB-430B-88AC-F83873519F6F",
+					PartitionLabel:   "BIOS Boot",
+					Major:            42,
+					Minor:            1,
+					DiskIndex:        1,
+				},
+				{
+					KernelDeviceNode: "/dev/node2",
+					StartInBytes:     4096 * 512,
+					SizeInBytes:      2457600 * 512,
+					PartitionType:    "C12A7328-F81F-11D2-BA4B-00A0C93EC93B",
+					PartitionUUID:    "44C3D5C3-CAE1-4306-83E8-DF437ACDB32F",
+					PartitionLabel:   "Recovery",
+					Major:            42,
+					Minor:            2,
+					DiskIndex:        2,
+					FilesystemType:   "vfat",
+					FilesystemUUID:   "A644-B807",
+					FilesystemLabel:  "ubuntu-seed",
+				},
+				{
+					KernelDeviceNode: "/dev/node3",
+					StartInBytes:     2461696 * 512,
+					SizeInBytes:      2457600 * 512,
+					PartitionType:    "0FC63DAF-8483-4772-8E79-3D69D8477DE4",
+					PartitionUUID:    "F940029D-BFBB-4887-9D44-321E85C63866",
+					PartitionLabel:   "Writable",
+					Major:            42,
+					Minor:            3,
+					DiskIndex:        3,
+					FilesystemType:   "ext4",
+					FilesystemUUID:   "8781-433a",
+					FilesystemLabel:  "ubuntu-data",
 				},
-
-				StartOffset: 1024*1024 + 1024*1024,
-				Index:       2,
 			},
+		},
+	}
 
-			Node: "/dev/node2",
-			Size: 2457600 * 512,
+	// make empty device rescan file
+
+	err := os.MkdirAll(filepath.Join(devPath, "device"), 0755)
+	c.Assert(err, IsNil)
+
+	err = os.WriteFile(filepath.Join(devPath, "device", "rescan"), nil, 0755)
+	c.Assert(err, IsNil)
+
+	restore := disks.MockDeviceNameToDiskMapping(m)
+	defer restore()
+
+	cmdSfdisk := testutil.MockCommand(c, "sfdisk", "")
+	defer cmdSfdisk.Restore()
+
+	cmdUdevadm := testutil.MockCommand(c, "udevadm", "")
+	defer cmdUdevadm.Restore()
+
+	dl, err := gadget.OnDiskVolumeFromDevice("/dev/node")
+	c.Assert(err, IsNil)
+
+	err = gadgettest.MakeMockGadget(s.gadgetRoot, gadgetContent)
+	c.Assert(err, IsNil)
+
+	// add the file to indicate we should do the device/rescan trick
+	err = os.WriteFile(filepath.Join(s.gadgetRoot, "meta", "force-partition-table-reload-via-device-rescan"), nil, 0755)
+	c.Assert(err, IsNil)
+
+	gInfo, err := gadget.ReadInfoAndValidate(s.gadgetRoot, uc20Mod, nil)
+	c.Assert(err, IsNil)
+
+	err = install.RemoveCreatedPartitions(s.gadgetRoot, gInfo.Volumes["pc"], dl)
+	c.Assert(err, IsNil)
+
+	c.Assert(cmdSfdisk.Calls(), DeepEquals, [][]string{
+		{"sfdisk", "--no-reread", "--delete", "/dev/node", "3"},
+	})
+
+	// didn't call partx
+	c.Assert(s.cmdPartx.Calls(), HasLen, 0)
+
+	// but we did write to the sysfs file
+	c.Assert(filepath.Join(devPath, "device", "rescan"), testutil.FileEquals, "1\n")
+
+	// check that the OnDiskVolume was updated as expected
+	c.Assert(dl.Structure, DeepEquals, []gadget.OnDiskStructure{
+		{
+			Name:        "BIOS Boot",
+			Type:        "21686148-6449-6E6F-744E-656564454649",
+			StartOffset: 1024 * 1024,
+			DiskIndex:   1,
+			Node:        "/dev/node1",
+			Size:        1024 * 1024,
+		},
+		{
+			PartitionFSLabel: "ubuntu-seed",
+			Name:             "Recovery",
+			Type:             "C12A7328-F81F-11D2-BA4B-00A0C93EC93B",
+			PartitionFSType:  "vfat",
+			StartOffset:      1024*1024 + 1024*1024,
+			DiskIndex:        2,
+			Node:             "/dev/node2",
+			Size:             2457600 * 512,
 		},
 	})
 }
@@ -488,7 +742,7 @@ func (s *partitionTestSuite) TestRemovePartitionsNonAdjacent(c *C) {
 					PartitionLabel:   "BIOS Boot",
 					Major:            42,
 					Minor:            1,
-					StructureIndex:   1,
+					DiskIndex:        1,
 				},
 				{
 					KernelDeviceNode: "/dev/node2",
@@ -499,7 +753,7 @@ func (s *partitionTestSuite) TestRemovePartitionsNonAdjacent(c *C) {
 					PartitionLabel:   "Writable",
 					Major:            42,
 					Minor:            2,
-					StructureIndex:   2,
+					DiskIndex:        2,
 					FilesystemType:   "ext4",
 					FilesystemUUID:   "8781-433a",
 					FilesystemLabel:  "ubuntu-data",
@@ -513,7 +767,7 @@ func (s *partitionTestSuite) TestRemovePartitionsNonAdjacent(c *C) {
 					PartitionLabel:   "Recovery",
 					Major:            42,
 					Minor:            3,
-					StructureIndex:   3,
+					DiskIndex:        3,
 					FilesystemType:   "vfat",
 					FilesystemUUID:   "A644-B807",
 					FilesystemLabel:  "ubuntu-seed",
@@ -534,12 +788,12 @@ func (s *partitionTestSuite) TestRemovePartitionsNonAdjacent(c *C) {
 	dl, err := gadget.OnDiskVolumeFromDevice("/dev/node")
 	c.Assert(err, IsNil)
 
-	err = makeMockGadget(s.gadgetRoot, gadgetContentDifferentOrder)
+	err = gadgettest.MakeMockGadget(s.gadgetRoot, gadgetContentDifferentOrder)
 	c.Assert(err, IsNil)
-	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
+	gInfo, err := gadget.ReadInfoAndValidate(s.gadgetRoot, uc20Mod, nil)
 	c.Assert(err, IsNil)
 
-	err = install.RemoveCreatedPartitions(pv, dl)
+	err = install.RemoveCreatedPartitions(s.gadgetRoot, gInfo.Volumes["pc"], dl)
 	c.Assert(err, IsNil)
 
 	c.Assert(cmdSfdisk.Calls(), DeepEquals, [][]string{
@@ -549,36 +803,22 @@ func (s *partitionTestSuite) TestRemovePartitionsNonAdjacent(c *C) {
 	// check that the OnDiskVolume was updated as expected
 	c.Assert(dl.Structure, DeepEquals, []gadget.OnDiskStructure{
 		{
-			LaidOutStructure: gadget.LaidOutStructure{
-				VolumeStructure: &gadget.VolumeStructure{
-					Name: "BIOS Boot",
-					Size: 1024 * 1024,
-					Type: "21686148-6449-6E6F-744E-656564454649",
-					ID:   "2E59D969-52AB-430B-88AC-F83873519F6F",
-				},
-				StartOffset: 1024 * 1024,
-				Index:       1,
-			},
-			Node: "/dev/node1",
-			Size: 1024 * 1024,
+			Name:        "BIOS Boot",
+			Type:        "21686148-6449-6E6F-744E-656564454649",
+			StartOffset: 1024 * 1024,
+			DiskIndex:   1,
+			Node:        "/dev/node1",
+			Size:        1024 * 1024,
 		},
 		{
-			LaidOutStructure: gadget.LaidOutStructure{
-				VolumeStructure: &gadget.VolumeStructure{
-					Label:      "ubuntu-seed",
-					Name:       "Recovery",
-					Size:       2457600 * 512,
-					Type:       "C12A7328-F81F-11D2-BA4B-00A0C93EC93B",
-					ID:         "44C3D5C3-CAE1-4306-83E8-DF437ACDB32F",
-					Filesystem: "vfat",
-				},
-
-				StartOffset: 1024*1024 + 1024*1024 + 2457600*512,
-				Index:       3,
-			},
-
-			Node: "/dev/node3",
-			Size: 2457600 * 512,
+			PartitionFSLabel: "ubuntu-seed",
+			Name:             "Recovery",
+			Type:             "C12A7328-F81F-11D2-BA4B-00A0C93EC93B",
+			PartitionFSType:  "vfat",
+			StartOffset:      1024*1024 + 1024*1024 + 2457600*512,
+			Node:             "/dev/node3",
+			DiskIndex:        3,
+			Size:             2457600 * 512,
 		},
 	})
 }
@@ -595,14 +835,14 @@ func (s *partitionTestSuite) TestEnsureNodesExist(c *C) {
 		c.Logf("utErr:%q err:%q", tc.utErr, tc.err)
 
 		node := filepath.Join(c.MkDir(), "node")
-		err := ioutil.WriteFile(node, nil, 0644)
+		err := os.WriteFile(node, nil, 0644)
 		c.Assert(err, IsNil)
 
 		cmdUdevadm := testutil.MockCommand(c, "udevadm", fmt.Sprintf(mockUdevadmScript, tc.utErr))
 		defer cmdUdevadm.Restore()
 
-		ds := []gadget.OnDiskStructure{{Node: node}}
-		err = install.EnsureNodesExist(ds, 10*time.Millisecond)
+		nodes := []string{node}
+		err = install.EnsureNodesExist(nodes, 10*time.Millisecond)
 		if tc.err == "" {
 			c.Assert(err, IsNil)
 		} else {
@@ -620,10 +860,10 @@ func (s *partitionTestSuite) TestEnsureNodesExistTimeout(c *C) {
 	defer cmdUdevadm.Restore()
 
 	node := filepath.Join(c.MkDir(), "node")
-	ds := []gadget.OnDiskStructure{{Node: node}}
+	nodes := []string{node}
 	t := time.Now()
 	timeout := 1 * time.Second
-	err := install.EnsureNodesExist(ds, timeout)
+	err := install.EnsureNodesExist(nodes, timeout)
 	c.Assert(err, ErrorMatches, fmt.Sprintf("device %s not available", node))
 	c.Assert(time.Since(t) >= timeout, Equals, true)
 	c.Assert(cmdUdevadm.Calls(), HasLen, 0)
@@ -666,6 +906,105 @@ const gptGadgetContentWithSave = `volumes:
         size: 1200M
 `
 
+const gptGadgetContentWithGap = `volumes:
+  pc:
+    bootloader: grub
+    partial: [ structure ]
+    structure:
+      - name: Recovery
+        offset: 2M
+        role: system-seed
+        filesystem: vfat
+        # UEFI will boot the ESP partition by default first
+        type: EF,C12A7328-F81F-11D2-BA4B-00A0C93EC93B
+        size: 1200M
+        content:
+          - source: grubx64.efi
+            target: EFI/boot/grubx64.efi
+      - name: Save
+        role: system-save
+        filesystem: ext4
+        type: 83,0FC63DAF-8483-4772-8E79-3D69D8477DE4
+        size: 128M
+      - name: Writable
+        role: system-data
+        filesystem: ext4
+        type: 83,0FC63DAF-8483-4772-8E79-3D69D8477DE4
+        size: 1200M
+`
+
+const gptGadgetContentWithMinSize = `volumes:
+  pc:
+    bootloader: grub
+    structure:
+      - name: mbr
+        type: mbr
+        size: 440
+        content:
+          - image: pc-boot.img
+      - name: BIOS Boot
+        type: DA,21686148-6449-6E6F-744E-656564454649
+        size: 1M
+        offset: 1M
+        offset-write: mbr+92
+        content:
+          - image: pc-core.img
+      - name: Recovery
+        role: system-seed
+        filesystem: vfat
+        # UEFI will boot the ESP partition by default first
+        type: EF,C12A7328-F81F-11D2-BA4B-00A0C93EC93B
+        size: 1200M
+        content:
+          - source: grubx64.efi
+            target: EFI/boot/grubx64.efi
+      - name: Save
+        role: system-save
+        filesystem: ext4
+        type: 83,0FC63DAF-8483-4772-8E79-3D69D8477DE4
+        min-size: 128M
+        size: 256M
+      - name: Writable
+        role: system-data
+        filesystem: ext4
+        type: 83,0FC63DAF-8483-4772-8E79-3D69D8477DE4
+        size: 1200M
+`
+
+const gptGadgetContentWithRangeForSeed = `volumes:
+  pc:
+    bootloader: grub
+    structure:
+      - name: mbr
+        type: mbr
+        size: 440
+        content:
+          - image: pc-boot.img
+      - name: BIOS Boot
+        type: DA,21686148-6449-6E6F-744E-656564454649
+        size: 1M
+        offset: 1M
+      - name: Recovery
+        role: system-seed
+        filesystem: vfat
+        type: EF,C12A7328-F81F-11D2-BA4B-00A0C93EC93B
+        min-size: 1000M
+        size: 1400M
+        content:
+          - source: grubx64.efi
+            target: EFI/boot/grubx64.efi
+      - name: Save
+        role: system-save
+        filesystem: ext4
+        type: 83,0FC63DAF-8483-4772-8E79-3D69D8477DE4
+        size: 128M
+      - name: Writable
+        role: system-data
+        filesystem: ext4
+        type: 83,0FC63DAF-8483-4772-8E79-3D69D8477DE4
+        size: 1200M
+`
+
 func (s *partitionTestSuite) TestCreatedDuringInstallGPT(c *C) {
 	m := map[string]*disks.MockDiskMapping{
 		"node": {
@@ -686,7 +1025,7 @@ func (s *partitionTestSuite) TestCreatedDuringInstallGPT(c *C) {
 					PartitionLabel:   "BIOS Boot",
 					Major:            42,
 					Minor:            1,
-					StructureIndex:   1,
+					DiskIndex:        1,
 				},
 				{
 					KernelDeviceNode: "/dev/node2",
@@ -697,7 +1036,7 @@ func (s *partitionTestSuite) TestCreatedDuringInstallGPT(c *C) {
 					PartitionLabel:   "ubuntu-seed",
 					Major:            42,
 					Minor:            2,
-					StructureIndex:   2,
+					DiskIndex:        2,
 					FilesystemType:   "vfat",
 					FilesystemUUID:   "A644-B807",
 					FilesystemLabel:  "ubuntu-seed",
@@ -711,7 +1050,7 @@ func (s *partitionTestSuite) TestCreatedDuringInstallGPT(c *C) {
 					PartitionLabel:   "ubuntu-boot",
 					Major:            42,
 					Minor:            3,
-					StructureIndex:   3,
+					DiskIndex:        3,
 					FilesystemType:   "ext4",
 					FilesystemUUID:   "8781-433a",
 					FilesystemLabel:  "ubuntu-boot",
@@ -725,7 +1064,7 @@ func (s *partitionTestSuite) TestCreatedDuringInstallGPT(c *C) {
 					PartitionLabel:   "ubuntu-data",
 					Major:            42,
 					Minor:            4,
-					StructureIndex:   4,
+					DiskIndex:        4,
 					FilesystemType:   "ext4",
 					FilesystemUUID:   "8123-433a",
 					FilesystemLabel:  "ubuntu-data",
@@ -737,7 +1076,7 @@ func (s *partitionTestSuite) TestCreatedDuringInstallGPT(c *C) {
 	restore := disks.MockDeviceNameToDiskMapping(m)
 	defer restore()
 
-	err := makeMockGadget(s.gadgetRoot, gptGadgetContentWithSave)
+	err := gadgettest.MakeMockGadget(s.gadgetRoot, gptGadgetContentWithSave)
 	c.Assert(err, IsNil)
 	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
 	c.Assert(err, IsNil)
@@ -745,7 +1084,21 @@ func (s *partitionTestSuite) TestCreatedDuringInstallGPT(c *C) {
 	dl, err := gadget.OnDiskVolumeFromDevice("node")
 	c.Assert(err, IsNil)
 
-	list := install.CreatedDuringInstall(pv, dl)
+	list := install.CreatedDuringInstall(pv.Volume, dl)
+	// only save and writable should show up
+	c.Check(list, DeepEquals, []string{"/dev/node3", "/dev/node4"})
+
+	// min-size for ubuntu-save for this gadget will match the third partition size
+	// (but size wouldn't)
+	err = gadgettest.MakeMockGadget(s.gadgetRoot, gptGadgetContentWithMinSize)
+	c.Assert(err, IsNil)
+	pv, err = gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
+	c.Assert(err, IsNil)
+
+	dl, err = gadget.OnDiskVolumeFromDevice("node")
+	c.Assert(err, IsNil)
+
+	list = install.CreatedDuringInstall(pv.Volume, dl)
 	// only save and writable should show up
 	c.Check(list, DeepEquals, []string{"/dev/node3", "/dev/node4"})
 }
@@ -808,7 +1161,7 @@ func (s *partitionTestSuite) TestCreatedDuringInstallMBR(c *C) {
 					PartitionLabel:   "Recovery",
 					Major:            42,
 					Minor:            1,
-					StructureIndex:   1,
+					DiskIndex:        1,
 					FilesystemType:   "vfat",
 					FilesystemUUID:   "A644-B807",
 					FilesystemLabel:  "ubuntu-seed",
@@ -821,7 +1174,7 @@ func (s *partitionTestSuite) TestCreatedDuringInstallMBR(c *C) {
 					PartitionLabel:   "Boot",
 					Major:            42,
 					Minor:            2,
-					StructureIndex:   2,
+					DiskIndex:        2,
 					FilesystemType:   "vfat",
 					FilesystemUUID:   "A644-B807",
 					FilesystemLabel:  "ubuntu-boot",
@@ -834,7 +1187,7 @@ func (s *partitionTestSuite) TestCreatedDuringInstallMBR(c *C) {
 					PartitionLabel:   "Save",
 					Major:            42,
 					Minor:            3,
-					StructureIndex:   3,
+					DiskIndex:        3,
 					FilesystemType:   "ext4",
 					FilesystemUUID:   "8781-433a",
 					FilesystemLabel:  "ubuntu-save",
@@ -847,7 +1200,7 @@ func (s *partitionTestSuite) TestCreatedDuringInstallMBR(c *C) {
 					PartitionLabel:   "Data",
 					Major:            42,
 					Minor:            4,
-					StructureIndex:   4,
+					DiskIndex:        4,
 					FilesystemType:   "ext4",
 					FilesystemUUID:   "8123-433a",
 					FilesystemLabel:  "ubuntu-data",
@@ -862,11 +1215,11 @@ func (s *partitionTestSuite) TestCreatedDuringInstallMBR(c *C) {
 	dl, err := gadget.OnDiskVolumeFromDevice("node")
 	c.Assert(err, IsNil)
 
-	err = makeMockGadget(s.gadgetRoot, mbrGadgetContentWithSave)
+	err = gadgettest.MakeMockGadget(s.gadgetRoot, mbrGadgetContentWithSave)
 	c.Assert(err, IsNil)
 	pv, err := gadgettest.MustLayOutSingleVolumeFromGadget(s.gadgetRoot, "", uc20Mod)
 	c.Assert(err, IsNil)
 
-	list := install.CreatedDuringInstall(pv, dl)
+	list := install.CreatedDuringInstall(pv.Volume, dl)
 	c.Assert(list, DeepEquals, []string{"/dev/node2", "/dev/node3", "/dev/node4"})
 }
