@@ -43,6 +43,9 @@ type LayoutOptions struct {
 	// created and content gets written later.
 	IgnoreContent bool
 
+	// GadgetRootDir must be used only to find assets, not to load
+	// gadget.yaml, as we might be using information provided by an
+	// installer.
 	GadgetRootDir string
 	KernelRootDir string
 
@@ -58,8 +61,6 @@ const NonMBRStartOffset = 1 * quantity.OffsetMiB
 // structures within it
 type LaidOutVolume struct {
 	*Volume
-	// Size is the total size of the volume
-	Size quantity.Size
 	// LaidOutStructure is a list of structures within the volume, sorted
 	// by their start offsets
 	LaidOutStructure []LaidOutStructure
@@ -77,15 +78,16 @@ type PartiallyLaidOutVolume struct {
 // LaidOutStructure describes a VolumeStructure coming from the gadget plus the
 // OnDiskStructure that describes how it would be applied to a given disk and
 // additional content used when writing/updating data in the structure.
+//
+// Note that we need to be careful while using the fields in OnDiskStructure as
+// some times LaidOutStructure is created before we have information about the
+// finally matched partition. This is especially important for StartOffset and
+// Size fields. TODO We want to eventually create LaidOutStructure only after
+// this information is available.
 type LaidOutStructure struct {
 	OnDiskStructure
 	// VolumeStructure is the volume structure defined in gadget.yaml
 	VolumeStructure *VolumeStructure
-	// AbsoluteOffsetWrite is the resolved absolute position of offset-write
-	// for this structure element within the enclosing volume
-	AbsoluteOffsetWrite *quantity.Offset
-	// Index of the structure definition in gadget YAML, note this starts at 0.
-	YamlIndex int
 	// LaidOutContent is a list of raw content inside the structure
 	LaidOutContent []LaidOutContent
 	// ResolvedContent is a list of filesystem content that has all
@@ -93,13 +95,6 @@ type LaidOutStructure struct {
 	ResolvedContent []ResolvedContent
 }
 
-// IsRoleMBR returns whether a structure's role is MBR or not.
-// meh this function is weirdly placed, not sure what to do w/o making schemaMBR
-// constant exported
-func IsRoleMBR(ls LaidOutStructure) bool {
-	return ls.Role() == schemaMBR
-}
-
 // These accessors return currently what comes in the gadget, but will use
 // OnDiskVolume data when the latter is made part of LaidOutStructure.
 
@@ -126,7 +121,7 @@ func (l LaidOutStructure) Label() string {
 
 // Filesystem for formatting the structure.
 func (l LaidOutStructure) Filesystem() string {
-	return l.VolumeStructure.Filesystem
+	return l.VolumeStructure.LinuxFilesystem()
 }
 
 // Role for the structure as specified in the gadget.
@@ -134,7 +129,7 @@ func (l LaidOutStructure) Role() string {
 	return l.VolumeStructure.Role
 }
 
-// HasFilesystem returns true if the structure is using a filesystem.
+// HasFilesystem returns true if the gadget expects a filesystem.
 func (l *LaidOutStructure) HasFilesystem() bool {
 	return l.VolumeStructure.HasFilesystem()
 }
@@ -146,15 +141,9 @@ func (l *LaidOutStructure) IsPartition() bool {
 }
 
 func (p LaidOutStructure) String() string {
-	return fmtIndexAndName(p.YamlIndex, p.Name())
+	return fmtIndexAndName(p.VolumeStructure.YamlIndex, p.Name())
 }
 
-type byStartOffset []LaidOutStructure
-
-func (b byStartOffset) Len() int           { return len(b) }
-func (b byStartOffset) Swap(i, j int)      { b[i], b[j] = b[j], b[i] }
-func (b byStartOffset) Less(i, j int) bool { return b[i].StartOffset < b[j].StartOffset }
-
 // LaidOutContent describes raw content that has been placed within the
 // encompassing structure and volume
 //
@@ -165,9 +154,6 @@ type LaidOutContent struct {
 
 	// StartOffset defines the start offset of this content image
 	StartOffset quantity.Offset
-	// AbsoluteOffsetWrite is the resolved absolute position of offset-write
-	// for this content element within the enclosing volume
-	AbsoluteOffsetWrite *quantity.Offset
 	// Size is the maximum size occupied by this image
 	Size quantity.Size
 	// Index of the content in structure declaration inside gadget YAML
@@ -193,55 +179,35 @@ type ResolvedContent struct {
 	KernelUpdate bool
 }
 
-func layoutVolumeStructures(volume *Volume) (structures []LaidOutStructure, byName map[string]*LaidOutStructure, err error) {
-	structures = make([]LaidOutStructure, len(volume.Structure))
-	byName = make(map[string]*LaidOutStructure, len(volume.Structure))
-
-	for idx := range volume.Structure {
-		ps := LaidOutStructure{
-			VolumeStructure: &volume.Structure[idx],
-			YamlIndex:       idx,
+func layoutVSFromDiskData(volume *Volume, gadgetToDiskStruct map[int]*OnDiskStructure) (sts []LaidOutStructure, err error) {
+	sts = make([]LaidOutStructure, len(volume.Structure))
+	for i := range volume.Structure {
+		gs := &volume.Structure[i]
+		ds, ok := gadgetToDiskStruct[gs.YamlIndex]
+		if !ok {
+			return nil, fmt.Errorf("internal error: partition %q not in disk map", gs.Name)
 		}
-
-		if ps.Name() != "" {
-			byName[ps.Name()] = &ps
+		los := LaidOutStructure{
+			OnDiskStructure: *ds,
+			VolumeStructure: &volume.Structure[i],
 		}
-		// Fill the parts of OnDiskStructure that do not depend on the disk
-		// or on whether we are encrypting or not.
-		// TODO Eventually fill everything here by passing all needed info
-		ps.OnDiskStructure = OnDiskStructure{
-			Name:        ps.VolumeStructure.Name,
-			Type:        ps.VolumeStructure.Type,
-			StartOffset: *volume.Structure[idx].Offset,
-			Size:        ps.VolumeStructure.Size,
-		}
-
-		structures[idx] = ps
+		sts[i] = los
 	}
 
-	// sort by starting offset
-	sort.Sort(byStartOffset(structures))
-
-	previousEnd := quantity.Offset(0)
-	for idx, ps := range structures {
-		if ps.StartOffset < previousEnd {
-			return nil, nil, fmt.Errorf("cannot lay out volume, structure %v overlaps with preceding structure %v", ps, structures[idx-1])
-		}
-		previousEnd = ps.StartOffset + quantity.Offset(ps.VolumeStructure.Size)
-
-		offsetWrite, err := resolveOffsetWrite(ps.VolumeStructure.OffsetWrite, byName)
-		if err != nil {
-			return nil, nil, fmt.Errorf("cannot resolve offset-write of structure %v: %v", ps, err)
-		}
-		structures[idx].AbsoluteOffsetWrite = offsetWrite
-	}
-
-	return structures, byName, nil
+	return sts, nil
 }
 
-// LayoutVolumePartially attempts to lay out only the structures in the volume.
-func LayoutVolumePartially(volume *Volume) (*PartiallyLaidOutVolume, error) {
-	structures, _, err := layoutVolumeStructures(volume)
+func layoutVolumeStructures(volume *Volume, gadgetToDiskStruct map[int]*OnDiskStructure) (
+	structures []LaidOutStructure, err error) {
+	if len(gadgetToDiskStruct) == 0 {
+		return nil, fmt.Errorf("cannot lay out: internal error: no disk structures provided")
+	}
+	return layoutVSFromDiskData(volume, gadgetToDiskStruct)
+}
+
+// layoutVolumePartially attempts to lay out only the structures in the volume.
+func layoutVolumePartially(volume *Volume, gadgetToDiskStruct map[int]*OnDiskStructure) (*PartiallyLaidOutVolume, error) {
+	structures, err := layoutVolumeStructures(volume, gadgetToDiskStruct)
 	if err != nil {
 		return nil, err
 	}
@@ -253,23 +219,22 @@ func LayoutVolumePartially(volume *Volume) (*PartiallyLaidOutVolume, error) {
 	return vol, nil
 }
 
-func setOnDiskLabelAndTypeInLaidOuts(los []LaidOutStructure, encType secboot.EncryptionType) {
-	for i := range los {
-		los[i].PartitionFSLabel = los[i].Label()
-		los[i].PartitionFSType = los[i].Filesystem()
-		if encType != secboot.EncryptionTypeNone {
-			switch los[i].Role() {
-			case SystemData, SystemSave:
-				los[i].PartitionFSLabel += "-enc"
-				los[i].PartitionFSType = "crypto_LUKS"
-			}
+func setOnDiskLabelAndTypeInLaidOut(los *LaidOutStructure, encType secboot.EncryptionType) {
+	los.PartitionFSLabel = los.Label()
+	los.PartitionFSType = los.Filesystem()
+	if encType != secboot.EncryptionTypeNone {
+		switch los.Role() {
+		case SystemData, SystemSave:
+			los.PartitionFSLabel += "-enc"
+			los.PartitionFSType = "crypto_LUKS"
 		}
 	}
 }
 
 // LayoutVolume attempts to completely lay out the volume, that is the
-// structures and their content, using provided options.
-func LayoutVolume(volume *Volume, opts *LayoutOptions) (*LaidOutVolume, error) {
+// structures and their content, using provided map of gadget
+// structures to disk structures and options.
+func LayoutVolume(volume *Volume, gadgetToDiskStruct map[int]*OnDiskStructure, opts *LayoutOptions) (*LaidOutVolume, error) {
 	var err error
 	if opts == nil {
 		opts = &LayoutOptions{}
@@ -291,83 +256,85 @@ func LayoutVolume(volume *Volume, opts *LayoutOptions) (*LaidOutVolume, error) {
 		}
 	}
 
-	structures, byName, err := layoutVolumeStructures(volume)
+	structures, err := layoutVolumeStructures(volume, gadgetToDiskStruct)
 	if err != nil {
 		return nil, err
 	}
 
-	farthestEnd := quantity.Offset(0)
-	fartherstOffsetWrite := quantity.Offset(0)
-
-	for idx, ps := range structures {
-		if ps.AbsoluteOffsetWrite != nil && *ps.AbsoluteOffsetWrite > fartherstOffsetWrite {
-			fartherstOffsetWrite = *ps.AbsoluteOffsetWrite
+	for idx := range structures {
+		if err := fillLaidoutStructure(&structures[idx], kernelInfo, opts); err != nil {
+			return nil, err
 		}
-		if end := ps.StartOffset + quantity.Offset(ps.VolumeStructure.Size); end > farthestEnd {
-			farthestEnd = end
-		}
-
-		// Set appropriately label and type details
-		// TODO: set this in layoutVolumeStructures in the future.
-		setOnDiskLabelAndTypeInLaidOuts(structures, opts.EncType)
-
-		// Lay out raw content. This can be skipped when only partition
-		// creation is needed and is safe because each volume structure
-		// has a size so even without the structure content the layout
-		// can be calculated.
-		if !opts.IgnoreContent {
-			content, err := layOutStructureContent(opts.GadgetRootDir, &structures[idx], byName)
-			if err != nil {
-				return nil, err
-			}
-			for _, c := range content {
-				if c.AbsoluteOffsetWrite != nil && *c.AbsoluteOffsetWrite > fartherstOffsetWrite {
-					fartherstOffsetWrite = *c.AbsoluteOffsetWrite
-				}
-			}
-			structures[idx].LaidOutContent = content
-		}
-
-		// resolve filesystem content
-		if doResolveContent {
-			resolvedContent, err := resolveVolumeContent(opts.GadgetRootDir, opts.KernelRootDir, kernelInfo, &structures[idx], nil)
-			if err != nil {
-				return nil, err
-			}
-			structures[idx].ResolvedContent = resolvedContent
-		}
-	}
-
-	volumeSize := quantity.Size(farthestEnd)
-	if fartherstOffsetWrite+quantity.Offset(SizeLBA48Pointer) > farthestEnd {
-		volumeSize = quantity.Size(fartherstOffsetWrite) + SizeLBA48Pointer
 	}
 
 	vol := &LaidOutVolume{
 		Volume:           volume,
-		Size:             volumeSize,
 		LaidOutStructure: structures,
 	}
 	return vol, nil
 }
 
-func resolveVolumeContent(gadgetRootDir, kernelRootDir string, kernelInfo *kernel.Info, ps *LaidOutStructure, filter ResolvedContentFilterFunc) ([]ResolvedContent, error) {
-	if !ps.HasFilesystem() {
+func fillLaidoutStructure(los *LaidOutStructure, kernelInfo *kernel.Info, opts *LayoutOptions) (err error) {
+	setOnDiskLabelAndTypeInLaidOut(los, opts.EncType)
+	// Lay out raw content. This can be skipped when only partition
+	// creation is needed and is safe because each volume structure
+	// has a size so even without the structure content the layout
+	// can be calculated.
+	var content []LaidOutContent
+	if !opts.IgnoreContent && !los.HasFilesystem() {
+		content, err = layOutStructureContent(opts.GadgetRootDir, los)
+		if err != nil {
+			return err
+		}
+	}
+
+	// resolve filesystem content
+	var resolvedContent []ResolvedContent
+	doResolveContent := !(opts.IgnoreContent || opts.SkipResolveContent)
+	if doResolveContent {
+		resolvedContent, err = resolveVolumeContent(opts.GadgetRootDir, opts.KernelRootDir, kernelInfo, los.VolumeStructure, nil)
+		if err != nil {
+			return err
+		}
+	}
+
+	los.LaidOutContent = content
+	los.ResolvedContent = resolvedContent
+	return nil
+}
+
+// LayoutVolumeStructure lays out a structure given disk, gadget and kernel
+// snaps information, and some options.
+func LayoutVolumeStructure(dgpair *OnDiskAndGadgetStructurePair, kernelInfo *kernel.Info, opts *LayoutOptions) (*LaidOutStructure, error) {
+	los := &LaidOutStructure{
+		OnDiskStructure: *dgpair.DiskStructure,
+		VolumeStructure: dgpair.GadgetStructure,
+	}
+
+	if err := fillLaidoutStructure(los, kernelInfo, opts); err != nil {
+		return nil, err
+	}
+
+	return los, nil
+}
+
+func resolveVolumeContent(gadgetRootDir, kernelRootDir string, kernelInfo *kernel.Info, vs *VolumeStructure, filter ResolvedContentFilterFunc) ([]ResolvedContent, error) {
+	if !vs.HasFilesystem() {
 		// structures without a file system are not resolved here
 		return nil, nil
 	}
-	if len(ps.VolumeStructure.Content) == 0 {
+	if len(vs.Content) == 0 {
 		return nil, nil
 	}
 
-	content := make([]ResolvedContent, 0, len(ps.VolumeStructure.Content))
-	for idx := range ps.VolumeStructure.Content {
-		resolvedSource, kupdate, err := resolveContentPathOrRef(gadgetRootDir, kernelRootDir, kernelInfo, ps.VolumeStructure.Content[idx].UnresolvedSource)
+	content := make([]ResolvedContent, 0, len(vs.Content))
+	for idx := range vs.Content {
+		resolvedSource, kupdate, err := resolveContentPathOrRef(gadgetRootDir, kernelRootDir, kernelInfo, vs.Content[idx].UnresolvedSource)
 		if err != nil {
-			return nil, fmt.Errorf("cannot resolve content for structure %v at index %v: %v", ps, idx, err)
+			return nil, fmt.Errorf("cannot resolve content for structure #%d (%q) at index %v: %v", vs.YamlIndex, vs.Name, idx, err)
 		}
 		rc := ResolvedContent{
-			VolumeContent:  &ps.VolumeStructure.Content[idx],
+			VolumeContent:  &vs.Content[idx],
 			ResolvedSource: resolvedSource,
 			KernelUpdate:   kupdate,
 		}
@@ -386,13 +353,12 @@ func resolveVolumeContent(gadgetRootDir, kernelRootDir string, kernelInfo *kerne
 // an absolute path, a flag indicating whether the content is part of
 // a kernel update, or an error.
 func resolveContentPathOrRef(gadgetRootDir, kernelRootDir string, kernelInfo *kernel.Info, pathOrRef string) (resolved string, kupdate bool, err error) {
-
 	// TODO: add kernelRootDir == "" error too once all the higher
 	//       layers in devicestate call gadget.Update() with a
 	//       kernel dir set
 	switch {
 	case gadgetRootDir == "":
-		return "", false, fmt.Errorf("internal error: gadget root dir cannot beempty")
+		return "", false, fmt.Errorf("internal error: gadget root dir cannot be empty")
 	case pathOrRef == "":
 		return "", false, fmt.Errorf("cannot use empty source")
 	}
@@ -456,7 +422,7 @@ func getImageSize(path string) (quantity.Size, error) {
 	return quantity.Size(stat.Size()), nil
 }
 
-func layOutStructureContent(gadgetRootDir string, ps *LaidOutStructure, known map[string]*LaidOutStructure) ([]LaidOutContent, error) {
+func layOutStructureContent(gadgetRootDir string, ps *LaidOutStructure) ([]LaidOutContent, error) {
 	if ps.HasFilesystem() {
 		// structures with a filesystem do not need any extra layout
 		return nil, nil
@@ -490,18 +456,11 @@ func layOutStructureContent(gadgetRootDir string, ps *LaidOutStructure, known ma
 			actualSize = c.Size
 		}
 
-		offsetWrite, err := resolveOffsetWrite(c.OffsetWrite, known)
-		if err != nil {
-			return nil, fmt.Errorf("cannot resolve offset-write of structure %v content %q: %v", ps, c.Image, err)
-		}
-
 		content[idx] = LaidOutContent{
 			VolumeContent: &ps.VolumeStructure.Content[idx],
 			Size:          actualSize,
 			StartOffset:   ps.StartOffset + start,
 			Index:         idx,
-			// break for gofmt < 1.11
-			AbsoluteOffsetWrite: offsetWrite,
 		}
 		previousEnd = start + quantity.Offset(actualSize)
 		if quantity.Size(previousEnd) > ps.VolumeStructure.Size {
@@ -522,22 +481,28 @@ func layOutStructureContent(gadgetRootDir string, ps *LaidOutStructure, known ma
 	return content, nil
 }
 
-func resolveOffsetWrite(offsetWrite *RelativeOffset, knownStructs map[string]*LaidOutStructure) (*quantity.Offset, error) {
-	if offsetWrite == nil {
-		return nil, nil
+// checkGadgetContentImages checks that images provided in the gadget fit the
+// declared gadget content sizes.
+func checkGadgetContentImages(gadgetRootDir string, vs *VolumeStructure) error {
+	if vs.HasFilesystem() {
+		return nil
+	}
+	if len(vs.Content) == 0 {
+		return nil
 	}
 
-	var relativeToOffset quantity.Offset
-	if offsetWrite.RelativeTo != "" {
-		otherStruct, ok := knownStructs[offsetWrite.RelativeTo]
-		if !ok {
-			return nil, fmt.Errorf("refers to an unknown structure %q", offsetWrite.RelativeTo)
+	for _, c := range vs.Content {
+		fileSize, err := getImageSize(filepath.Join(gadgetRootDir, c.Image))
+		if err != nil {
+			return fmt.Errorf("structure #%d (%q): content %q: %v", vs.YamlIndex, vs.Name, c.Image, err)
+		}
+
+		if c.Size != 0 && c.Size < fileSize {
+			return fmt.Errorf("structure #%d (%q): content %q size %v is larger than declared %v", vs.YamlIndex, vs.Name, c.Image, fileSize, c.Size)
 		}
-		relativeToOffset = otherStruct.StartOffset
 	}
 
-	resolvedOffsetWrite := relativeToOffset + offsetWrite.Offset
-	return &resolvedOffsetWrite, nil
+	return nil
 }
 
 // ShiftStructureTo translates the starting offset of a laid out structure and
@@ -557,13 +522,12 @@ func ShiftStructureTo(ps LaidOutStructure, offset quantity.Offset) LaidOutStruct
 	return newPs
 }
 
-func isLayoutCompatible(current, new *PartiallyLaidOutVolume) error {
+func isLayoutCompatible(current, new *Volume) error {
 	if current.ID != new.ID {
 		return fmt.Errorf("incompatible ID change from %v to %v", current.ID, new.ID)
 	}
-	if current.Schema != new.Schema {
-		return fmt.Errorf("incompatible schema change from %v to %v",
-			current.Schema, new.Schema)
+	if err := checkCompatibleSchema(current, new); err != nil {
+		return err
 	}
 	if current.Bootloader != new.Bootloader {
 		return fmt.Errorf("incompatible bootloader change from %v to %v",
@@ -572,17 +536,15 @@ func isLayoutCompatible(current, new *PartiallyLaidOutVolume) error {
 
 	// XXX: the code below asssumes both volumes have the same number of
 	// structures, this limitation may be lifted later
-	if len(current.LaidOutStructure) != len(new.LaidOutStructure) {
+	if len(current.Structure) != len(new.Structure) {
 		return fmt.Errorf("incompatible change in the number of structures from %v to %v",
-			len(current.LaidOutStructure), len(new.LaidOutStructure))
+			len(current.Structure), len(new.Structure))
 	}
 
 	// at the structure level we expect the volume to be identical
-	for i := range current.LaidOutStructure {
-		from := &current.LaidOutStructure[i]
-		to := &new.LaidOutStructure[i]
-		if err := canUpdateStructure(from, to, new.Schema); err != nil {
-			return fmt.Errorf("incompatible structure %v change: %v", to, err)
+	for i := range current.Structure {
+		if err := canUpdateStructure(current, i, new, i); err != nil {
+			return fmt.Errorf("incompatible structure #%d (%q) change: %v", new.Structure[i].YamlIndex, new.Structure[i].Name, err)
 		}
 	}
 	return nil
