@@ -40,7 +40,6 @@ package apparmor
 import (
 	"bytes"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path"
 	"path/filepath"
@@ -65,7 +64,7 @@ var (
 	kernelFeatures        = apparmor_sandbox.KernelFeatures
 	parserFeatures        = apparmor_sandbox.ParserFeatures
 	loadProfiles          = apparmor_sandbox.LoadProfiles
-	unloadProfiles        = apparmor_sandbox.UnloadProfiles
+	removeCachedProfiles  = apparmor_sandbox.RemoveCachedProfiles
 
 	// make sure that apparmor profile fulfills the late discarding backend
 	// interface
@@ -162,10 +161,10 @@ func snapConfineFromSnapProfile(info *snap.Info) (dir, glob string, content map[
 	// We must test the ".real" suffix first, this is a workaround for
 	// https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=858004
 	vanillaProfilePath := filepath.Join(info.MountDir(), "/etc/apparmor.d/usr.lib.snapd.snap-confine.real")
-	vanillaProfileText, err := ioutil.ReadFile(vanillaProfilePath)
+	vanillaProfileText, err := os.ReadFile(vanillaProfilePath)
 	if os.IsNotExist(err) {
 		vanillaProfilePath = filepath.Join(info.MountDir(), "/etc/apparmor.d/usr.lib.snapd.snap-confine")
-		vanillaProfileText, err = ioutil.ReadFile(vanillaProfilePath)
+		vanillaProfileText, err = os.ReadFile(vanillaProfilePath)
 	}
 	if err != nil {
 		return "", "", nil, fmt.Errorf("cannot open apparmor profile for vanilla snap-confine: %s", err)
@@ -176,6 +175,24 @@ func snapConfineFromSnapProfile(info *snap.Info) (dir, glob string, content map[
 	patchedProfileText := bytes.Replace(
 		vanillaProfileText, []byte("/usr/lib/snapd/snap-confine"), []byte(snapConfineInCore), -1)
 
+	// Replace the path to the vanilla snap-confine apparmor snippets
+	patchedProfileText = bytes.Replace(
+		patchedProfileText, []byte("/var/lib/snapd/apparmor/snap-confine"), []byte(apparmor_sandbox.SnapConfineAppArmorDir), -1)
+
+	// To support non standard homedirs we currently use the home.d tunables, which are
+	// written to the system apparmor directory. However snapd vendors its own apparmor, which
+	// uses the readonly filesystem, which we cannot modify with our own snippets. So we force
+	// include the home.d tunables from /etc if necessary.
+	// We should be safely able to use "#include if exists" as the vendored apparmor supports this.
+	// XXX: Replace include home tunables until we have a better solution
+	features, _ := parserFeatures()
+	if strutil.ListContains(features, "snapd-internal") {
+		patchedProfileText = bytes.Replace(
+			patchedProfileText,
+			[]byte("#include <tunables/global>"),
+			[]byte("#include <tunables/global>\n#include if exists \"/etc/apparmor.d/tunables/home.d/\""), -1)
+	}
+
 	// Also replace the test providing access to verbatim
 	// /usr/lib/snapd/snap-confine, which is necessary because to execute snaps
 	// from strict snaps, we need to be able read and map
@@ -232,7 +249,7 @@ func snapConfineProfileName(snapName string, rev snap.Revision) string {
 //
 // Additionally it will cleanup stale apparmor profiles it created.
 func (b *Backend) setupSnapConfineReexec(info *snap.Info) error {
-	if err := os.MkdirAll(dirs.SnapConfineAppArmorDir, 0755); err != nil {
+	if err := os.MkdirAll(apparmor_sandbox.SnapConfineAppArmorDir, 0755); err != nil {
 		return fmt.Errorf("cannot create snap-confine policy directory: %s", err)
 	}
 	dir, glob, content, err := snapConfineFromSnapProfile(info)
@@ -246,10 +263,10 @@ func (b *Backend) setupSnapConfineReexec(info *snap.Info) error {
 
 	changed, removed, errEnsure := osutil.EnsureDirState(dir, glob, content)
 	if len(changed) == 0 {
-		// XXX: because NFS workaround is handled separately the same correct
-		// snap-confine profile may need to be re-loaded. This is because the
-		// profile contains include directives and those load a second file
-		// that has changed outside of the scope of EnsureDirState.
+		// XXX: because remote file system workaround is handled separately the
+		// same correct snap-confine profile may need to be re-loaded. This is
+		// because the profile contains include directives and those load a
+		// second file that has changed outside of the scope of EnsureDirState.
 		//
 		// To counter that, always reload the profile by pretending it had
 		// changed.
@@ -267,15 +284,15 @@ func (b *Backend) setupSnapConfineReexec(info *snap.Info) error {
 		aaFlags = apparmor_sandbox.SkipKernelLoad
 	}
 	errReload := loadProfiles(pathnames, cache, aaFlags)
-	errUnload := unloadProfiles(removed, cache)
+	errRemoveCached := removeCachedProfiles(removed, cache)
 	if errEnsure != nil {
 		return fmt.Errorf("cannot synchronize snap-confine apparmor profile: %s", errEnsure)
 	}
 	if errReload != nil {
 		return fmt.Errorf("cannot reload snap-confine apparmor profile: %s", errReload)
 	}
-	if errUnload != nil {
-		return fmt.Errorf("cannot unload snap-confine apparmor profile: %s", errReload)
+	if errRemoveCached != nil {
+		return fmt.Errorf("cannot remove cached snap-confine apparmor profile: %s", errReload)
 	}
 	return nil
 }
@@ -326,13 +343,15 @@ type profilePathsResults struct {
 	removed   []string
 }
 
-func (b *Backend) prepareProfiles(snapInfo *snap.Info, opts interfaces.ConfinementOptions, repo *interfaces.Repository) (prof *profilePathsResults, err error) {
-	snapName := snapInfo.InstanceName()
-	spec, err := repo.SnapSpecification(b.Name(), snapName)
+func (b *Backend) prepareProfiles(appSet *interfaces.SnapAppSet, opts interfaces.ConfinementOptions, repo *interfaces.Repository) (prof *profilePathsResults, err error) {
+	snapName := appSet.InstanceName()
+	spec, err := repo.SnapSpecification(b.Name(), appSet)
 	if err != nil {
 		return nil, fmt.Errorf("cannot obtain apparmor specification for snap %q: %s", snapName, err)
 	}
 
+	snapInfo := appSet.Info()
+
 	// Add snippets for parallel snap installation mapping
 	spec.(*Specification).AddOvername(snapInfo)
 
@@ -384,7 +403,7 @@ func (b *Backend) prepareProfiles(snapInfo *snap.Info, opts interfaces.Confineme
 		return nil, fmt.Errorf("cannot create directory for apparmor profiles %q: %s", dir, err)
 	}
 	changed, removedPaths, errEnsure := osutil.EnsureDirStateGlobs(dir, globs, content)
-	// XXX: in the old code this error was reported late, after doing load/unload.
+	// XXX: in the old code this error was reported late, after doing load/removeCached.
 	if errEnsure != nil {
 		return nil, fmt.Errorf("cannot synchronize security files for snap %q: %s", snapName, errEnsure)
 	}
@@ -420,12 +439,14 @@ func (b *Backend) prepareProfiles(snapInfo *snap.Info, opts interfaces.Confineme
 //
 // This method should be called after changing plug, slots, connections between
 // them or application present in the snap.
-func (b *Backend) Setup(snapInfo *snap.Info, opts interfaces.ConfinementOptions, repo *interfaces.Repository, tm timings.Measurer) error {
-	prof, err := b.prepareProfiles(snapInfo, opts, repo)
+func (b *Backend) Setup(appSet *interfaces.SnapAppSet, opts interfaces.ConfinementOptions, repo *interfaces.Repository, tm timings.Measurer) error {
+	prof, err := b.prepareProfiles(appSet, opts, repo)
 	if err != nil {
 		return err
 	}
 
+	snapInfo := appSet.Info()
+
 	// Load all changed profiles with a flag that asks apparmor to skip reading
 	// the cache (since we know those changed for sure).  This allows us to
 	// work despite time being wrong (e.g. in the past). For more details see
@@ -450,14 +471,14 @@ func (b *Backend) Setup(snapInfo *snap.Info, opts interfaces.ConfinementOptions,
 	timings.Run(tm, "load-profiles[unchanged]", fmt.Sprintf("load unchanged security profiles of snap %q", snapInfo.InstanceName()), func(nesttm timings.Measurer) {
 		errReloadOther = loadProfiles(prof.unchanged, apparmor_sandbox.CacheDir, aaFlags)
 	})
-	errUnload := unloadProfiles(prof.removed, apparmor_sandbox.CacheDir)
+	errRemoveCached := removeCachedProfiles(prof.removed, apparmor_sandbox.CacheDir)
 	if errReloadChanged != nil {
 		return errReloadChanged
 	}
 	if errReloadOther != nil {
 		return errReloadOther
 	}
-	return errUnload
+	return errRemoveCached
 }
 
 // SetupMany creates and loads apparmor profiles for multiple snaps.
@@ -467,12 +488,12 @@ func (b *Backend) Setup(snapInfo *snap.Info, opts interfaces.ConfinementOptions,
 // collects and returns them all.
 //
 // This method is useful mainly for regenerating profiles.
-func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string) interfaces.ConfinementOptions, repo *interfaces.Repository, tm timings.Measurer) []error {
+func (b *Backend) SetupMany(appSets []*interfaces.SnapAppSet, confinement func(snapName string) interfaces.ConfinementOptions, repo *interfaces.Repository, tm timings.Measurer) []error {
 	var allChangedPaths, allUnchangedPaths, allRemovedPaths []string
 	var fallback bool
-	for _, snapInfo := range snaps {
-		opts := confinement(snapInfo.InstanceName())
-		prof, err := b.prepareProfiles(snapInfo, opts, repo)
+	for _, set := range appSets {
+		opts := confinement(set.InstanceName())
+		prof, err := b.prepareProfiles(set, opts, repo)
 		if err != nil {
 			fallback = true
 			break
@@ -488,7 +509,7 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 			aaFlags |= apparmor_sandbox.SkipKernelLoad
 		}
 		var errReloadChanged error
-		timings.Run(tm, "load-profiles[changed-many]", fmt.Sprintf("load changed security profiles of %d snaps", len(snaps)), func(nesttm timings.Measurer) {
+		timings.Run(tm, "load-profiles[changed-many]", fmt.Sprintf("load changed security profiles of %d snaps", len(appSets)), func(nesttm timings.Measurer) {
 			errReloadChanged = loadProfiles(allChangedPaths, apparmor_sandbox.CacheDir, aaFlags)
 		})
 
@@ -497,11 +518,11 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 			aaFlags |= apparmor_sandbox.SkipKernelLoad
 		}
 		var errReloadOther error
-		timings.Run(tm, "load-profiles[unchanged-many]", fmt.Sprintf("load unchanged security profiles %d snaps", len(snaps)), func(nesttm timings.Measurer) {
+		timings.Run(tm, "load-profiles[unchanged-many]", fmt.Sprintf("load unchanged security profiles %d snaps", len(appSets)), func(nesttm timings.Measurer) {
 			errReloadOther = loadProfiles(allUnchangedPaths, apparmor_sandbox.CacheDir, aaFlags)
 		})
 
-		errUnload := unloadProfiles(allRemovedPaths, apparmor_sandbox.CacheDir)
+		errRemoveCached := removeCachedProfiles(allRemovedPaths, apparmor_sandbox.CacheDir)
 		if errReloadChanged != nil {
 			logger.Noticef("failed to batch-reload changed profiles: %s", errReloadChanged)
 			fallback = true
@@ -510,8 +531,8 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 			logger.Noticef("failed to batch-reload unchanged profiles: %s", errReloadOther)
 			fallback = true
 		}
-		if errUnload != nil {
-			logger.Noticef("failed to batch-unload profiles: %s", errUnload)
+		if errRemoveCached != nil {
+			logger.Noticef("failed to batch-remove cached profiles: %s", errRemoveCached)
 			fallback = true
 		}
 	}
@@ -519,28 +540,51 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 	var errors []error
 	// if an error was encountered when processing all profiles at once, re-try them one by one
 	if fallback {
-		for _, snapInfo := range snaps {
-			opts := confinement(snapInfo.InstanceName())
-			if err := b.Setup(snapInfo, opts, repo, tm); err != nil {
-				errors = append(errors, fmt.Errorf("cannot setup profiles for snap %q: %s", snapInfo.InstanceName(), err))
+		for _, set := range appSets {
+			instanceName := set.InstanceName()
+			opts := confinement(instanceName)
+			if err := b.Setup(set, opts, repo, tm); err != nil {
+				errors = append(errors, fmt.Errorf("cannot setup profiles for snap %q: %s", instanceName, err))
 			}
 		}
 	}
 	return errors
 }
 
-// Remove removes and unloads apparmor profiles of a given snap.
+// Removes all AppArmor profiles from disk but does not unload them from the
+// kernel - currently it is not possible to ensure that all services of a snap
+// are stopped and so it is not possible to safely unload the profile for a snap
+// from the kernel and so we only remove it from the cache on disk
+func RemoveAllSnapAppArmorProfiles() error {
+	dir := dirs.SnapAppArmorDir
+	globs := []string{"snap.*", "snap-update-ns.*", "snap-confine.*"}
+	cache := apparmor_sandbox.CacheDir
+	_, removed, errEnsure := osutil.EnsureDirStateGlobs(dir, globs, nil)
+	errRemoveCached := removeCachedProfiles(removed, cache)
+	switch {
+	case errEnsure != nil && errRemoveCached != nil:
+		return fmt.Errorf("cannot remove apparmor profiles: %s (and also %s)", errEnsure, errRemoveCached)
+	case errEnsure != nil:
+		return fmt.Errorf("cannot remove apparmor profiles: %s", errEnsure)
+	case errRemoveCached != nil:
+		return errRemoveCached
+	default:
+		return nil
+	}
+}
+
+// Remove removes the apparmor profiles of a given snap from disk and the cache.
 func (b *Backend) Remove(snapName string) error {
 	dir := dirs.SnapAppArmorDir
 	globs := profileGlobs(snapName)
 	cache := apparmor_sandbox.CacheDir
 	_, removed, errEnsure := osutil.EnsureDirStateGlobs(dir, globs, nil)
-	// always try to unload affected profiles
-	errUnload := unloadProfiles(removed, cache)
+	// always try to remove affected profiles from the cache
+	errRemoveCached := removeCachedProfiles(removed, cache)
 	if errEnsure != nil {
 		return fmt.Errorf("cannot synchronize security files for snap %q: %s", snapName, errEnsure)
 	}
-	return errUnload
+	return errRemoveCached
 }
 
 func (b *Backend) RemoveLate(snapName string, rev snap.Revision, typ snap.Type) error {
@@ -552,14 +596,15 @@ func (b *Backend) RemoveLate(snapName string, rev snap.Revision, typ snap.Type)
 
 	globs := []string{snapConfineProfileName(snapName, rev)}
 	_, removed, errEnsure := osutil.EnsureDirStateGlobs(dirs.SnapAppArmorDir, globs, nil)
-	// XXX: unloadProfiles() does not unload profiles from the kernel, but
-	// only removes profiles from the cache
-	// always try to unload the affected profile
-	errUnload := unloadProfiles(removed, apparmor_sandbox.CacheDir)
+	// XXX: we should also try and unload the profile from the kernel
+	// instead of just removing it from the cache but currently it is not
+	// possible to ensure all snap services are stopped at this time and so
+	// it is not safe to unload the profile
+	errRemoveCached := removeCachedProfiles(removed, apparmor_sandbox.CacheDir)
 	if errEnsure != nil {
 		return fmt.Errorf("cannot remove security profiles for snap %q (%s): %s", snapName, rev, errEnsure)
 	}
-	return errUnload
+	return errRemoveCached
 }
 
 var (
@@ -567,11 +612,6 @@ var (
 	coreRuntimePattern = regexp.MustCompile("^core([0-9][0-9])?$")
 )
 
-const (
-	attachPattern  = "(attach_disconnected,mediate_deleted)"
-	attachComplain = "(attach_disconnected,mediate_deleted,complain)"
-)
-
 func (b *Backend) deriveContent(spec *Specification, snapInfo *snap.Info, opts interfaces.ConfinementOptions) (content map[string]osutil.FileState) {
 	content = make(map[string]osutil.FileState, len(snapInfo.Apps)+len(snapInfo.Hooks)+1)
 
@@ -585,6 +625,9 @@ func (b *Backend) deriveContent(spec *Specification, snapInfo *snap.Info, opts i
 		securityTag := hookInfo.SecurityTag()
 		b.addContent(securityTag, snapInfo, "hook."+hookInfo.Name, opts, spec.SnippetForTag(securityTag), content, spec)
 	}
+
+	// TODO: something with component hooks will need to happen here
+
 	// Add profile for snap-update-ns if we have any apps or hooks.
 	// If we have neither then we don't have any need to create an executing environment.
 	// This applies to, for example, kernel snaps or gadget snaps (unless they have hooks).
@@ -612,6 +655,16 @@ func addUpdateNSProfile(snapInfo *snap.Info, snippets string, content map[string
 				snippets += strings.Replace(apparmor_sandbox.OverlayRootSnippet, "###UPPERDIR###", overlayRoot, -1)
 			}
 			return snippets
+		case "###INCLUDE_SYSTEM_TUNABLES_HOME_D_WITH_VENDORED_APPARMOR###":
+			// XXX: refactor this so that we don't have to duplicate this part.
+			// TODO: rewrite this whole mess with go templates.
+			features, _ := parserFeatures()
+			if strutil.ListContains(features, "snapd-internal") {
+				return `#include if exists "/etc/apparmor.d/tunables/home.d"`
+			}
+			return ""
+		default:
+			// TODO: Warn that an invalid pattern is being used.
 		}
 		return ""
 	})
@@ -643,13 +696,6 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 		policy = classicTemplate
 		ignoreSnippets = true
 	}
-	// If a snap is in devmode (or is using classic confinement) then make the
-	// profile non-enforcing where violations are logged but not denied.
-	// This is also done for classic so that no confinement applies. Just in
-	// case the profile we start with is not permissive enough.
-	if (opts.DevMode || opts.Classic) && !opts.JailMode {
-		policy = strings.Replace(policy, attachPattern, attachComplain, -1)
-	}
 	policy = templatePattern.ReplaceAllStringFunc(policy, func(placeholder string) string {
 		switch placeholder {
 		case "###DEVMODE_SNAP_CONFINE###":
@@ -742,7 +788,7 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 				// initial seed change and continue on. This code will be
 				// removed/adapted before it is merged to the main branch,
 				// it is only meant to exist on the security release branch.
-				msg := fmt.Sprintf("neither snapd nor core snap available while preparing apparmor profile for devmode snap %s, panicing to restart snapd to continue seeding", snapInfo.InstanceName())
+				msg := fmt.Sprintf("neither snapd nor core snap available while preparing apparmor profile for devmode snap %s, panicking to restart snapd to continue seeding", snapInfo.InstanceName())
 				panic(msg)
 			}
 
@@ -802,10 +848,56 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 				return `#include if exists "/var/lib/snapd/apparmor/snap-tuning"`
 			}
 			return ""
+		// XXX: Remove this when we have a better solution to including the system
+		// tunables. See snapConfineFromSnapProfile() for a more detailed explanation.
+		case "###INCLUDE_SYSTEM_TUNABLES_HOME_D_WITH_VENDORED_APPARMOR###":
+			features, _ := parserFeatures()
+			if strutil.ListContains(features, "snapd-internal") {
+				return `#include if exists "/etc/apparmor.d/tunables/home.d"`
+			}
+			return ""
 		case "###VAR###":
 			return templateVariables(snapInfo, securityTag, cmdName)
 		case "###PROFILEATTACH###":
 			return fmt.Sprintf("profile \"%s\"", securityTag)
+		case "###FLAGS###":
+			// default flags
+			flags := []string{"attach_disconnected", "mediate_deleted"}
+			if spec.Unconfined() == UnconfinedEnabled {
+				// need both parser and kernel support for unconfined
+				pfeatures, _ := parserFeatures()
+				kfeatures, _ := kernelFeatures()
+				if strutil.ListContains(pfeatures, "unconfined") &&
+					strutil.ListContains(kfeatures, "policy:unconfined_restrictions") {
+					flags = append(flags, "unconfined")
+				}
+			}
+			// If a snap is in devmode (or is using classic confinement) then make the
+			// profile non-enforcing where violations are logged but not denied.
+			// This is also done for classic so that no confinement applies. Just in
+			// case the profile we start with is not permissive enough.
+			if (opts.DevMode || opts.Classic) && !opts.JailMode {
+				if !strutil.ListContains(flags, "unconfined") {
+					// Profile modes unconfined and complain
+					// conflict with each other and are
+					// rejected by the parser, in any case
+					// this is fine since we already
+					// requested unconfined based on the
+					// spec and complain would no enforce
+					// any rules anyway.
+					flags = append(flags, "complain")
+				}
+			}
+			if len(flags) > 0 {
+				return "flags=(" + strings.Join(flags, ",") + ")"
+			} else {
+				return ""
+			}
+		case "###PYCACHEDENY###":
+			if spec.SuppressPycacheDeny() {
+				return ""
+			}
+			return pycacheDenySnippet
 		case "###CHANGEPROFILE_RULE###":
 			features, _ := parserFeatures()
 			if strutil.ListContains(features, "unsafe") {
@@ -824,12 +916,13 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 				// ignoring all apparmor snippets as they may conflict with the
 				// super-broad template we are starting with.
 			} else {
-				// Check if NFS is mounted at or under $HOME. Because NFS is not
-				// transparent to apparmor we must alter the profile to counter that and
-				// allow access to SNAP_USER_* files.
+				// Check if a remote file system is mounted at or under $HOME.
+				// Because some file systems, like NFS, are not transparent to
+				// apparmor we must alter the profile to counter that and allow
+				// access to SNAP_USER_* files.
 				tagSnippets = snippetForTag
-				if nfs, _ := osutil.IsHomeUsingNFS(); nfs {
-					tagSnippets += apparmor_sandbox.NfsSnippet
+				if isRemote, _ := osutil.IsHomeUsingRemoteFS(); isRemote {
+					tagSnippets += apparmor_sandbox.RemoteFSSnippet
 				}
 
 				if overlayRoot, _ := isRootWritableOverlay(); overlayRoot != "" {
@@ -884,8 +977,8 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 }
 
 // NewSpecification returns a new, empty apparmor specification.
-func (b *Backend) NewSpecification() interfaces.Specification {
-	return &Specification{}
+func (b *Backend) NewSpecification(appSet *interfaces.SnapAppSet) interfaces.Specification {
+	return &Specification{appSet: appSet}
 }
 
 // SandboxFeatures returns the list of apparmor features supported by the kernel.
