@@ -27,7 +27,9 @@ import (
 	"github.com/snapcore/snapd/interfaces/kmod"
 	"github.com/snapcore/snapd/interfaces/seccomp"
 	"github.com/snapcore/snapd/release"
+	apparmor_sandbox "github.com/snapcore/snapd/sandbox/apparmor"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/strutil"
 )
 
 const dockerSupportSummary = `allows operating as the Docker daemon`
@@ -55,6 +57,11 @@ const dockerSupportConnectedPlugAppArmorCore = `
 /system-data/var/snap/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/@{SNAP_REVISION}/{,**} rwl,
 `
 
+const dockerSupportConnectedPlugAppArmorUserNS = `
+# allow use of user namespaces
+userns,
+`
+
 const dockerSupportConnectedPlugAppArmor = `
 # Description: allow operating as the Docker daemon/containerd. This policy is
 # intentionally not restrictive and is here to help guard against programming
@@ -1007,6 +1014,13 @@ func (iface *dockerSupportInterface) AppArmorConnectedPlug(spec *apparmor.Specif
 	// The 'change_profile unsafe' rules conflict with the 'ix' rules in
 	// the home interface, so suppress them (LP: #1797786)
 	spec.SetSuppressHomeIx()
+	// New enough docker & containers it launches appear to get
+	// denial for writing pycache inside the container... which I
+	// guess to apparmor it looks like a snap. This is harmless, as
+	// docker snap no longer ships any python, and thus will not
+	// try to modify, otherwise immutable, pycache inside the
+	// snaps.
+	spec.SetSuppressPycacheDeny()
 	spec.AddSnippet(dockerSupportConnectedPlugAppArmor)
 	if privileged {
 		spec.AddSnippet(dockerSupportPrivilegedAppArmor)
@@ -1014,6 +1028,17 @@ func (iface *dockerSupportInterface) AppArmorConnectedPlug(spec *apparmor.Specif
 	if !release.OnClassic {
 		spec.AddSnippet(dockerSupportConnectedPlugAppArmorCore)
 	}
+	// if apparmor supports userns mediation then add this too
+	if apparmor_sandbox.ProbedLevel() != apparmor_sandbox.Unsupported {
+		features, err := apparmor_sandbox.ParserFeatures()
+		if err != nil {
+			return err
+		}
+		if strutil.ListContains(features, "userns") {
+			spec.AddSnippet(dockerSupportConnectedPlugAppArmorUserNS)
+		}
+	}
+
 	spec.SetUsesPtraceTrace()
 	return nil
 }
