@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016-2017 Canonical Ltd
+ * Copyright (C) 2016-2023 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -233,8 +233,7 @@ capability setuid,
 
 # resolvconf
 /{,usr/}sbin/resolvconf ixr,
-/run/resolvconf/{,**} rk,
-/run/resolvconf/** w,
+/run/resolvconf/{,**} rwk,
 /etc/resolvconf/{,**} r,
 /{,usr/}lib/resolvconf/* ix,
 # Required by resolvconf
@@ -286,7 +285,7 @@ capability sys_admin, # for setns()
 network netlink raw,
 
 / r,
-/run/netns/ r,     # only 'r' since snap-confine will create this for us
+/run/netns/ rk,     # no 'w' since snap-confine will create this for us
 /run/netns/* rw,
 mount options=(rw, rshared) -> /run/netns/,
 mount options=(rw, bind) /run/netns/ -> /run/netns/,
@@ -300,6 +299,7 @@ capability sys_ptrace,
 # 'ip netns exec foo /bin/sh'
 mount options=(rw, rslave) /,
 mount options=(rw, rslave), # LP: #1648245
+mount fstype=sysfs,
 umount /sys/,
 
 # Eg, nsenter --net=/run/netns/... <command>
