@@ -37,10 +37,18 @@ func (s *cpSuite) TestCpMulti(c *C) {
 }
 
 func (s *cpSuite) TestDoCpErr(c *C) {
-	f1, err := os.Open(s.f1)
+	c.Assert(os.WriteFile(s.f2, nil, 0444), IsNil)
+
+	src, err := os.Open(s.f1)
 	c.Assert(err, IsNil)
-	st, err := f1.Stat()
+	defer src.Close()
+
+	roFd, err := os.Open(s.f2)
 	c.Assert(err, IsNil)
+	defer roFd.Close()
+
 	// force an error by asking it to write to a readonly stream
-	c.Check(osutil.DoCopyFile(f1, os.Stdin, st), NotNil)
+	st, err := src.Stat()
+	c.Assert(err, IsNil)
+	c.Check(osutil.DoCopyFile(src, roFd, st), NotNil)
 }
