@@ -22,7 +22,6 @@ package devicestate_test
 import (
 	"errors"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"strings"
@@ -61,6 +60,9 @@ import (
 	"github.com/snapcore/snapd/overlord/storecontext"
 	"github.com/snapcore/snapd/release"
 	"github.com/snapcore/snapd/secboot/keys"
+	"github.com/snapcore/snapd/seed"
+	"github.com/snapcore/snapd/seed/seedtest"
+	"github.com/snapcore/snapd/seed/seedwriter"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/snap/snaptest"
 	"github.com/snapcore/snapd/snapdenv"
@@ -103,6 +105,19 @@ type deviceMgrBaseSuite struct {
 	// test the actual functions, it can just call this in it's SetUpTest, see
 	// devicestate_cloudinit_test.go for details
 	restoreCloudInitStatusRestore func()
+
+	restoreProcessAutoImportAssertion func()
+}
+
+// mockRestartAndSettle expects the state to be locked
+func (s *deviceMgrBaseSuite) mockRestartAndSettle(c *C, st *state.State, chg *state.Change) {
+	restart.MockPending(st, restart.RestartUnset)
+	restart.MockAfterRestartForChange(chg)
+
+	st.Unlock()
+	defer st.Lock()
+	err := s.o.Settle(settleTimeout)
+	c.Check(err, IsNil)
 }
 
 type deviceMgrSuite struct {
@@ -131,6 +146,28 @@ func (sto *fakeStore) Assertion(assertType *asserts.AssertionType, key []string,
 	return ref.Resolve(sto.db.Find)
 }
 
+func (sto *fakeStore) SeqFormingAssertion(assertType *asserts.AssertionType, key []string, sequence int, user *auth.UserState) (asserts.Assertion, error) {
+	sto.pokeStateLock()
+
+	ref := &asserts.AtSequence{
+		Type:        assertType,
+		SequenceKey: key,
+		Sequence:    sequence,
+		Revision:    asserts.RevisionNotKnown,
+		Pinned:      sequence > 0,
+	}
+
+	if sequence <= 0 {
+		hdrs, err := asserts.HeadersFromSequenceKey(ref.Type, ref.SequenceKey)
+		if err != nil {
+			return nil, err
+		}
+		return sto.db.FindSequence(ref.Type, hdrs, -1, ref.Type.MaxSupportedFormat())
+	}
+
+	return ref.Resolve(sto.db.Find)
+}
+
 var (
 	brandPrivKey, _  = assertstest.GenerateKey(752)
 	brandPrivKey2, _ = assertstest.GenerateKey(752)
@@ -250,6 +287,11 @@ func (s *deviceMgrBaseSuite) setupBaseTest(c *C, classic bool) {
 	s.AddCleanup(devicestate.MockSecbootMarkSuccessful(func() error {
 		return nil
 	}))
+
+	s.restoreProcessAutoImportAssertion = devicestate.MockProcessAutoImportAssertion(func(*state.State, seed.Seed, asserts.RODatabase, func(batch *asserts.Batch) error) error {
+		return nil
+	})
+	s.AddCleanup(s.restoreProcessAutoImportAssertion)
 }
 
 func (s *deviceMgrBaseSuite) newStore(devBE storecontext.DeviceBackend) snapstate.StoreService {
@@ -265,6 +307,7 @@ func (s *deviceMgrBaseSuite) settle(c *C) {
 func (s *deviceMgrBaseSuite) seeding() {
 	chg := s.state.NewChange("seed", "Seed system")
 	chg.SetStatus(state.DoingStatus)
+	s.state.Set("seeded", false)
 }
 
 func (s *deviceMgrBaseSuite) makeModelAssertionInState(c *C, brandID, model string, extras map[string]interface{}) *asserts.Model {
@@ -591,7 +634,7 @@ func (s *deviceMgrSuite) TestDeviceManagerEnsureBootOkBootloaderHappy(c *C) {
 	snapstate.Set(s.state, "core", &snapstate.SnapState{
 		SnapType: "os",
 		Active:   true,
-		Sequence: []*snap.SideInfo{siCore1},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{siCore1}),
 		Current:  siCore1.Revision,
 	})
 
@@ -623,7 +666,7 @@ func (s *deviceMgrSuite) TestDeviceManagerEnsureBootOkUpdateBootRevisionsHappy(c
 	snapstate.Set(s.state, "kernel", &snapstate.SnapState{
 		SnapType: "kernel",
 		Active:   true,
-		Sequence: []*snap.SideInfo{siKernel1},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{siKernel1}),
 		Current:  siKernel1.Revision,
 	})
 
@@ -632,7 +675,7 @@ func (s *deviceMgrSuite) TestDeviceManagerEnsureBootOkUpdateBootRevisionsHappy(c
 	snapstate.Set(s.state, "core", &snapstate.SnapState{
 		SnapType: "os",
 		Active:   true,
-		Sequence: []*snap.SideInfo{siCore1, siCore2},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{siCore1, siCore2}),
 		Current:  siCore2.Revision,
 	})
 
@@ -1073,7 +1116,7 @@ func makeInstalledMockCoreSnapWithSnapdControl(c *C, st *state.State) *snap.Info
 	sideInfoCore11 := &snap.SideInfo{RealName: "core", Revision: snap.R(11), SnapID: "core-id"}
 	snapstate.Set(st, "core", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{sideInfoCore11},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{sideInfoCore11}),
 		Current:  sideInfoCore11.Revision,
 		SnapType: "os",
 	})
@@ -1107,7 +1150,7 @@ func makeInstalledMockSnap(c *C, st *state.State, yml string) *snap.Info {
 	sideInfo11 := &snap.SideInfo{RealName: "snap-with-snapd-control", Revision: snap.R(11), SnapID: "snap-with-snapd-control-id"}
 	snapstate.Set(st, "snap-with-snapd-control", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{sideInfo11},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{sideInfo11}),
 		Current:  sideInfo11.Revision,
 		SnapType: "app",
 	})
@@ -1121,7 +1164,7 @@ func makeInstalledMockKernelSnap(c *C, st *state.State, yml string) *snap.Info {
 	sideInfo11 := &snap.SideInfo{RealName: "pc-kernel", Revision: snap.R(11), SnapID: "pc-kernel-id"}
 	snapstate.Set(st, "pc-kernel", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{sideInfo11},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{sideInfo11}),
 		Current:  sideInfo11.Revision,
 		SnapType: "kernel",
 	})
@@ -1543,7 +1586,7 @@ func (s *deviceMgrSuite) TestDeviceManagerStartupUC20UbuntuSaveSystemCtlFails(c
 
 	err = mgr.StartUp()
 	c.Assert(err, NotNil)
-	c.Check(err.Error(), Equals, "cannot set up ubuntu-save: systemctl command [start var-lib-snapd-save.mount] failed with exit status 1: failed\n")
+	c.Check(err.Error(), Equals, "cannot set up ubuntu-save: systemctl command [start var-lib-snapd-save.mount] failed with exit status 1: failed")
 	c.Check(sysctlCmd.Calls(), DeepEquals, [][]string{
 		{"systemctl", "start", "var-lib-snapd-save.mount"},
 	})
@@ -2117,20 +2160,20 @@ func (s *deviceMgrSuite) TestDeviceManagerEnsurePostFactoryResetEncrypted(c *C)
 
 	// encrypted system
 	mockSnapFDEFile(c, "marker", nil)
-	err := ioutil.WriteFile(filepath.Join(dirs.SnapFDEDir, "ubuntu-save.key"),
+	err := os.WriteFile(filepath.Join(dirs.SnapFDEDir, "ubuntu-save.key"),
 		[]byte("save-key"), 0644)
 	c.Assert(err, IsNil)
 	c.Assert(os.MkdirAll(boot.InitramfsSeedEncryptionKeyDir, 0755), IsNil)
-	err = ioutil.WriteFile(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key"),
+	err = os.WriteFile(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key"),
 		[]byte("old"), 0644)
 	c.Assert(err, IsNil)
-	err = ioutil.WriteFile(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key.factory-reset"),
+	err = os.WriteFile(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key.factory-reset"),
 		[]byte("save"), 0644)
 	c.Assert(err, IsNil)
 	// matches the .factory key
 	factoryResetMarkercontent := []byte(`{"fallback-save-key-sha3-384":"d192153f0a50e826c6eb400c8711750ed0466571df1d151aaecc8c73095da7ec104318e7bf74d5e5ae2940827bf8402b"}
 `)
-	c.Assert(ioutil.WriteFile(filepath.Join(dirs.SnapDeviceDir, "factory-reset"), factoryResetMarkercontent, 0644), IsNil)
+	c.Assert(os.WriteFile(filepath.Join(dirs.SnapDeviceDir, "factory-reset"), factoryResetMarkercontent, 0644), IsNil)
 
 	completeCalls := 0
 	restore := devicestate.MockMarkFactoryResetComplete(func(encrypted bool) error {
@@ -2172,7 +2215,7 @@ func (s *deviceMgrSuite) TestDeviceManagerEnsurePostFactoryResetEncrypted(c *C)
 	c.Check(os.Rename(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key.factory-reset"),
 		filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key")),
 		IsNil)
-	c.Assert(ioutil.WriteFile(filepath.Join(dirs.SnapDeviceDir, "factory-reset"), factoryResetMarkercontent, 0644), IsNil)
+	c.Assert(os.WriteFile(filepath.Join(dirs.SnapDeviceDir, "factory-reset"), factoryResetMarkercontent, 0644), IsNil)
 
 	devicestate.SetPostFactoryResetRan(s.mgr, false)
 	err = s.mgr.Ensure()
@@ -2195,16 +2238,16 @@ func (s *deviceMgrSuite) TestDeviceManagerEnsurePostFactoryResetEncryptedError(c
 	// encrypted system
 	mockSnapFDEFile(c, "marker", nil)
 	c.Assert(os.MkdirAll(boot.InitramfsSeedEncryptionKeyDir, 0755), IsNil)
-	err := ioutil.WriteFile(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key"),
+	err := os.WriteFile(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key"),
 		[]byte("old"), 0644)
 	c.Check(err, IsNil)
-	err = ioutil.WriteFile(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key.factory-reset"),
+	err = os.WriteFile(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key.factory-reset"),
 		[]byte("save"), 0644)
 	c.Check(err, IsNil)
 	// does not match the save key
 	factoryResetMarkercontent := []byte(`{"fallback-save-key-sha3-384":"uh-oh"}
 `)
-	c.Assert(ioutil.WriteFile(filepath.Join(dirs.SnapDeviceDir, "factory-reset"), factoryResetMarkercontent, 0644), IsNil)
+	c.Assert(os.WriteFile(filepath.Join(dirs.SnapDeviceDir, "factory-reset"), factoryResetMarkercontent, 0644), IsNil)
 
 	completeCalls := 0
 	restore := devicestate.MockMarkFactoryResetComplete(func(encrypted bool) error {
@@ -2245,7 +2288,7 @@ func (s *deviceMgrSuite) TestDeviceManagerEnsurePostFactoryResetUnencrypted(c *C
 
 	// mock the factory reset marker of a system that isn't encrypted
 	c.Assert(os.MkdirAll(dirs.SnapDeviceDir, 0755), IsNil)
-	c.Assert(ioutil.WriteFile(filepath.Join(dirs.SnapDeviceDir, "factory-reset"), []byte("{}"), 0644), IsNil)
+	c.Assert(os.WriteFile(filepath.Join(dirs.SnapDeviceDir, "factory-reset"), []byte("{}"), 0644), IsNil)
 
 	completeCalls := 0
 	restore := devicestate.MockMarkFactoryResetComplete(func(encrypted bool) error {
@@ -2408,3 +2451,297 @@ func (s *deviceMgrSuite) TestEnsureExpiredUsersRemovedNotUnseeded(c *C) {
 	s.mockSystemUser(c, "remove-me", time.Now().Add(-(time.Minute * 5)))
 	s.testExpiredUserNotRemoved(c)
 }
+
+func (s *deviceMgrSuite) cacheDeviceCore20Seed(c *C) {
+
+	// now create a minimal uc20 seed dir with snaps/assertions
+	ss := &seedtest.SeedSnaps{
+		StoreSigning: s.storeSigning,
+		Brands:       s.brands,
+	}
+
+	// now create a minimal uc20 seed dir with snaps/assertions
+	seed20 := &seedtest.TestingSeed20{
+		SeedSnaps: *ss,
+		SeedDir:   dirs.SnapSeedDir,
+	}
+
+	gadgetYaml := `
+volumes:
+    volume-id:
+        bootloader: grub
+        structure:
+        - name: ubuntu-seed
+          role: system-seed
+          type: EF,C12A7328-F81F-11D2-BA4B-00A0C93EC93B
+          size: 1G
+        - name: ubuntu-data
+          role: system-data
+          type: 83,0FC63DAF-8483-4772-8E79-3D69D8477DE4
+          size: 2G
+`
+	makeSnap := func(yamlKey string) {
+		var files [][]string
+		if yamlKey == "pc=20" {
+			files = append(files, []string{"meta/gadget.yaml", gadgetYaml})
+		}
+		seed20.MakeAssertedSnap(c, seedtest.SampleSnapYaml[yamlKey], files, snap.R(1), "canonical", seed20.StoreSigning.Database)
+	}
+
+	makeSnap("snapd")
+	makeSnap("pc-kernel=20")
+	makeSnap("core20")
+	makeSnap("pc=20")
+	optSnapPath := snaptest.MakeTestSnapWithFiles(c, seedtest.SampleSnapYaml["optional20-a"], nil)
+
+	model := map[string]interface{}{
+		"display-name": "my model",
+		"architecture": "amd64",
+		"base":         "core20",
+		"grade":        "dangerous",
+		"snaps": []interface{}{
+			map[string]interface{}{
+				"name":            "pc-kernel",
+				"id":              seed20.AssertedSnapID("pc-kernel"),
+				"type":            "kernel",
+				"default-channel": "20",
+			},
+			map[string]interface{}{
+				"name":            "pc",
+				"id":              seed20.AssertedSnapID("pc"),
+				"type":            "gadget",
+				"default-channel": "20",
+			},
+			map[string]interface{}{
+				"name": "snapd",
+				"id":   seed20.AssertedSnapID("snapd"),
+				"type": "snapd",
+			},
+			map[string]interface{}{
+				"name": "core20",
+				"id":   seed20.AssertedSnapID("core20"),
+				"type": "base",
+			}},
+	}
+
+	modelAs := seed20.MakeSeed(c, "20220401", "my-brand", "my-model", model, []*seedwriter.OptionsSnap{{Path: optSnapPath}})
+	c.Assert(modelAs, NotNil)
+
+	// mock /he modeenv file
+	m := boot.Modeenv{
+		Mode:           "install",
+		RecoverySystem: "20220401",
+	}
+	err := m.WriteTo("")
+	c.Assert(err, IsNil)
+
+	// reload device seed
+	_, _, err = devicestate.ReloadEarlyDeviceSeed(s.mgr, state.ErrNoState)
+	c.Assert(err, IsNil)
+
+	// not fully realistic but avoids more mocking
+	devicestate.SetBootOkRan(s.mgr, true)
+}
+
+func (s *deviceMgrSuite) TestHandleAutoImportAssertionClassic(c *C) {
+	a := devicestate.MockProcessAutoImportAssertion(func(*state.State, seed.Seed, asserts.RODatabase, func(batch *asserts.Batch) error) error {
+		panic("trying to process auto-import-assertion on classic system")
+	})
+	defer a()
+
+	release.OnClassic = true
+	s.mockSystemMode(c, "run")
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	err := devicestate.EnsureAutoImportAssertions(s.mgr)
+	c.Check(err, IsNil)
+
+	// ensure state has not been changed
+	var autoImported bool
+	err = s.state.Get("asserts-early-auto-imported", &autoImported)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
+	c.Assert(autoImported, Equals, false)
+}
+
+func (s *deviceMgrSuite) testHandleAutoImportAssertionInstallModes(c *C, mode string) {
+	a := devicestate.MockProcessAutoImportAssertion(func(*state.State, seed.Seed, asserts.RODatabase, func(batch *asserts.Batch) error) error {
+		panic("trying to process auto-import-assertion in install modes")
+	})
+	defer a()
+
+	s.mockSystemMode(c, mode)
+	s.state.Lock()
+	s.cacheDeviceCore20Seed(c)
+	s.state.Set("seeded", nil)
+	s.state.Unlock()
+
+	err := devicestate.EnsureAutoImportAssertions(s.mgr)
+	c.Check(err, IsNil)
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	// ensure state has not been changed
+	var autoImported bool
+	err = s.state.Get("asserts-early-auto-imported", &autoImported)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
+	c.Assert(autoImported, Equals, false)
+}
+
+func (s *deviceMgrSuite) TestHandleAutoImportAssertionInstallMode(c *C) {
+	s.testHandleAutoImportAssertionInstallModes(c, "install")
+}
+
+func (s *deviceMgrSuite) TestHandleAutoImportAssertionFactoryResetMode(c *C) {
+	s.testHandleAutoImportAssertionInstallModes(c, "factory-reset")
+}
+
+func (s *deviceMgrSuite) TestHandleAutoImportAssertionWhenDone(c *C) {
+	a := devicestate.MockProcessAutoImportAssertion(func(*state.State, seed.Seed, asserts.RODatabase, func(batch *asserts.Batch) error) error {
+		panic("trying to process auto-import-assertion after it was already processed")
+	})
+	defer a()
+
+	s.mockSystemMode(c, "run")
+
+	// set state as processed
+	s.state.Lock()
+	s.state.Set("asserts-early-auto-imported", true)
+	s.cacheDeviceCore20Seed(c)
+	s.seeding()
+	s.state.Unlock()
+
+	err := devicestate.EnsureAutoImportAssertions(s.mgr)
+	c.Check(err, IsNil)
+
+	// check state has not changed
+	s.state.Lock()
+	defer s.state.Unlock()
+	var autoImported bool
+	err = s.state.Get("asserts-early-auto-imported", &autoImported)
+	c.Assert(err, IsNil)
+	c.Assert(autoImported, Equals, true)
+}
+
+func (s *deviceMgrSuite) TestHandleAutoImportAssertionNoSeedCache(c *C) {
+	a := devicestate.MockProcessAutoImportAssertion(func(*state.State, seed.Seed, asserts.RODatabase, func(batch *asserts.Batch) error) error {
+		panic("trying to process auto-import-assertion without cached system seed")
+	})
+	defer a()
+
+	s.mockSystemMode(c, "run")
+
+	err := devicestate.EnsureAutoImportAssertions(s.mgr)
+	c.Check(err, IsNil)
+
+	// ensure state has not been changed
+	s.state.Lock()
+	defer s.state.Unlock()
+	var autoImported bool
+	err = s.state.Get("asserts-early-auto-imported", &autoImported)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
+	c.Assert(autoImported, Equals, false)
+}
+
+func (s *deviceMgrSuite) TestHandleAutoImportAssertionFailed(c *C) {
+	a := devicestate.MockProcessAutoImportAssertion(func(*state.State, seed.Seed, asserts.RODatabase, func(batch *asserts.Batch) error) error {
+		return fmt.Errorf("failed to add user from system user assertions")
+	})
+	defer a()
+
+	s.mockSystemMode(c, "run")
+
+	s.state.Lock()
+	s.cacheDeviceCore20Seed(c)
+	s.seeding()
+	s.state.Unlock()
+
+	logbuf, restore := logger.MockLogger()
+	defer restore()
+
+	err := s.mgr.Ensure()
+	c.Check(err, IsNil)
+
+	// ensure state has not been changed
+	s.state.Lock()
+	defer s.state.Unlock()
+	var autoImported bool
+	err = s.state.Get("asserts-early-auto-imported", &autoImported)
+	c.Assert(err, IsNil)
+	c.Assert(autoImported, Equals, true)
+	c.Assert(logbuf.String(), testutil.Contains, `failed to add user from system user assertions`)
+}
+
+func (s *deviceMgrSuite) TestHandleAutoImportAssertionAlreadySeeded(c *C) {
+	a := devicestate.MockProcessAutoImportAssertion(func(*state.State, seed.Seed, asserts.RODatabase, func(batch *asserts.Batch) error) error {
+		return fmt.Errorf("failed to process auto import assertion")
+	})
+	defer a()
+
+	s.mockSystemMode(c, "run")
+
+	s.state.Lock()
+	s.cacheDeviceCore20Seed(c)
+	s.state.Set("seeded", true)
+	s.state.Unlock()
+
+	err := s.mgr.Ensure()
+	c.Check(err, IsNil)
+
+	// ensure state has not been changed
+	s.state.Lock()
+	defer s.state.Unlock()
+	var autoImported bool
+	err = s.state.Get("asserts-early-auto-imported", &autoImported)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
+	c.Assert(autoImported, Equals, false)
+}
+
+func (s *deviceMgrSuite) TestHandleAutoImportAssertionHappy(c *C) {
+
+	s.mockSystemMode(c, "run")
+
+	s.state.Lock()
+	s.cacheDeviceCore20Seed(c)
+	s.seeding()
+	s.state.Unlock()
+
+	err := s.mgr.Ensure()
+	c.Check(err, IsNil)
+
+	// check state is set as done
+	s.state.Lock()
+	defer s.state.Unlock()
+	var autoImported bool
+	err = s.state.Get("asserts-early-auto-imported", &autoImported)
+	c.Assert(err, IsNil)
+	c.Assert(autoImported, Equals, true)
+}
+
+func (s *deviceMgrSuite) TestDefaultRecoverySystem(c *C) {
+	// no recovery system set
+	s.state.Lock()
+	s.state.Set("default-recovery-system", nil)
+	s.state.Unlock()
+
+	_, err := s.mgr.DefaultRecoverySystem()
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
+
+	expectedSystem := devicestate.DefaultRecoverySystem{
+		System:   "label",
+		Model:    "model",
+		BrandID:  "brand",
+		Revision: 1,
+	}
+
+	// recovery system set
+	s.state.Lock()
+	s.state.Set("default-recovery-system", expectedSystem)
+	s.state.Unlock()
+
+	system, err := s.mgr.DefaultRecoverySystem()
+	c.Assert(err, IsNil)
+	c.Check(*system, Equals, expectedSystem)
+}
