@@ -23,13 +23,10 @@ import (
 	"bytes"
 	"compress/gzip"
 	"crypto"
-	"encoding/base64"
 	"encoding/hex"
 	"encoding/json"
 	"errors"
 	"fmt"
-	"io"
-	"io/ioutil"
 	"os"
 	"os/exec"
 	"path/filepath"
@@ -47,19 +44,16 @@ import (
 	"github.com/snapcore/snapd/gadget/install"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil"
-	"github.com/snapcore/snapd/overlord/assertstate"
+	installLogic "github.com/snapcore/snapd/overlord/install"
 	"github.com/snapcore/snapd/overlord/restart"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/progress"
-	"github.com/snapcore/snapd/randutil"
 	"github.com/snapcore/snapd/secboot"
 	"github.com/snapcore/snapd/secboot/keys"
 	"github.com/snapcore/snapd/seed"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/snap/snapfile"
-	"github.com/snapcore/snapd/snap/squashfs"
-	"github.com/snapcore/snapd/sysconfig"
 	"github.com/snapcore/snapd/systemd"
 	"github.com/snapcore/snapd/timings"
 )
@@ -76,51 +70,13 @@ var (
 	installWriteContent                  = install.WriteContent
 	installEncryptPartitions             = install.EncryptPartitions
 	installSaveStorageTraits             = install.SaveStorageTraits
+	installMatchDisksToGadgetVolumes     = install.MatchDisksToGadgetVolumes
 	secbootStageEncryptionKeyChange      = secboot.StageEncryptionKeyChange
 	secbootTransitionEncryptionKeyChange = secboot.TransitionEncryptionKeyChange
 
-	sysconfigConfigureTargetSystem = sysconfig.ConfigureTargetSystem
+	installLogicPrepareRunSystemData = installLogic.PrepareRunSystemData
 )
 
-func setSysconfigCloudOptions(opts *sysconfig.Options, gadgetDir string, model *asserts.Model) {
-	ubuntuSeedCloudCfg := filepath.Join(boot.InitramfsUbuntuSeedDir, "data/etc/cloud/cloud.cfg.d")
-
-	grade := model.Grade()
-
-	// we always set the cloud-init src directory if it exists, it is
-	// automatically ignored by sysconfig in the case it shouldn't be used
-	if osutil.IsDirectory(ubuntuSeedCloudCfg) {
-		opts.CloudInitSrcDir = ubuntuSeedCloudCfg
-	}
-
-	switch {
-	// if the gadget has a cloud.conf file, always use that regardless of grade
-	case sysconfig.HasGadgetCloudConf(gadgetDir):
-		opts.AllowCloudInit = true
-
-	// next thing is if are in secured grade and didn't have gadget config, we
-	// disable cloud-init always, clouds should have their own config via
-	// gadgets for grade secured
-	case grade == asserts.ModelSecured:
-		opts.AllowCloudInit = false
-
-	// all other cases we allow cloud-init to run, either through config that is
-	// available at runtime via a CI-DATA USB drive, or via config on
-	// ubuntu-seed if that is allowed by the model grade, etc.
-	default:
-		opts.AllowCloudInit = true
-	}
-}
-
-func writeModel(model *asserts.Model, where string) error {
-	f, err := os.OpenFile(where, os.O_WRONLY|os.O_CREATE|os.O_EXCL, 0644)
-	if err != nil {
-		return err
-	}
-	defer f.Close()
-	return asserts.NewEncoder(f).Encode(model)
-}
-
 func writeLogs(rootdir string, fromMode string) error {
 	// XXX: would be great to use native journal format but it's tied
 	//      to machine-id, we could journal -o export but there
@@ -156,31 +112,6 @@ func writeLogs(rootdir string, fromMode string) error {
 	return nil
 }
 
-func writeTimesyncdClock(srcRootDir, dstRootDir string) error {
-	// keep track of the time
-	const timesyncClockInRoot = "/var/lib/systemd/timesync/clock"
-	clockSrc := filepath.Join(srcRootDir, timesyncClockInRoot)
-	clockDst := filepath.Join(dstRootDir, timesyncClockInRoot)
-	if err := os.MkdirAll(filepath.Dir(clockDst), 0755); err != nil {
-		return fmt.Errorf("cannot store the clock: %v", err)
-	}
-	if !osutil.FileExists(clockSrc) {
-		logger.Noticef("timesyncd clock timestamp %v does not exist", clockSrc)
-		return nil
-	}
-	// clock file is owned by a specific user/group, thus preserve
-	// attributes of the source
-	if err := osutil.CopyFile(clockSrc, clockDst, osutil.CopyFlagPreserveAll); err != nil {
-		return fmt.Errorf("cannot copy clock: %v", err)
-	}
-	// the file is empty however, its modification timestamp is used to set
-	// up the current time
-	if err := os.Chtimes(clockDst, timeNow(), timeNow()); err != nil {
-		return fmt.Errorf("cannot update clock timestamp: %v", err)
-	}
-	return nil
-}
-
 func writeTimings(st *state.State, rootdir, fromMode string) error {
 	changeKind := "install-system"
 	logPath := filepath.Join(rootdir, "var/log/install-timings.txt.gz")
@@ -260,30 +191,6 @@ func writeTimings(st *state.State, rootdir, fromMode string) error {
 	return nil
 }
 
-// buildInstallObserver creates an observer for gadget assets if
-// applicable, otherwise the returned gadget.ContentObserver is nil.
-// The observer if any is also returned as non-nil trustedObserver if
-// encryption is in use.
-func buildInstallObserver(model *asserts.Model, gadgetDir string, useEncryption bool) (
-	observer gadget.ContentObserver, trustedObserver *boot.TrustedAssetsInstallObserver, err error) {
-
-	// observer will be a nil interface by default
-	trustedObserver, err = boot.TrustedAssetsInstallObserverForModel(model, gadgetDir, useEncryption)
-	if err != nil && err != boot.ErrObserverNotApplicable {
-		return nil, nil, fmt.Errorf("cannot setup asset install observer: %v", err)
-	}
-	if err == nil {
-		observer = trustedObserver
-		if !useEncryption {
-			// there will be no key sealing, so past the
-			// installation pass no other methods need to be called
-			trustedObserver = nil
-		}
-	}
-
-	return observer, trustedObserver, nil
-}
-
 func (m *DeviceManager) doSetupUbuntuSave(t *state.Task, _ *tomb.Tomb) error {
 	st := t.State()
 	st.Lock()
@@ -358,7 +265,7 @@ func (m *DeviceManager) doSetupRunSystem(t *state.Task, _ *tomb.Tomb) error {
 		return fmt.Errorf("cannot use gadget: %v", err)
 	}
 
-	installObserver, trustedInstallObserver, err := buildInstallObserver(model, gadgetDir, useEncryption)
+	installObserver, trustedInstallObserver, err := installLogic.BuildInstallObserver(model, gadgetDir, useEncryption)
 	if err != nil {
 		return err
 	}
@@ -376,12 +283,12 @@ func (m *DeviceManager) doSetupRunSystem(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	if trustedInstallObserver != nil {
-		if err := prepareEncryptedSystemData(model, installedSystem.KeyForRole, trustedInstallObserver); err != nil {
+		if err := installLogic.PrepareEncryptedSystemData(model, installedSystem.KeyForRole, trustedInstallObserver); err != nil {
 			return err
 		}
 	}
 
-	if err := prepareRunSystemData(model, gadgetDir, perfTimings); err != nil {
+	if err := installLogicPrepareRunSystemData(model, gadgetDir, perfTimings); err != nil {
 		return err
 	}
 
@@ -413,137 +320,6 @@ func (m *DeviceManager) doSetupRunSystem(t *state.Task, _ *tomb.Tomb) error {
 	return nil
 }
 
-func prepareEncryptedSystemData(model *asserts.Model, keyForRole map[string]keys.EncryptionKey, trustedInstallObserver *boot.TrustedAssetsInstallObserver) error {
-	// validity check
-	if len(keyForRole) == 0 || keyForRole[gadget.SystemData] == nil || keyForRole[gadget.SystemSave] == nil {
-		return fmt.Errorf("internal error: system encryption keys are unset")
-	}
-	dataEncryptionKey := keyForRole[gadget.SystemData]
-	saveEncryptionKey := keyForRole[gadget.SystemSave]
-
-	// make note of the encryption keys
-	trustedInstallObserver.ChosenEncryptionKeys(dataEncryptionKey, saveEncryptionKey)
-
-	// keep track of recovery assets
-	if err := trustedInstallObserver.ObserveExistingTrustedRecoveryAssets(boot.InitramfsUbuntuSeedDir); err != nil {
-		return fmt.Errorf("cannot observe existing trusted recovery assets: %v", err)
-	}
-	if err := saveKeys(model, keyForRole); err != nil {
-		return err
-	}
-	// write markers containing a secret to pair data and save
-	if err := writeMarkers(model); err != nil {
-		return err
-	}
-	return nil
-}
-
-func prepareRunSystemData(model *asserts.Model, gadgetDir string, perfTimings timings.Measurer) error {
-	// keep track of the model we installed
-	err := os.MkdirAll(filepath.Join(boot.InitramfsUbuntuBootDir, "device"), 0755)
-	if err != nil {
-		return fmt.Errorf("cannot store the model: %v", err)
-	}
-	err = writeModel(model, filepath.Join(boot.InitramfsUbuntuBootDir, "device/model"))
-	if err != nil {
-		return fmt.Errorf("cannot store the model: %v", err)
-	}
-
-	// preserve systemd-timesyncd clock timestamp, so that RTC-less devices
-	// can start with a more recent time on the next boot
-	if err := writeTimesyncdClock(dirs.GlobalRootDir, boot.InstallHostWritableDir(model)); err != nil {
-		return fmt.Errorf("cannot seed timesyncd clock: %v", err)
-	}
-
-	// configure the run system
-	opts := &sysconfig.Options{TargetRootDir: boot.InstallHostWritableDir(model), GadgetDir: gadgetDir}
-	// configure cloud init
-	setSysconfigCloudOptions(opts, gadgetDir, model)
-	timings.Run(perfTimings, "sysconfig-configure-target-system", "Configure target system", func(timings.Measurer) {
-		err = sysconfigConfigureTargetSystem(model, opts)
-	})
-	if err != nil {
-		return err
-	}
-
-	// TODO: FIXME: this should go away after we have time to design a proper
-	//              solution
-
-	if !model.Classic() {
-		// on some specific devices, we need to create these directories in
-		// _writable_defaults in order to allow the install-device hook to install
-		// some files there, this eventually will go away when we introduce a proper
-		// mechanism not using system-files to install files onto the root
-		// filesystem from the install-device hook
-		if err := fixupWritableDefaultDirs(boot.InstallHostWritableDir(model)); err != nil {
-			return err
-		}
-	}
-
-	return nil
-}
-
-func fixupWritableDefaultDirs(systemDataDir string) error {
-	// the _writable_default directory is used to put files in place on
-	// ubuntu-data from install mode, so we abuse it here for a specific device
-	// to let that device install files with system-files and the install-device
-	// hook
-
-	// eventually this will be a proper, supported, designed mechanism instead
-	// of just this hack, but this hack is just creating the directories, since
-	// the system-files interface only allows creating the file, not creating
-	// the directories leading up to that file, and since the file is deeply
-	// nested we would effectively have to give all permission to the device
-	// to create any file on ubuntu-data which we don't want to do, so we keep
-	// this restriction to let the device create one specific file, and then
-	// we behind the scenes just create the directories for the device
-
-	for _, subDirToCreate := range []string{"/etc/udev/rules.d", "/etc/modprobe.d", "/etc/modules-load.d/", "/etc/systemd/network"} {
-		dirToCreate := sysconfig.WritableDefaultsDir(systemDataDir, subDirToCreate)
-
-		if err := os.MkdirAll(dirToCreate, 0755); err != nil {
-			return err
-		}
-	}
-
-	return nil
-}
-
-// writeMarkers writes markers containing the same secret to pair data and save.
-func writeMarkers(model *asserts.Model) error {
-	// ensure directory for markers exists
-	if err := os.MkdirAll(boot.InstallHostFDEDataDir(model), 0755); err != nil {
-		return err
-	}
-	if err := os.MkdirAll(boot.InstallHostFDESaveDir, 0755); err != nil {
-		return err
-	}
-
-	// generate a secret random marker
-	markerSecret, err := randutil.CryptoTokenBytes(32)
-	if err != nil {
-		return fmt.Errorf("cannot create ubuntu-data/save marker secret: %v", err)
-	}
-
-	return device.WriteEncryptionMarkers(boot.InstallHostFDEDataDir(model), boot.InstallHostFDESaveDir, markerSecret)
-}
-
-func saveKeys(model *asserts.Model, keyForRole map[string]keys.EncryptionKey) error {
-	saveEncryptionKey := keyForRole[gadget.SystemSave]
-	if saveEncryptionKey == nil {
-		// no system-save support
-		return nil
-	}
-	// ensure directory for keys exists
-	if err := os.MkdirAll(boot.InstallHostFDEDataDir(model), 0755); err != nil {
-		return err
-	}
-	if err := saveEncryptionKey.Save(device.SaveKeyUnder(boot.InstallHostFDEDataDir(model))); err != nil {
-		return fmt.Errorf("cannot store system save key: %v", err)
-	}
-	return nil
-}
-
 // RebootOptions can be attached to restart-system-to-run-mode tasks to control
 // their restart behavior.
 type RebootOptions struct {
@@ -578,7 +354,7 @@ func (m *DeviceManager) doRestartSystemToRunMode(t *state.Task, _ *tomb.Tomb) er
 	}
 	model := deviceCtx.Model()
 
-	preseeded, err := maybeApplyPreseededData(st, boot.InitramfsUbuntuSeedDir, modeEnv.RecoverySystem, boot.InstallHostWritableDir(model))
+	preseeded, err := maybeApplyPreseededData(model, boot.InitramfsUbuntuSeedDir, modeEnv.RecoverySystem, boot.InstallHostWritableDir(model))
 	if err != nil {
 		logger.Noticef("failed to apply preseed data: %v", err)
 		return err
@@ -642,221 +418,37 @@ func (m *DeviceManager) doRestartSystemToRunMode(t *state.Task, _ *tomb.Tomb) er
 	return nil
 }
 
-func readPreseedAssertion(st *state.State, model *asserts.Model, ubuntuSeedDir, sysLabel string) (*asserts.Preseed, error) {
-	f, err := os.Open(filepath.Join(ubuntuSeedDir, "systems", sysLabel, "preseed"))
-	if err != nil {
-		return nil, fmt.Errorf("cannot read preseed assertion: %v", err)
-	}
-
-	// main seed assertions are loaded in the assertion db of install mode; add preseed assertion from
-	// systems/<label>/preseed file on top of it via a temporary db.
-	tmpDb := assertstate.TemporaryDB(st)
-	batch := asserts.NewBatch(nil)
-	_, err = batch.AddStream(f)
-	if err != nil {
-		return nil, err
-	}
-
-	var preseedAs *asserts.Preseed
-	err = batch.CommitToAndObserve(tmpDb, func(as asserts.Assertion) {
-		if as.Type() == asserts.PreseedType {
-			preseedAs = as.(*asserts.Preseed)
-		}
-	}, nil)
-	if err != nil {
-		return nil, err
-	}
-
-	switch {
-	case preseedAs == nil:
-		return nil, fmt.Errorf("internal error: preseed assertion file is present but preseed assertion not found")
-	case preseedAs.SystemLabel() != sysLabel:
-		return nil, fmt.Errorf("preseed assertion system label %q doesn't match system label %q", preseedAs.SystemLabel(), sysLabel)
-	case preseedAs.Model() != model.Model():
-		return nil, fmt.Errorf("preseed assertion model %q doesn't match the model %q", preseedAs.Model(), model.Model())
-	case preseedAs.BrandID() != model.BrandID():
-		return nil, fmt.Errorf("preseed assertion brand %q doesn't match model brand %q", preseedAs.BrandID(), model.BrandID())
-	case preseedAs.Series() != model.Series():
-		return nil, fmt.Errorf("preseed assertion series %q doesn't match model series %q", preseedAs.Series(), model.Series())
-	}
-
-	return preseedAs, nil
-}
-
 var seedOpen = seed.Open
 
-// TODO: consider reusing this kind of handler for UC20 seeding
-type preseedSnapHandler struct {
-	writableDir string
-}
-
-func (p *preseedSnapHandler) HandleUnassertedSnap(name, path string, _ timings.Measurer) (string, error) {
-	pinfo := snap.MinimalPlaceInfo(name, snap.Revision{N: -1})
-	targetPath := filepath.Join(p.writableDir, pinfo.MountFile())
-	mountDir := filepath.Join(p.writableDir, pinfo.MountDir())
-
-	sq := squashfs.New(path)
-	opts := &snap.InstallOptions{MustNotCrossDevices: true}
-	if _, err := sq.Install(targetPath, mountDir, opts); err != nil {
-		return "", fmt.Errorf("cannot install snap %q: %v", name, err)
-	}
-
-	return targetPath, nil
-}
-
-func (p *preseedSnapHandler) HandleAndDigestAssertedSnap(name, path string, essType snap.Type, snapRev *asserts.SnapRevision, _ func(string, uint64) (snap.Revision, error), _ timings.Measurer) (string, string, uint64, error) {
-	pinfo := snap.MinimalPlaceInfo(name, snap.Revision{N: snapRev.SnapRevision()})
-	targetPath := filepath.Join(p.writableDir, pinfo.MountFile())
-	mountDir := filepath.Join(p.writableDir, pinfo.MountDir())
-
-	logger.Debugf("copying: %q to %q; mount dir=%q", path, targetPath, mountDir)
-
-	srcFile, err := os.Open(path)
+func maybeApplyPreseededData(model *asserts.Model, ubuntuSeedDir, sysLabel, writableDir string) (preseeded bool, err error) {
+	sysSeed, err := seedOpen(ubuntuSeedDir, sysLabel)
 	if err != nil {
-		return "", "", 0, err
+		return false, err
 	}
-	defer srcFile.Close()
+	// this function is for UC20+ only so sysSeed ia always PreseedCapable
+	preseedSeed := sysSeed.(seed.PreseedCapable)
 
-	destFile, err := osutil.NewAtomicFile(targetPath, 0644, 0, osutil.NoChown, osutil.NoChown)
-	if err != nil {
-		return "", "", 0, fmt.Errorf("cannot create atomic file: %v", err)
-	}
-	defer destFile.Cancel()
-
-	finfo, err := srcFile.Stat()
-	if err != nil {
-		return "", "", 0, err
-	}
-
-	destFile.SetModTime(finfo.ModTime())
-
-	h := crypto.SHA3_384.New()
-	w := io.MultiWriter(h, destFile)
-
-	size, err := io.CopyBuffer(w, srcFile, make([]byte, 2*1024*1024))
-	if err != nil {
-		return "", "", 0, err
-	}
-	if err := destFile.Commit(); err != nil {
-		return "", "", 0, fmt.Errorf("cannot copy snap %q: %v", name, err)
-	}
-
-	sq := squashfs.New(targetPath)
-	opts := &snap.InstallOptions{MustNotCrossDevices: true}
-	// since Install target path is the same as source path passed to squashfs.New,
-	// Install isn't going to copy the blob, but we call it to set up mount directory etc.
-	if _, err := sq.Install(targetPath, mountDir, opts); err != nil {
-		return "", "", 0, fmt.Errorf("cannot install snap %q: %v", name, err)
-	}
-
-	sha3_384, err := asserts.EncodeDigest(crypto.SHA3_384, h.Sum(nil))
-	if err != nil {
-		return "", "", 0, fmt.Errorf("cannot encode snap %q digest: %v", path, err)
-	}
-	return targetPath, sha3_384, uint64(size), nil
-}
-
-var maybeApplyPreseededData = func(st *state.State, ubuntuSeedDir, sysLabel, writableDir string) (preseeded bool, err error) {
-	preseedArtifact := filepath.Join(ubuntuSeedDir, "systems", sysLabel, "preseed.tgz")
-	if !osutil.FileExists(preseedArtifact) {
+	if !preseedSeed.HasArtifact("preseed.tgz") {
 		return false, nil
 	}
 
-	model, err := findModel(st)
-	if err != nil {
-		return false, fmt.Errorf("preseed error: cannot find model: %v", err)
-	}
-
-	preseedAs, err := readPreseedAssertion(st, model, ubuntuSeedDir, sysLabel)
-	if err != nil {
+	if err := preseedSeed.LoadAssertions(nil, nil); err != nil {
 		return false, err
 	}
-
-	// TODO: consider a writer that feeds the file to stdin of tar and calculates the digest at the same time.
-	sha3_384, _, err := osutil.FileDigest(preseedArtifact, crypto.SHA3_384)
-	if err != nil {
-		return false, fmt.Errorf("cannot calculate preseed artifact digest: %v", err)
+	_, sig := model.Signature()
+	_, seedModelSig := preseedSeed.Model().Signature()
+	if !bytes.Equal(sig, seedModelSig) {
+		return false, fmt.Errorf("system seed %q model does not match model in use", sysLabel)
 	}
 
-	digest, err := base64.RawURLEncoding.DecodeString(preseedAs.ArtifactSHA3_384())
-	if err != nil {
-		return false, fmt.Errorf("cannot decode preseed artifact digest")
-	}
-	if !bytes.Equal(sha3_384, digest) {
-		return false, fmt.Errorf("invalid preseed artifact digest")
-	}
-
-	logger.Noticef("apply preseed data: %q, %q", writableDir, preseedArtifact)
-	cmd := exec.Command("tar", "--extract", "--preserve-permissions", "--preserve-order", "--gunzip", "--directory", writableDir, "-f", preseedArtifact)
-	if err := cmd.Run(); err != nil {
+	if err := applyPreseededData(preseedSeed, writableDir); err != nil {
 		return false, err
 	}
-
-	logger.Noticef("copying snaps")
-
-	deviceSeed, err := seedOpen(ubuntuSeedDir, sysLabel)
-	if err != nil {
-		return false, err
-	}
-	tm := timings.New(nil)
-
-	if err := deviceSeed.LoadAssertions(nil, nil); err != nil {
-		return false, err
-	}
-
-	if err := os.MkdirAll(filepath.Join(writableDir, "var/lib/snapd/snaps"), 0755); err != nil {
-		return false, err
-	}
-
-	snapHandler := &preseedSnapHandler{writableDir: writableDir}
-	if err := deviceSeed.LoadMeta("run", snapHandler, tm); err != nil {
-		return false, err
-	}
-
-	preseedSnaps := make(map[string]*asserts.PreseedSnap)
-	for _, ps := range preseedAs.Snaps() {
-		preseedSnaps[ps.Name] = ps
-	}
-
-	checkSnap := func(ssnap *seed.Snap) error {
-		ps, ok := preseedSnaps[ssnap.SnapName()]
-		if !ok {
-			return fmt.Errorf("snap %q not present in the preseed assertion", ssnap.SnapName())
-		}
-		if ps.Revision != ssnap.SideInfo.Revision.N {
-			rev := snap.Revision{N: ps.Revision}
-			return fmt.Errorf("snap %q has wrong revision %s (expected: %s)", ssnap.SnapName(), ssnap.SideInfo.Revision, rev)
-		}
-		if ps.SnapID != ssnap.SideInfo.SnapID {
-			return fmt.Errorf("snap %q has wrong snap id %q (expected: %q)", ssnap.SnapName(), ssnap.SideInfo.SnapID, ps.SnapID)
-		}
-		return nil
-	}
-
-	esnaps := deviceSeed.EssentialSnaps()
-	msnaps, err := deviceSeed.ModeSnaps("run")
-	if err != nil {
-		return false, err
-	}
-	if len(msnaps)+len(esnaps) != len(preseedSnaps) {
-		return false, fmt.Errorf("seed has %d snaps but %d snaps are required by preseed assertion", len(msnaps)+len(esnaps), len(preseedSnaps))
-	}
-
-	for _, esnap := range esnaps {
-		if err := checkSnap(esnap); err != nil {
-			return false, err
-		}
-	}
-
-	for _, ssnap := range msnaps {
-		if err := checkSnap(ssnap); err != nil {
-			return false, err
-		}
-	}
-
 	return true, nil
 }
 
+var applyPreseededData = installLogic.ApplyPreseededData
+
 func (m *DeviceManager) doFactoryResetRunSystem(t *state.Task, _ *tomb.Tomb) error {
 	st := t.State()
 	st.Lock()
@@ -997,12 +589,12 @@ func (m *DeviceManager) doFactoryResetRunSystem(t *state.Task, _ *tomb.Tomb) err
 		// keep track of the new ubuntu-save encryption key
 		installedSystem.KeyForRole[gadget.SystemSave] = saveEncryptionKey
 
-		if err := prepareEncryptedSystemData(model, installedSystem.KeyForRole, trustedInstallObserver); err != nil {
+		if err := installLogic.PrepareEncryptedSystemData(model, installedSystem.KeyForRole, trustedInstallObserver); err != nil {
 			return err
 		}
 	}
 
-	if err := prepareRunSystemData(model, gadgetDir, perfTimings); err != nil {
+	if err := installLogicPrepareRunSystemData(model, gadgetDir, perfTimings); err != nil {
 		return err
 	}
 
@@ -1225,7 +817,7 @@ func verifyFactoryResetMarkerInRun(marker string, hasEncryption bool) error {
 }
 
 func rotateEncryptionKeys() error {
-	kd, err := ioutil.ReadFile(filepath.Join(dirs.SnapFDEDir, "ubuntu-save.key"))
+	kd, err := os.ReadFile(filepath.Join(dirs.SnapFDEDir, "ubuntu-save.key"))
 	if err != nil {
 		return fmt.Errorf("cannot open encryption key file: %v", err)
 	}
@@ -1260,17 +852,16 @@ func mountSeedSnap(seedSn *seed.Snap) (mountpoint string, unmount func() error,
 		nil
 }
 
-func (m *DeviceManager) loadAndMountSystemLabelSnaps(systemLabel string) (
-	*System, map[snap.Type]*snap.Info, map[snap.Type]*seed.Snap, map[snap.Type]string, func(), error) {
+func (m *DeviceManager) loadAndMountSystemLabelSnaps(systemLabel string) (*systemAndEssentialSnaps, map[snap.Type]string, func(), error) {
 
 	essentialTypes := []snap.Type{snap.TypeKernel, snap.TypeBase, snap.TypeGadget}
-	sys, snapInfos, snapSeeds, err := m.loadSystemAndEssentialSnaps(systemLabel, essentialTypes)
+	systemAndSnaps, err := m.loadSystemAndEssentialSnaps(systemLabel, essentialTypes)
 	if err != nil {
-		return nil, nil, nil, nil, nil, err
+		return nil, nil, nil, err
 	}
 	// Unset revision here actually means that the snap is local.
 	// Assign then a local revision as seeding/installing the snap would do.
-	for _, snInfo := range snapInfos {
+	for _, snInfo := range systemAndSnaps.InfosByType {
 		if snInfo.Revision.Unset() {
 			snInfo.Revision = snap.R(-1)
 		}
@@ -1286,21 +877,25 @@ func (m *DeviceManager) loadAndMountSystemLabelSnaps(systemLabel string) (
 			}
 		}
 	}
-	for _, seedSn := range []*seed.Snap{snapSeeds[snap.TypeGadget], snapSeeds[snap.TypeKernel]} {
+
+	seedSnaps := systemAndSnaps.SeedSnapsByType
+
+	for _, seedSn := range []*seed.Snap{seedSnaps[snap.TypeGadget], seedSnaps[snap.TypeKernel]} {
 		mntPt, unmountSnap, err := mountSeedSnap(seedSn)
 		if err != nil {
 			unmount()
-			return nil, nil, nil, nil, nil, err
+			return nil, nil, nil, err
 		}
 		unmountFuncs = append(unmountFuncs, unmountSnap)
 		mntPtForType[seedSn.EssentialType] = mntPt
 	}
 
-	return sys, snapInfos, snapSeeds, mntPtForType, unmount, nil
+	return systemAndSnaps, mntPtForType, unmount, nil
 }
 
 // doInstallFinish performs the finish step of the install. It will
 // - install missing volumes structure content
+// - copy seed (only for UC)
 // - install gadget assets
 // - install kernel.efi
 // - make system bootable (including writing modeenv)
@@ -1334,17 +929,15 @@ func (m *DeviceManager) doInstallFinish(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	st.Unlock()
-	sys, snapInfos, snapSeeds, mntPtForType, unmount, err := m.loadAndMountSystemLabelSnaps(systemLabel)
+	systemAndSnaps, mntPtForType, unmount, err := m.loadAndMountSystemLabelSnaps(systemLabel)
 	st.Lock()
 	if err != nil {
 		return err
 	}
 	defer unmount()
 
-	// TODO validation of onVolumes versus gadget.yaml
-
 	// Check if encryption is mandatory
-	if sys.Model.StorageSafety() == asserts.StorageSafetyEncrypted && encryptSetupData == nil {
+	if systemAndSnaps.Model.StorageSafety() == asserts.StorageSafetyEncrypted && encryptSetupData == nil {
 		return fmt.Errorf("storage encryption required by model but has not been set up")
 	}
 	useEncryption := encryptSetupData != nil
@@ -1352,7 +945,40 @@ func (m *DeviceManager) doInstallFinish(t *state.Task, _ *tomb.Tomb) error {
 	logger.Debugf("starting install-finish for %q (using encryption: %t) on %v", systemLabel, useEncryption, onVolumes)
 
 	// TODO we probably want to pass a different location for the assets cache
-	installObserver, trustedInstallObserver, err := buildInstallObserver(sys.Model, mntPtForType[snap.TypeGadget], useEncryption)
+	installObserver, trustedInstallObserver, err := installLogic.BuildInstallObserver(systemAndSnaps.Model, mntPtForType[snap.TypeGadget], useEncryption)
+	if err != nil {
+		return err
+	}
+
+	gi, err := gadget.ReadInfoAndValidate(mntPtForType[snap.TypeGadget], systemAndSnaps.Model, nil)
+	if err != nil {
+		return err
+	}
+
+	// Import new information from the installer to the gadget data,
+	// including the target devices and information marked as partial in
+	// the gadget, so the gadget is not partially defined anymore if it
+	// was.
+	// TODO validation of onVolumes versus gadget.yaml, needs to happen here.
+	mergedVols, err := gadget.ApplyInstallerVolumesToGadget(onVolumes, gi.Volumes)
+	if err != nil {
+		return err
+	}
+
+	// Match gadget against the disk, so we make sure that the information
+	// reported by the installer is correct and that all partitions have
+	// been created.
+	volCompatOpts := &gadget.VolumeCompatibilityOptions{
+		// at this point all partitions should be created
+		AssumeCreatablePartitionsCreated: true,
+	}
+	if useEncryption {
+		volCompatOpts.ExpectedStructureEncryption = map[string]gadget.StructureEncryptionParameters{
+			"ubuntu-data": {Method: gadget.EncryptionLUKS},
+			"ubuntu-save": {Method: gadget.EncryptionLUKS},
+		}
+	}
+	volToGadgetToDiskStruct, err := installMatchDisksToGadgetVolumes(mergedVols, volCompatOpts)
 	if err != nil {
 		return err
 	}
@@ -1362,9 +988,9 @@ func (m *DeviceManager) doInstallFinish(t *state.Task, _ *tomb.Tomb) error {
 	if useEncryption {
 		encType = secboot.EncryptionTypeLUKS
 	}
-	// TODO for partial gadgets we should also use the data from onVolumes instead of
-	// using only what comes from gadget.yaml.
-	_, allLaidOutVols, err := gadget.LaidOutVolumesFromGadget(mntPtForType[snap.TypeGadget], mntPtForType[snap.TypeKernel], sys.Model, encType)
+	allLaidOutVols, err := gadget.LaidOutVolumesFromGadget(mergedVols,
+		mntPtForType[snap.TypeGadget], mntPtForType[snap.TypeKernel],
+		encType, volToGadgetToDiskStruct)
 	if err != nil {
 		return fmt.Errorf("on finish install: cannot layout volumes: %v", err)
 	}
@@ -1373,31 +999,46 @@ func (m *DeviceManager) doInstallFinish(t *state.Task, _ *tomb.Tomb) error {
 	timings.Run(perfTimings, "install-content", "Writing content to partitions", func(tm timings.Measurer) {
 		st.Unlock()
 		defer st.Lock()
-		_, err = installWriteContent(onVolumes, allLaidOutVols, encryptSetupData, installObserver, perfTimings)
+		_, err = installWriteContent(mergedVols, allLaidOutVols, encryptSetupData, installObserver, perfTimings)
 	})
 	if err != nil {
 		return fmt.Errorf("cannot write content: %v", err)
 	}
 
 	// Mount the partitions and find the system-seed{,-null} partition
-	seedMntDir, unmountParts, err := installMountVolumes(onVolumes, encryptSetupData)
+	seedMntDir, unmountParts, err := installMountVolumes(mergedVols, encryptSetupData)
 	if err != nil {
 		return fmt.Errorf("cannot mount partitions for installation: %v", err)
 	}
 	defer unmountParts()
 
-	if err := installSaveStorageTraits(sys.Model, allLaidOutVols, encryptSetupData); err != nil {
+	if !systemAndSnaps.Model.Classic() {
+		copier, ok := systemAndSnaps.Seed.(seed.Copier)
+		if !ok {
+			return fmt.Errorf("internal error: seed does not support copying: %s", systemAndSnaps.Label)
+		}
+
+		logger.Debugf("copying label %q to seed partition", systemAndSnaps.Label)
+		if err := copier.Copy(seedMntDir, systemAndSnaps.Label, perfTimings); err != nil {
+			return fmt.Errorf("cannot copy seed: %w", err)
+		}
+	}
+
+	if err := installSaveStorageTraits(systemAndSnaps.Model, mergedVols, encryptSetupData); err != nil {
 		return err
 	}
 
 	if useEncryption {
 		if trustedInstallObserver != nil {
-			if err := prepareEncryptedSystemData(sys.Model, install.KeysForRole(encryptSetupData), trustedInstallObserver); err != nil {
+			if err := installLogic.PrepareEncryptedSystemData(systemAndSnaps.Model, install.KeysForRole(encryptSetupData), trustedInstallObserver); err != nil {
 				return err
 			}
 		}
 	}
 
+	snapInfos := systemAndSnaps.InfosByType
+	snapSeeds := systemAndSnaps.SeedSnapsByType
+
 	bootWith := &boot.BootableSet{
 		Base:              snapInfos[snap.TypeBase],
 		BasePath:          snapSeeds[snap.TypeBase].Path,
@@ -1422,13 +1063,13 @@ func (m *DeviceManager) doInstallFinish(t *state.Task, _ *tomb.Tomb) error {
 		return err
 	}
 
-	// writes the model
-	if err := prepareRunSystemData(sys.Model, bootWith.UnpackedGadgetDir, perfTimings); err != nil {
+	// writes the model etc
+	if err := installLogicPrepareRunSystemData(systemAndSnaps.Model, bootWith.UnpackedGadgetDir, perfTimings); err != nil {
 		return err
 	}
 
 	logger.Debugf("making the installed system runnable for system label %s", systemLabel)
-	if err := bootMakeRunnableStandalone(sys.Model, bootWith, trustedInstallObserver); err != nil {
+	if err := bootMakeRunnableStandalone(systemAndSnaps.Model, bootWith, trustedInstallObserver, st.Unlocker()); err != nil {
 		return err
 	}
 
@@ -1454,7 +1095,7 @@ func (m *DeviceManager) doInstallSetupStorageEncryption(t *state.Task, _ *tomb.T
 	logger.Debugf("install-setup-storage-encryption for %q on %v", systemLabel, onVolumes)
 
 	st.Unlock()
-	sys, snapInfos, snapSeeds, mntPtForType, unmount, err := m.loadAndMountSystemLabelSnaps(systemLabel)
+	systemAndSeeds, mntPtForType, unmount, err := m.loadAndMountSystemLabelSnaps(systemLabel)
 	st.Lock()
 	if err != nil {
 		return err
@@ -1462,16 +1103,16 @@ func (m *DeviceManager) doInstallSetupStorageEncryption(t *state.Task, _ *tomb.T
 	defer unmount()
 
 	// Gadget information
-	snapf, err := snapfile.Open(snapSeeds[snap.TypeGadget].Path)
+	snapf, err := snapfile.Open(systemAndSeeds.SeedSnapsByType[snap.TypeGadget].Path)
 	if err != nil {
 		return fmt.Errorf("cannot open gadget snap: %v", err)
 	}
-	gadgetInfo, err := gadget.ReadInfoFromSnapFileNoValidate(snapf, sys.Model)
+	gadgetInfo, err := gadget.ReadInfoFromSnapFileNoValidate(snapf, systemAndSeeds.Model)
 	if err != nil {
 		return fmt.Errorf("reading gadget information: %v", err)
 	}
 
-	encryptInfo, err := m.encryptionSupportInfo(sys.Model, secboot.TPMProvisionFull, snapInfos[snap.TypeKernel], gadgetInfo)
+	encryptInfo, err := m.encryptionSupportInfo(systemAndSeeds.Model, secboot.TPMProvisionFull, systemAndSeeds.InfosByType[snap.TypeKernel], gadgetInfo)
 	if err != nil {
 		return err
 	}
@@ -1487,7 +1128,7 @@ func (m *DeviceManager) doInstallSetupStorageEncryption(t *state.Task, _ *tomb.T
 
 	// TODO:ICE: support secboot.EncryptionTypeLUKSWithICE in the API
 	encType := secboot.EncryptionTypeLUKS
-	encryptionSetupData, err := installEncryptPartitions(onVolumes, encType, sys.Model, mntPtForType[snap.TypeGadget], mntPtForType[snap.TypeKernel], perfTimings)
+	encryptionSetupData, err := installEncryptPartitions(onVolumes, encType, systemAndSeeds.Model, mntPtForType[snap.TypeGadget], mntPtForType[snap.TypeKernel], perfTimings)
 	if err != nil {
 		return err
 	}
