@@ -22,7 +22,6 @@ package devicestate_test
 import (
 	"bytes"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"time"
@@ -60,7 +59,7 @@ type remodelLogicBaseSuite struct {
 	brands       *assertstest.SigningAccounts
 
 	capturedDevBE storecontext.DeviceBackend
-	dummyStore    snapstate.StoreService
+	testStore     snapstate.StoreService
 }
 
 func (s *remodelLogicBaseSuite) SetUpTest(c *C) {
@@ -91,11 +90,11 @@ func (s *remodelLogicBaseSuite) SetUpTest(c *C) {
 		assertstatetest.AddMany(s.state, s.brands.AccountsAndKeys("my-brand")...)
 	}()
 
-	s.dummyStore = new(storetest.Store)
+	s.testStore = new(storetest.Store)
 
 	newStore := func(devBE storecontext.DeviceBackend) snapstate.StoreService {
 		s.capturedDevBE = devBE
-		return s.dummyStore
+		return s.testStore
 	}
 
 	hookMgr, err := hookstate.Manager(s.state, o.TaskRunner())
@@ -455,10 +454,10 @@ func (s *remodelLogicSuite) TestNewStoreRemodelContextStore(c *C) {
 	})
 
 	sto := remodCtx.Store()
-	c.Check(sto, Equals, s.dummyStore)
+	c.Check(sto, Equals, s.testStore)
 
 	// store is kept and not rebuilt
-	s.dummyStore = nil
+	s.testStore = nil
 
 	sto1 := remodCtx.Store()
 	c.Check(sto1, Equals, sto)
@@ -720,19 +719,19 @@ func (s *remodelLogicSuite) TestRemodelContextForTaskNo(c *C) {
 
 	// task is nil
 	remodCtx1, err := devicestate.DeviceCtx(s.state, nil, nil)
-	c.Check(err, Equals, state.ErrNoState)
+	c.Check(err, testutil.ErrorIs, state.ErrNoState)
 	c.Check(remodCtx1, IsNil)
 
 	// no change
 	t := s.state.NewTask("random-task", "...")
 	_, err = devicestate.DeviceCtx(s.state, t, nil)
-	c.Check(err, Equals, state.ErrNoState)
+	c.Check(err, testutil.ErrorIs, state.ErrNoState)
 
 	// not a remodel change
 	chg := s.state.NewChange("not-remodel", "...")
 	chg.AddTask(t)
 	_, err = devicestate.DeviceCtx(s.state, t, nil)
-	c.Check(err, Equals, state.ErrNoState)
+	c.Check(err, testutil.ErrorIs, state.ErrNoState)
 }
 
 func (s *remodelLogicSuite) setupForRereg(c *C) (oldModel, newModel *asserts.Model) {
@@ -896,13 +895,13 @@ func (s *remodelLogicSuite) TestReregRemodelContextNewSerial(c *C) {
 
 	// no new serial yet
 	_, err = devBE.Serial()
-	c.Assert(err, Equals, state.ErrNoState)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
 
 	chg := s.state.NewChange("remodel", "...")
 
 	remodCtx.Init(chg)
 
-	// sanity check
+	// validity check
 	device1, err := devBE.Device()
 	c.Assert(err, IsNil)
 	c.Check(device1, DeepEquals, &auth.DeviceState{
@@ -912,13 +911,13 @@ func (s *remodelLogicSuite) TestReregRemodelContextNewSerial(c *C) {
 
 	// still no new serial
 	_, err = devBE.Serial()
-	c.Assert(err, Equals, state.ErrNoState)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
 
 	newSerial := makeSerialAssertionInState(c, s.brands, s.state, "my-brand", "other-model", "serialserialserial2")
 
 	// same
 	_, err = devBE.Serial()
-	c.Check(err, Equals, state.ErrNoState)
+	c.Check(err, testutil.ErrorIs, state.ErrNoState)
 
 	// finish registration
 	regCtx := remodCtx.(devicestate.RegistrationContext)
@@ -960,7 +959,7 @@ func writeDeviceModelToUbuntuBoot(c *C, model *asserts.Model) {
 	var buf bytes.Buffer
 	c.Assert(asserts.NewEncoder(&buf).Encode(model), IsNil)
 	c.Assert(os.MkdirAll(filepath.Join(boot.InitramfsUbuntuBootDir, "device"), 0755), IsNil)
-	c.Assert(ioutil.WriteFile(filepath.Join(boot.InitramfsUbuntuBootDir, "device/model"),
+	c.Assert(os.WriteFile(filepath.Join(boot.InitramfsUbuntuBootDir, "device/model"),
 		buf.Bytes(), 0755),
 		IsNil)
 }
@@ -988,7 +987,7 @@ func (s *uc20RemodelLogicSuite) SetUpTest(c *C) {
 	err := m.WriteTo("")
 	c.Assert(err, IsNil)
 
-	restore := boot.MockResealKeyToModeenv(func(_ string, m *boot.Modeenv, expectReseal bool) error {
+	restore := boot.MockResealKeyToModeenv(func(_ string, m *boot.Modeenv, expectReseal bool, _ boot.Unlocker) error {
 		return fmt.Errorf("not expected to be called")
 	})
 	s.AddCleanup(restore)
@@ -1073,7 +1072,7 @@ func (s *uc20RemodelLogicSuite) TestReregRemodelContextUC20(c *C) {
 
 	remodCtx.Init(chg)
 
-	// sanity check
+	// validity check
 	device1, err := devBE.Device()
 	c.Assert(err, IsNil)
 	c.Check(device1, DeepEquals, &auth.DeviceState{
@@ -1089,7 +1088,7 @@ func (s *uc20RemodelLogicSuite) TestReregRemodelContextUC20(c *C) {
 	c.Assert(err, IsNil)
 
 	resealKeysCalls := 0
-	restore := boot.MockResealKeyToModeenv(func(_ string, m *boot.Modeenv, expectReseal bool) error {
+	restore := boot.MockResealKeyToModeenv(func(_ string, m *boot.Modeenv, expectReseal bool, u boot.Unlocker) error {
 		resealKeysCalls++
 		c.Check(m.CurrentRecoverySystems, DeepEquals, []string{"0000", "1234"})
 		c.Check(m.GoodRecoverySystems, DeepEquals, []string{"0000", "1234"})
@@ -1107,6 +1106,8 @@ func (s *uc20RemodelLogicSuite) TestReregRemodelContextUC20(c *C) {
 		default:
 			c.Fatalf("unexpected call #%v to reseal key to modeenv", resealKeysCalls)
 		}
+		// check unlocker
+		u()()
 		// this is running as part of post finish step, so the state has
 		// already been updated
 		serial, err = s.mgr.Serial()
@@ -1197,7 +1198,7 @@ func (s *uc20RemodelLogicSuite) TestUpdateRemodelContext(c *C) {
 	c.Assert(chg.Get("new-model", &encNewModel), IsNil)
 
 	resealKeysCalls := 0
-	restore := boot.MockResealKeyToModeenv(func(_ string, m *boot.Modeenv, expectReseal bool) error {
+	restore := boot.MockResealKeyToModeenv(func(_ string, m *boot.Modeenv, expectReseal bool, u boot.Unlocker) error {
 		resealKeysCalls++
 		c.Check(m.CurrentRecoverySystems, DeepEquals, []string{"0000", "1234"})
 		c.Check(m.GoodRecoverySystems, DeepEquals, []string{"0000", "1234"})
@@ -1217,6 +1218,8 @@ func (s *uc20RemodelLogicSuite) TestUpdateRemodelContext(c *C) {
 		default:
 			c.Fatalf("unexpected call #%v to reseal key to modeenv", resealKeysCalls)
 		}
+		// check unlocker
+		u()()
 		return nil
 	})
 	s.AddCleanup(restore)
@@ -1296,7 +1299,7 @@ func (s *uc20RemodelLogicSuite) TestSimpleRemodelErr(c *C) {
 	c.Assert(chg.Get("new-model", &encNewModel), IsNil)
 
 	resealKeysCalls := 0
-	restore := boot.MockResealKeyToModeenv(func(_ string, m *boot.Modeenv, expectReseal bool) error {
+	restore := boot.MockResealKeyToModeenv(func(_ string, m *boot.Modeenv, expectReseal bool, u boot.Unlocker) error {
 		resealKeysCalls++
 		c.Check(m.CurrentRecoverySystems, DeepEquals, []string{"0000", "1234"})
 		c.Check(m.GoodRecoverySystems, DeepEquals, []string{"0000", "1234"})
@@ -1311,6 +1314,8 @@ func (s *uc20RemodelLogicSuite) TestSimpleRemodelErr(c *C) {
 		default:
 			c.Fatalf("unexpected call #%v to reseal key to modeenv", resealKeysCalls)
 		}
+		// check unlocker
+		u()()
 		return nil
 	})
 	s.AddCleanup(restore)
