@@ -551,7 +551,9 @@ func MockConnectRetryTimeout(d time.Duration) (restore func()) {
 // snapstate.LinkSnapParticipant follow activation changes for snaps
 // so that we can track revisions with security profiles on disk for
 // temporarily inactive snaps.
-func OnSnapLinkageChanged(st *state.State, instanceName string) error {
+func OnSnapLinkageChanged(st *state.State, snapsup *snapstate.SnapSetup) error {
+	instanceName := snapsup.InstanceName()
+
 	var snapst snapstate.SnapState
 	if err := snapstate.Get(st, instanceName, &snapst); err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
