@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2017-2022 Canonical Ltd
+ * Copyright (C) 2017-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -20,12 +20,15 @@
 package snapstate
 
 import (
+	"bytes"
 	"context"
 	"errors"
 	"fmt"
 	"os"
+	"strings"
 	"time"
 
+	"github.com/snapcore/snapd/features"
 	"github.com/snapcore/snapd/httputil"
 	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/logger"
@@ -80,7 +83,9 @@ var refreshRetryDelay = 20 * time.Minute
 // of auto-refresh.
 type refreshCandidate struct {
 	SnapSetup
-	Version string `json:"version,omitempty"`
+	// Monitored signals whether this snap is currently being monitored for closure
+	// so its auto-refresh can be continued.
+	Monitored bool `json:"monitored,omitempty"`
 }
 
 func (rc *refreshCandidate) Type() snap.Type {
@@ -99,16 +104,23 @@ func (rc *refreshCandidate) InstanceName() string {
 	return rc.SnapSetup.InstanceName()
 }
 
-func (rc *refreshCandidate) Prereq(st *state.State) []string {
+func (rc *refreshCandidate) Prereq(*state.State, PrereqTracker) []string {
 	return rc.SnapSetup.Prereq
 }
 
-func (rc *refreshCandidate) SnapSetupForUpdate(st *state.State, params updateParamsFunc, userID int, globalFlags *Flags) (*SnapSetup, *SnapState, error) {
+func (rc *refreshCandidate) SnapSetupForUpdate(st *state.State, _ updateParamsFunc, _ int, globalFlags *Flags, _ PrereqTracker) (*SnapSetup, *SnapState, error) {
 	var snapst SnapState
 	if err := Get(st, rc.InstanceName(), &snapst); err != nil {
 		return nil, nil, err
 	}
-	return &rc.SnapSetup, &snapst, nil
+
+	snapsup := &rc.SnapSetup
+	if globalFlags != nil {
+		snapsup.Flags.IsAutoRefresh = globalFlags.IsAutoRefresh
+		snapsup.Flags.IsContinuedAutoRefresh = globalFlags.IsContinuedAutoRefresh
+	}
+
+	return snapsup, &snapst, nil
 }
 
 // soundness check
@@ -123,6 +135,8 @@ type autoRefresh struct {
 	nextRefresh         time.Time
 	lastRefreshAttempt  time.Time
 	managedDeniedLogged bool
+
+	restoredMonitoring bool
 }
 
 func newAutoRefresh(st *state.State) *autoRefresh {
@@ -152,9 +166,13 @@ func (m *autoRefresh) LastRefresh() (time.Time, error) {
 // EffectiveRefreshHold returns the time until to which refreshes are
 // held if refresh.hold configuration is set.
 func (m *autoRefresh) EffectiveRefreshHold() (time.Time, error) {
+	return effectiveRefreshHold(m.state)
+}
+
+func effectiveRefreshHold(st *state.State) (time.Time, error) {
 	var holdValue string
 
-	tr := config.NewTransaction(m.state)
+	tr := config.NewTransaction(st)
 	err := tr.Get("core", "refresh.hold", &holdValue)
 	if err != nil && !config.IsNoOption(err) {
 		return time.Time{}, err
@@ -210,16 +228,18 @@ func (m *autoRefresh) clearRefreshHold() {
 func (m *autoRefresh) AtSeed() error {
 	// on classic hold refreshes for 2h after seeding
 	if release.OnClassic {
-		var t1 time.Time
-		tr := config.NewTransaction(m.state)
-		err := tr.Get("core", "refresh.hold", &t1)
-		if !config.IsNoOption(err) {
-			// already set or error
+		holdTime, err := effectiveRefreshHold(m.state)
+		if err != nil {
 			return err
 		}
+		if !holdTime.IsZero() {
+			// already set
+			return nil
+		}
 		// TODO: have a policy that if the snapd exe itself
 		// is older than X weeks/months we skip the holding?
 		now := time.Now().UTC()
+		tr := config.NewTransaction(m.state)
 		tr.Set("core", "refresh.hold", now.Add(2*time.Hour))
 		tr.Commit()
 		m.nextRefresh = now
@@ -265,11 +285,31 @@ func (m *autoRefresh) canRefreshRespectingMetered(now, lastRefresh time.Time) (c
 	return false, nil
 }
 
+func isStoreOnline(s *state.State) (bool, error) {
+	tr := config.NewTransaction(s)
+
+	var access string
+	if err := tr.GetMaybe("core", "store.access", &access); err != nil {
+		return false, err
+	}
+
+	return access != "offline", nil
+}
+
 // Ensure ensures that we refresh all installed snaps periodically
 func (m *autoRefresh) Ensure() (err error) {
 	m.state.Lock()
 	defer m.state.Unlock()
 
+	online, err := isStoreOnline(m.state)
+	if err != nil || !online {
+		return err
+	}
+
+	if err := m.restoreMonitoring(); err != nil {
+		return fmt.Errorf("cannot restore monitoring: %v", err)
+	}
+
 	// see if it even makes sense to try to refresh
 	if CanAutoRefresh == nil {
 		return nil
@@ -278,27 +318,6 @@ func (m *autoRefresh) Ensure() (err error) {
 		return err
 	}
 
-	// is there a previously partially inhibited auto-refresh that can now be continued?
-	if attempt, ok := canContinueAutoRefresh(m.state); ok {
-		// override the auto-refresh delay if we're continuing an inhibited auto-refresh
-		// for the first time (because the snap just closed after we notified the user)
-		overrideDelay := attempt == 1
-		err := m.launchAutoRefresh(overrideDelay)
-		if err != nil {
-			if errors.Is(err, tooSoonError{}) {
-				// ignore error, retry the auto-refresh later
-				return nil
-			}
-
-			// we didn't auto-refresh, so keep flag but increase attempt counter
-			m.state.Cache("auto-refresh-continue-attempt", attempt+1)
-			return err
-		}
-		// clear the continue flag if the auto-refresh was scheduled successfully
-		m.state.Cache("auto-refresh-continue-attempt", nil)
-		return nil
-	}
-
 	// get lastRefresh and schedule
 	lastRefresh, err := m.LastRefresh()
 	if err != nil {
@@ -381,8 +400,7 @@ func (m *autoRefresh) Ensure() (err error) {
 				return nil
 			}
 
-			overrideDelay := false
-			err = m.launchAutoRefresh(overrideDelay)
+			err = m.launchAutoRefresh()
 			if _, ok := err.(*httputil.PersistentNetworkError); ok {
 				// refresh will be retried after refreshRetryDelay
 				return err
@@ -399,12 +417,48 @@ func (m *autoRefresh) Ensure() (err error) {
 	return err
 }
 
-func canContinueAutoRefresh(st *state.State) (int, bool) {
-	if cachedAttempt := st.Cached("auto-refresh-continue-attempt"); cachedAttempt != nil {
-		return cachedAttempt.(int), true
+func (m *autoRefresh) restoreMonitoring() error {
+	if m.restoredMonitoring {
+		return nil
 	}
 
-	return 0, false
+	// clear the old monitoring state; remove in a later snapd version
+	m.state.Set("monitored-snaps", nil)
+
+	var refreshHints map[string]*refreshCandidate
+	if err := m.state.Get("refresh-candidates", &refreshHints); err != nil && !errors.Is(err, &state.NoStateError{}) {
+		return fmt.Errorf("cannot get refresh-candidates: %v", err)
+	}
+
+	defer func() { m.restoredMonitoring = true }()
+
+	var monitored []*SnapSetup
+	for _, hint := range refreshHints {
+		if hint.Monitored {
+			monitored = append(monitored, &hint.SnapSetup)
+		}
+	}
+
+	if len(monitored) == 0 {
+		return nil
+	}
+
+	aborts := make(map[string]context.CancelFunc, len(monitored))
+	for _, snap := range monitored {
+		done := make(chan string, 1)
+		snapName := snap.InstanceName()
+		if err := cgroupMonitorSnapEnded(snapName, done); err != nil {
+			logger.Noticef("cannot restore monitoring for snap %q closure: %v", snapName, err)
+			continue
+		}
+
+		refreshCtx, abort := context.WithCancel(context.Background())
+		aborts[snapName] = abort
+		go continueRefreshOnSnapClose(m.state, snap.InstanceName(), done, refreshCtx)
+	}
+
+	m.state.Cache("monitored-snaps", aborts)
+	return nil
 }
 
 // isRefreshHeld returns whether an auto-refresh is currently held back or not,
@@ -545,13 +599,13 @@ func (tooSoonError) Is(err error) bool {
 }
 
 // launchAutoRefresh creates the auto-refresh taskset and a change for it.
-func (m *autoRefresh) launchAutoRefresh(overrideDelay bool) error {
+func (m *autoRefresh) launchAutoRefresh() error {
 	// Check that we have reasonable delays between attempts.
 	// If the store is under stress we need to make sure we do not
 	// hammer it too often
 	now := timeNow()
 	minAttempt := m.lastRefreshAttempt.Add(refreshRetryDelay)
-	if !overrideDelay && !m.lastRefreshAttempt.IsZero() && minAttempt.After(now) {
+	if !m.lastRefreshAttempt.IsZero() && minAttempt.After(now) {
 		return tooSoonError{}
 	}
 	m.lastRefreshAttempt = now
@@ -595,10 +649,19 @@ func (m *autoRefresh) launchAutoRefresh(overrideDelay bool) error {
 		return err
 	}
 
-	createPreDownloadChange(m.state, updateTss)
+	if _, err = createPreDownloadChange(m.state, updateTss); err != nil {
+		return err
+	}
 
 	if len(updateTss.Refresh) == 0 {
-		return nil
+		// NOTE: If all refresh candidates are blocked from auto-refresh by checks
+		// in softCheckNothingRunningForRefresh then no auto-refresh change will be
+		// created (i.e. len(updateTss.Refresh) == 0) and only a pre-download change
+		// is created for those snaps. This still means that the set of inhibited
+		// snaps could have changed so we are recording a notice about it here
+		// because it cannot be captured in processInhibitedAutoRefresh which only
+		// looks for auto-refresh changes.
+		return maybeAddRefreshInhibitNotice(m.state)
 	}
 
 	msg := autoRefreshSummary(updated)
@@ -620,20 +683,31 @@ func (m *autoRefresh) launchAutoRefresh(overrideDelay bool) error {
 
 // createPreDownloadChange creates a pre-download change if any relevant tasksets
 // exist in the UpdateTaskSets and returns whether or not a change was created.
-func createPreDownloadChange(st *state.State, updateTss *UpdateTaskSets) bool {
+func createPreDownloadChange(st *state.State, updateTss *UpdateTaskSets) (bool, error) {
 	if updateTss != nil && len(updateTss.PreDownload) > 0 {
-		preDlChg := st.NewChange("pre-download", i18n.G("Pre-download tasks for auto-refresh"))
+		var snapNames []string
+		for _, ts := range updateTss.PreDownload {
+			task := ts.Tasks()[0]
+			var snapsup *SnapSetup
+			if err := task.Get("snap-setup", &snapsup); err != nil {
+				return false, err
+			}
+			snapNames = append(snapNames, snapsup.InstanceName())
+		}
+
+		chgSummary := fmt.Sprintf(i18n.G("Pre-download %s for auto-refresh"), strutil.Quoted(snapNames))
+		preDlChg := st.NewChange("pre-download", chgSummary)
 		for _, ts := range updateTss.PreDownload {
 			preDlChg.AddAll(ts)
 		}
-		return true
+		return true, nil
 	}
-	return false
+	return false, nil
 }
 
 func autoRefreshInFlight(st *state.State) bool {
 	for _, chg := range st.Changes() {
-		if chg.Kind() == "auto-refresh" && !chg.Status().Ready() {
+		if chg.Kind() == "auto-refresh" && !chg.IsReady() {
 			return true
 		}
 	}
@@ -654,14 +728,45 @@ func getTime(st *state.State, timeKey string) (time.Time, error) {
 //
 // This allows the, possibly slow, communication with each snapd session agent,
 // to be performed without holding the snap state lock.
-var asyncPendingRefreshNotification = func(context context.Context, client *userclient.Client, refreshInfo *userclient.PendingSnapRefreshInfo) {
+var asyncPendingRefreshNotification = func(ctx context.Context, refreshInfo *userclient.PendingSnapRefreshInfo) {
+	logger.Debugf("notifying agents about pending refresh for snap %q", refreshInfo.InstanceName)
+
 	go func() {
-		if err := client.PendingRefreshNotification(context, refreshInfo); err != nil {
+		client := userclient.New()
+		if err := client.PendingRefreshNotification(ctx, refreshInfo); err != nil {
 			logger.Noticef("Cannot send notification about pending refresh: %v", err)
 		}
 	}()
 }
 
+// maybeAsyncPendingRefreshNotification broadcasts desktop notification in a goroutine.
+//
+// The notification is sent only if no snap has the marker "snap-refresh-observe"
+// interface connected and the "refresh-app-awareness-ux" experimental flag is disabled.
+func maybeAsyncPendingRefreshNotification(ctx context.Context, st *state.State, refreshInfo *userclient.PendingSnapRefreshInfo) {
+	tr := config.NewTransaction(st)
+	experimentalRefreshAppAwarenessUX, err := features.Flag(tr, features.RefreshAppAwarenessUX)
+	if err != nil && !config.IsNoOption(err) {
+		logger.Noticef("Cannot send notification about pending refresh: %v", err)
+		return
+	}
+	if experimentalRefreshAppAwarenessUX {
+		// use notices + warnings fallback flow instead
+		return
+	}
+
+	markerExists, err := HasActiveConnection(st, "snap-refresh-observe")
+	if err != nil {
+		logger.Noticef("Cannot send notification about pending refresh: %v", err)
+		return
+	}
+	if markerExists {
+		// found snap with marker interface, skip notification
+		return
+	}
+	asyncPendingRefreshNotification(ctx, refreshInfo)
+}
+
 type timedBusySnapError struct {
 	err           *BusySnapError
 	timeRemaining time.Duration
@@ -682,15 +787,16 @@ func (e *timedBusySnapError) Is(err error) bool {
 	return ok
 }
 
-// inhibitRefresh returns an error if refresh is inhibited by running apps.
+// inhibitRefresh returns whether a refresh is forced due to inhibition
+// timeout or an error if refresh is inhibited by running apps.
 //
 // Internally the snap state is updated to remember when the inhibition first
 // took place. Apps can inhibit refreshes for up to "maxInhibition", beyond
 // that period the refresh will go ahead despite application activity.
-func inhibitRefresh(st *state.State, snapst *SnapState, snapsup *SnapSetup, info *snap.Info) error {
+func inhibitRefresh(st *state.State, snapst *SnapState, snapsup *SnapSetup, info *snap.Info) (inhibitionTimeout bool, err error) {
 	checkerErr := refreshAppsCheck(info)
 	if checkerErr == nil {
-		return nil
+		return false, nil
 	}
 
 	// carries the remaining inhibition time along with the BusySnapError
@@ -699,7 +805,7 @@ func inhibitRefresh(st *state.State, snapst *SnapState, snapsup *SnapSetup, info
 	// if it's not a snap busy error or the refresh is manual, surface the error
 	// to the user instead of notifying or delaying the refresh
 	if !snapsup.IsAutoRefresh || !errors.As(checkerErr, &busyErr.err) {
-		return checkerErr
+		return false, checkerErr
 	}
 
 	// Decide on what to do depending on the state of the snap and the remaining
@@ -721,23 +827,159 @@ func inhibitRefresh(st *state.State, snapst *SnapState, snapsup *SnapSetup, info
 		// increasing frequency, allowing the user to understand the urgency.
 		busyErr.timeRemaining = (maxInhibition - now.Sub(*snapst.RefreshInhibitedTime)).Truncate(time.Second)
 	default:
+		// XXX: should we drop this notification?
 		// if the refresh inhibition window has ended, notify the user that the
 		// refresh is happening now and ignore the error
 		refreshInfo := busyErr.PendingSnapRefreshInfo()
-		asyncPendingRefreshNotification(context.TODO(), userclient.New(), refreshInfo)
+		maybeAsyncPendingRefreshNotification(context.TODO(), st, refreshInfo)
 		// important to return "nil" type here instead of
 		// setting busyErr to nil as otherwise we return a nil
 		// interface which is not the nil type
+		return true, nil
+	}
+
+	return false, busyErr
+}
+
+// IsSnapMonitored checks if there's already a goroutine waiting for this snap to close.
+func IsSnapMonitored(st *state.State, snapName string) bool {
+	return monitoringAbort(st, snapName) != nil
+}
+
+func processInhibitedAutoRefresh(chg *state.Change, old state.Status, new state.Status) {
+	if chg.Kind() != "auto-refresh" || !new.Ready() {
+		return
+	}
+
+	if err := maybeAddRefreshInhibitNotice(chg.State()); err != nil {
+		logger.Debugf(`internal error: failed to add "refresh-inhibit" notice: %v`, err)
+	}
+}
+
+// maybeAddRefreshInhibitNotice records a refresh-inhibit notice if the set of
+// inhibited snaps was changed since the last notice.
+func maybeAddRefreshInhibitNotice(st *state.State) error {
+	var lastRecordedInhibitedSnaps map[string]bool
+	if err := st.Get("last-recorded-inhibited-snaps", &lastRecordedInhibitedSnaps); err != nil && !errors.Is(err, state.ErrNoState) {
+		return err
+	}
+
+	snapStates, err := All(st)
+	if err != nil {
+		return err
+	}
+
+	curInhibitedSnaps := make(map[string]bool, len(lastRecordedInhibitedSnaps))
+	for _, snapst := range snapStates {
+		if snapst.RefreshInhibitedTime == nil {
+			continue
+		}
+		curInhibitedSnaps[snapst.InstanceName()] = true
+	}
+
+	changed := len(lastRecordedInhibitedSnaps) != len(curInhibitedSnaps)
+	if !changed {
+		for snapName := range curInhibitedSnaps {
+			if !lastRecordedInhibitedSnaps[snapName] {
+				changed = true
+				break
+			}
+		}
+	}
+
+	if changed {
+		if _, err := st.AddNotice(nil, state.RefreshInhibitNotice, "-", nil); err != nil {
+			return err
+		}
+		st.Set("last-recorded-inhibited-snaps", curInhibitedSnaps)
+	}
+
+	if err := maybeAddRefreshInhibitWarningFallback(st, curInhibitedSnaps); err != nil {
+		logger.Noticef("Cannot add refresh inhibition warning: %v", err)
+	}
+
+	return nil
+}
+
+// maybeAddRefreshInhibitWarningFallback records a warning if the set of
+// inhibited snaps was changed since the last notice.
+//
+// The warning is recorded only if:
+// 	1. There is at least 1 inhibited snap.
+// 	2. The "refresh-app-awareness-ux" experimental flag is enabled.
+// 	3. No snap exists with the marker "snap-refresh-observe" interface connected.
+//
+// Note: If no snaps are inhibited then existing inhibition warning
+// will be removed.
+func maybeAddRefreshInhibitWarningFallback(st *state.State, inhibitedSnaps map[string]bool) error {
+	if len(inhibitedSnaps) == 0 {
+		// no more inhibited snaps, remove inhibition warning if it exists.
+		return removeRefreshInhibitWarning(st)
+	}
+
+	tr := config.NewTransaction(st)
+	experimentalRefreshAppAwarenessUX, err := features.Flag(tr, features.RefreshAppAwarenessUX)
+	if err != nil && !config.IsNoOption(err) {
+		return err
+	}
+	if !experimentalRefreshAppAwarenessUX {
+		// snapd will send notifications directly, check maybeAsyncPendingRefreshNotification
 		return nil
 	}
 
-	return busyErr
+	markerExists, err := HasActiveConnection(st, "snap-refresh-observe")
+	if err != nil {
+		return err
+	}
+	if markerExists {
+		// do nothing
+		return nil
+	}
+
+	// let's fallback to issuing warnings if no snap exists with the
+	// marker snap-refresh-observe interface connected.
+
+	// remove inhibition warning if it exists.
+	if err := removeRefreshInhibitWarning(st); err != nil {
+		return err
+	}
+
+	// building warning message
+	var snapsBuf bytes.Buffer
+	i := 0
+	for snap := range inhibitedSnaps {
+		if i > 0 {
+			snapsBuf.WriteString(", ")
+		}
+		snapsBuf.WriteString(snap)
+		i++
+	}
+	message := fmt.Sprintf("cannot refresh (%s) due running apps; close running apps to continue refresh.", snapsBuf.String())
+
+	// wait some time before showing the same warning to the user again after okaying.
+	st.AddWarning(message, &state.AddWarningOptions{RepeatAfter: 24 * time.Hour})
+
+	return nil
+}
+
+// removeRefreshInhibitWarning removes inhibition warning if it exists.
+func removeRefreshInhibitWarning(st *state.State) error {
+	// XXX: is it worth it to check for unexpected multiple matches?
+	for _, warning := range st.AllWarnings() {
+		if !strings.HasSuffix(warning.String(), "close running apps to continue refresh.") {
+			continue
+		}
+		if err := st.RemoveWarning(warning.String()); err != nil && !errors.Is(err, state.ErrNoState) {
+			return err
+		}
+		return nil
+	}
+	return nil
 }
 
 // for testing outside of snapstate
-func MockRefreshCandidate(snapSetup *SnapSetup, version string) interface{} {
+func MockRefreshCandidate(snapSetup *SnapSetup) interface{} {
 	return &refreshCandidate{
 		SnapSetup: *snapSetup,
-		Version:   version,
 	}
 }
