@@ -71,6 +71,7 @@ type fakeOp struct {
 
 	otherInstances         bool
 	unlinkFirstInstallUndo bool
+	unlinkSkipBinaries     bool
 	skipKernelExtraction   bool
 
 	services         []string
@@ -84,6 +85,12 @@ type fakeOp struct {
 
 	dirOpts  *dirs.SnapDirOptions
 	undoInfo *backend.UndoInfo
+
+	compsToInstall, currentComps []*snap.ComponentSideInfo
+	compsToRemove, finalComps    []*snap.ComponentSideInfo
+
+	containerName     string
+	containerFileName string
 }
 
 type fakeOps []fakeOp
@@ -243,6 +250,8 @@ func (f *fakeStore) snap(spec snapSpec) (*snap.Info, error) {
 		confinement = "classic"
 	case "some-epoch-snap":
 		epoch = snap.E("42")
+	case "firmware-updater":
+		snapID = "EI0D1KHjP8XiwMZKqSjuh6W8zvcowUVP"
 	case "snapd-desktop-integration":
 		snapID = "IrwRHakqtzhFRHJOOPxKVPU0Kk7Erhcu"
 	}
@@ -259,7 +268,7 @@ func (f *fakeStore) snap(spec snapSpec) (*snap.Info, error) {
 			SnapID:   snapID,
 			Revision: spec.Revision,
 		},
-		Version: spec.Name,
+		Version: spec.Name + "Ver",
 		DownloadInfo: snap.DownloadInfo{
 			DownloadURL: "https://some-server.com/some/path.snap",
 			Size:        5,
@@ -367,6 +376,9 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 		name = "some-snap-now-classic"
 	case "some-snap-was-classic-id":
 		name = "some-snap-was-classic"
+	case "some-snap-with-new-base-id":
+		name = "some-snap-with-new-base"
+		base = "core22"
 	case "core-snap-id":
 		name = "core"
 		typ = snap.TypeOS
@@ -393,7 +405,7 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 		name = "snap-content-plug"
 	case "snap-content-slot-id":
 		name = "snap-content-slot"
-	case "snapd-snap-id":
+	case "snapd-snap-id", "snapd-without-version-id":
 		name = "snapd"
 		typ = snap.TypeSnapd
 	case "kernel-id":
@@ -423,6 +435,9 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 	// for validation-sets testing
 	case "bgtKhntON3vR7kwEbVPsILm7bUViPDzx":
 		name = "some-other-snap"
+	case "some-base-snap-id":
+		name = "some-base-snap"
+		base = "some-base"
 	case "provenance-snap-id":
 		name = "provenance-snap"
 	default:
@@ -454,7 +469,7 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 			SnapID:   cand.snapID,
 			Revision: revno,
 		},
-		Version: name,
+		Version: name + "Ver",
 		DownloadInfo: snap.DownloadInfo{
 			DownloadURL: "https://some-server.com/some/path.snap",
 		},
@@ -464,6 +479,10 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 		Base:          base,
 	}
 
+	if strings.HasSuffix(cand.snapID, "-without-version-id") {
+		info.Version = ""
+	}
+
 	if name == "outdated-consumer" {
 		info.Plugs = map[string]*snap.PlugInfo{
 			"content-plug": {
@@ -575,18 +594,22 @@ func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.Curren
 
 	refreshErrors := make(map[string]error)
 	installErrors := make(map[string]error)
+	downloadErrors := make(map[string]error)
 	var res []store.SnapActionResult
 	for _, a := range sorted {
-		if a.Action != "install" && a.Action != "refresh" {
+		switch a.Action {
+		case "install", "refresh", "download":
+		default:
 			panic("not supported")
 		}
+
 		if a.InstanceName == "" {
 			return nil, nil, fmt.Errorf("internal error: action without instance name")
 		}
 
 		snapName, instanceKey := snap.SplitInstanceName(a.InstanceName)
 
-		if a.Action == "install" {
+		if a.Action == "install" || a.Action == "download" {
 			spec := snapSpec{
 				Name:     snapName,
 				Channel:  a.Channel,
@@ -595,7 +618,11 @@ func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.Curren
 			}
 			info, err := f.snap(spec)
 			if err != nil {
-				installErrors[a.InstanceName] = err
+				if a.Action == "install" {
+					installErrors[a.InstanceName] = err
+				} else {
+					downloadErrors[a.InstanceName] = err
+				}
 				continue
 			}
 			f.fakeBackend.appendOp(&fakeOp{
@@ -668,17 +695,21 @@ func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.Curren
 		res = append(res, store.SnapActionResult{Info: info})
 	}
 
-	if len(refreshErrors)+len(installErrors) > 0 || len(res) == 0 {
+	if len(refreshErrors)+len(installErrors)+len(downloadErrors) > 0 || len(res) == 0 {
 		if len(refreshErrors) == 0 {
 			refreshErrors = nil
 		}
 		if len(installErrors) == 0 {
 			installErrors = nil
 		}
+		if len(downloadErrors) == 0 {
+			downloadErrors = nil
+		}
 		return res, nil, &store.SnapActionError{
-			NoResults: len(refreshErrors)+len(installErrors)+len(res) == 0,
+			NoResults: len(refreshErrors)+len(installErrors)+len(downloadErrors)+len(res) == 0,
 			Refresh:   refreshErrors,
 			Install:   installErrors,
+			Download:  downloadErrors,
 		}
 	}
 
@@ -706,7 +737,7 @@ func (f *fakeStore) Download(ctx context.Context, name, targetFn string, snapInf
 		macaroon = user.StoreMacaroon
 	}
 	// only add the options if they contain anything interesting
-	if *dlOpts == (store.DownloadOptions{}) {
+	if dlOpts != nil && *dlOpts == (store.DownloadOptions{}) {
 		dlOpts = nil
 	}
 	f.downloads = append(f.downloads, fakeDownload{
@@ -884,6 +915,83 @@ func (f *fakeSnappyBackend) SetupSnap(snapFilePath, instanceName string, si *sna
 	return snapType, &backend.InstallRecord{}, nil
 }
 
+func (f *fakeSnappyBackend) SetupKernelSnap(instanceName string, rev snap.Revision, meter progress.Meter) (err error) {
+	meter.Notify("prepare-kernel-snap")
+	f.appendOp(&fakeOp{
+		op: "prepare-kernel-snap",
+	})
+	return nil
+}
+
+func (f *fakeSnappyBackend) RemoveKernelSnapSetup(instanceName string, rev snap.Revision, meter progress.Meter) error {
+	meter.Notify("remove-kernel-snap-setup")
+	f.appendOp(&fakeOp{
+		op: "remove-kernel-snap-setup",
+	})
+	return nil
+}
+
+func (f *fakeSnappyBackend) SetupComponent(compFilePath string, compPi snap.ContainerPlaceInfo, dev snap.Device, meter progress.Meter) (installRecord *backend.InstallRecord, err error) {
+	meter.Notify("setup-component")
+	f.appendOp(&fakeOp{
+		op:                "setup-component",
+		containerName:     compPi.ContainerName(),
+		containerFileName: compPi.Filename(),
+	})
+	if strings.HasSuffix(compPi.ContainerName(), "+broken") {
+		return nil, fmt.Errorf("cannot set-up component %q", compPi.ContainerName())
+	}
+	return &backend.InstallRecord{}, nil
+}
+
+func (f *fakeSnappyBackend) SetupKernelModulesComponents(compsToInstall, currentComps []*snap.ComponentSideInfo, ksnapName string, ksnapRev snap.Revision, meter progress.Meter) (err error) {
+	meter.Notify("setup-kernel-modules-components")
+	f.appendOp(&fakeOp{
+		op:             "setup-kernel-modules-components",
+		compsToInstall: compsToInstall,
+		currentComps:   currentComps,
+	})
+	if strings.HasSuffix(ksnapName, "+broken") {
+		return fmt.Errorf("cannot set-up kernel-modules for %s", ksnapName)
+	}
+	return nil
+}
+
+func (f *fakeSnappyBackend) RemoveKernelModulesComponentsSetup(compsToRemove, finalComps []*snap.ComponentSideInfo, ksnapName string, ksnapRev snap.Revision, meter progress.Meter) (err error) {
+	meter.Notify("remove-kernel-modules-components-setup")
+	f.appendOp(&fakeOp{
+		op:            "remove-kernel-modules-components-setup",
+		compsToRemove: compsToRemove,
+		finalComps:    finalComps,
+	})
+	if strings.HasSuffix(ksnapName, "+reverterr") {
+		return fmt.Errorf("cannot remove set-up of kernel-modules for %s", ksnapName)
+	}
+	return nil
+}
+
+func (f *fakeSnappyBackend) UndoSetupComponent(cpi snap.ContainerPlaceInfo, installRecord *backend.InstallRecord, dev snap.Device, meter progress.Meter) error {
+	meter.Notify("undo-setup-component")
+	f.appendOp(&fakeOp{
+		op:                "undo-setup-component",
+		containerName:     cpi.ContainerName(),
+		containerFileName: cpi.Filename(),
+	})
+	if strings.HasSuffix(cpi.ContainerName(), "+brokenundo") {
+		return fmt.Errorf("cannot undo set-up of component %q", cpi.ContainerName())
+	}
+	return nil
+}
+
+func (f *fakeSnappyBackend) RemoveComponentDir(cpi snap.ContainerPlaceInfo) error {
+	f.appendOp(&fakeOp{
+		op:                "remove-component-dir",
+		containerName:     cpi.ContainerName(),
+		containerFileName: cpi.Filename(),
+	})
+	return nil
+}
+
 func (f *fakeSnappyBackend) ReadInfo(name string, si *snap.SideInfo) (*snap.Info, error) {
 	if name == "borken" && si.Revision == snap.R(2) {
 		return nil, errors.New(`cannot read info for "borken" snap`)
@@ -898,6 +1006,7 @@ func (f *fakeSnappyBackend) ReadInfo(name string, si *snap.SideInfo) (*snap.Info
 	// naive emulation for now, always works
 	info := &snap.Info{
 		SuggestedName: snapName,
+		Version:       snapName + "Ver",
 		SideInfo:      *si,
 		Architectures: []string{"all"},
 		SnapType:      snap.TypeApp,
@@ -1065,6 +1174,14 @@ func (f *fakeSnappyBackend) LinkSnap(info *snap.Info, dev snap.Device, linkCtx b
 	return boot.RebootInfo{RebootRequired: reboot}, nil
 }
 
+func (f *fakeSnappyBackend) LinkComponent(cpi snap.ContainerPlaceInfo, snapRev snap.Revision) error {
+	f.appendOp(&fakeOp{
+		op:   "link-component",
+		path: cpi.MountDir(),
+	})
+	return f.maybeErrForLastOp()
+}
+
 func svcSnapMountDir(svcs []*snap.AppInfo) string {
 	if len(svcs) == 0 {
 		return "<no services>"
@@ -1101,37 +1218,27 @@ func (f *fakeSnappyBackend) StopServices(svcs []*snap.AppInfo, reason snap.Servi
 	return f.maybeErrForLastOp()
 }
 
-func (f *fakeSnappyBackend) ServicesEnableState(info *snap.Info, meter progress.Meter) (map[string]bool, error) {
+func (f *fakeSnappyBackend) QueryDisabledServices(info *snap.Info, meter progress.Meter) ([]string, error) {
+	var l []string
+
 	// return the disabled services as disabled and nothing else
 	m := make(map[string]bool)
 	for _, svc := range f.servicesCurrentlyDisabled {
 		m[svc] = false
 	}
 
+	for name, enabled := range m {
+		if !enabled {
+			l = append(l, name)
+		}
+	}
+
 	f.appendOp(&fakeOp{
 		op:               "current-snap-service-states",
 		disabledServices: f.servicesCurrentlyDisabled,
 	})
 
-	return m, f.maybeErrForLastOp()
-}
-
-func (f *fakeSnappyBackend) QueryDisabledServices(info *snap.Info, meter progress.Meter) ([]string, error) {
-	var l []string
-
-	m, err := f.ServicesEnableState(info, meter)
-	if err != nil {
-		return nil, err
-	}
-	for name, enabled := range m {
-		if !enabled {
-			l = append(l, name)
-		}
-	}
-
-	// XXX: add a fakeOp here?
-
-	return l, nil
+	return l, f.maybeErrForLastOp()
 }
 
 func (f *fakeSnappyBackend) UndoSetupSnap(s snap.PlaceInfo, typ snap.Type, installRecord *backend.InstallRecord, dev snap.Device, p progress.Meter) error {
@@ -1182,6 +1289,15 @@ func (f *fakeSnappyBackend) UnlinkSnap(info *snap.Info, linkCtx backend.LinkCont
 		path: info.MountDir(),
 
 		unlinkFirstInstallUndo: linkCtx.FirstInstall,
+		unlinkSkipBinaries:     linkCtx.SkipBinaries,
+	})
+	return f.maybeErrForLastOp()
+}
+
+func (f *fakeSnappyBackend) UnlinkComponent(cpi snap.ContainerPlaceInfo, snapRev snap.Revision) error {
+	f.appendOp(&fakeOp{
+		op:   "unlink-component",
+		path: cpi.MountDir(),
 	})
 	return f.maybeErrForLastOp()
 }
@@ -1220,7 +1336,7 @@ func (f *fakeSnappyBackend) RemoveSnapSaveData(info *snap.Info, _ snap.Device) e
 	return f.maybeErrForLastOp()
 }
 
-func (f *fakeSnappyBackend) RemoveSnapDataDir(info *snap.Info, otherInstances bool) error {
+func (f *fakeSnappyBackend) RemoveSnapDataDir(info *snap.Info, otherInstances bool, opts *dirs.SnapDirOptions) error {
 	f.ops = append(f.ops, fakeOp{
 		op:             "remove-snap-data-dir",
 		name:           info.InstanceName(),
@@ -1230,10 +1346,10 @@ func (f *fakeSnappyBackend) RemoveSnapDataDir(info *snap.Info, otherInstances bo
 	return f.maybeErrForLastOp()
 }
 
-func (f *fakeSnappyBackend) RemoveSnapMountUnits(s snap.PlaceInfo, meter progress.Meter) error {
+func (f *fakeSnappyBackend) RemoveContainerMountUnits(s snap.ContainerPlaceInfo, meter progress.Meter) error {
 	f.ops = append(f.ops, fakeOp{
 		op:   "remove-snap-mount-units",
-		name: s.InstanceName(),
+		name: s.ContainerName(),
 	})
 	return f.maybeErrForLastOp()
 }
@@ -1256,6 +1372,13 @@ func (f *fakeSnappyBackend) DiscardSnapNamespace(snapName string) error {
 	return f.maybeErrForLastOp()
 }
 
+func (f *fakeSnappyBackend) RemoveAllSnapAppArmorProfiles() error {
+	f.appendOp(&fakeOp{
+		op: "remove-apparmor-profiles",
+	})
+	return f.maybeErrForLastOp()
+}
+
 func (f *fakeSnappyBackend) RemoveSnapInhibitLock(snapName string) error {
 	f.appendOp(&fakeOp{
 		op:   "remove-inhibit-lock",
