@@ -23,6 +23,7 @@ package seedwriter
 import (
 	"errors"
 	"fmt"
+	"sort"
 	"strings"
 
 	"github.com/snapcore/snapd/asserts"
@@ -46,6 +47,23 @@ type Options struct {
 	// TestSkipCopyUnverifiedModel is set to support naive tests
 	// using an unverified model, the resulting image is broken
 	TestSkipCopyUnverifiedModel bool
+
+	// Manifest is used to track snaps and validation sets that have
+	// been seeded. It can be pre-provided to provide specific revisions
+	// and validation-set sequences.
+	Manifest *Manifest
+	// ManifestPath if set, specifies the file path where the
+	// seed.manifest file should be written.
+	ManifestPath string
+}
+
+// manifest returns either the manifest already provided by the
+// options, or if not provided, returns a newly initialized manifest.
+func (opts *Options) manifest() *Manifest {
+	if opts.Manifest == nil {
+		return NewManifest()
+	}
+	return opts.Manifest
 }
 
 // OptionsSnap represents an options-referred snap with its option values.
@@ -184,6 +202,7 @@ type Writer struct {
 
 	availableSnaps  *naming.SnapSet
 	availableByMode map[string]*naming.SnapSet
+	byModeSnaps     map[string][]*SeedSnap
 
 	// toDownload tracks which set of snaps SnapsToDownload should compute
 	// next
@@ -199,6 +218,11 @@ type Writer struct {
 	systemSnap                      *SeedSnap
 	kernelSnap                      *SeedSnap
 	noKernelSnap                    bool
+	// manifest is the manifest of the seed used to track
+	// seeded snaps and validation-sets. It may either be
+	// initialized from the one provided in options, or it
+	// may be initialized to a new copy.
+	manifest *Manifest
 }
 
 type policy interface {
@@ -214,8 +238,6 @@ type policy interface {
 
 	checkBase(s *snap.Info, modes []string, availableByMode map[string]*naming.SnapSet) error
 
-	checkAvailable(snpRef naming.SnapRef, modes []string, availableByMode map[string]*naming.SnapSet) bool
-
 	checkClassicSnap(sn *SeedSnap) error
 
 	needsImplicitSnaps(availableByMode map[string]*naming.SnapSet) (bool, error)
@@ -252,6 +274,7 @@ func New(model *asserts.Model, opts *Options) (*Writer, error) {
 
 		byNameOptSnaps:  naming.NewSnapSet(nil),
 		byRefLocalSnaps: naming.NewSnapSet(nil),
+		manifest:        opts.manifest(),
 	}
 
 	var treeImpl tree
@@ -443,31 +466,82 @@ func IsSytemDirectoryExistsError(err error) bool {
 	return ok
 }
 
-// Start starts the seed writing. It creates a RefAssertsFetcher using
-// newFetcher and uses it to fetch model related assertions. For convenience it
-// returns the fetcher possibly for use to fetch seed snap assertions, a task
-// that the writer delegates as well as snap downloading. The writer assumes
-// that the snap assertions will end up in the given db (writing assertion
-// database). When the system seed directory is already present,
+func (w *Writer) validationSetFromManifest(vsm *asserts.ModelValidationSet) *ManifestValidationSet {
+	for _, vs := range w.manifest.AllowedValidationSets() {
+		if vs.AccountID == vsm.AccountID && vs.Name == vsm.Name {
+			return vs
+		}
+	}
+	return nil
+}
+
+// finalValidationSetAtSequence returns the final AtSequence for an
+// validation set. If any restrictions have been set in the manifest
+// then we must use the sequence and pinning status from that instead
+// of whats set in the model.
+func (w *Writer) finalValidationSetAtSequence(vsm *asserts.ModelValidationSet) (*asserts.AtSequence, error) {
+	atSeq := vsm.AtSequence()
+
+	// Check the manifest for a matching entry, to handle any restrictions that
+	// might have been setup.
+	vs := w.validationSetFromManifest(vsm)
+	if vs == nil {
+		return atSeq, nil
+	}
+
+	// If the model has the validation-set pinned, this can't be
+	// changed by the manifest.
+	if vsm.Sequence > 0 && vs.Sequence != vsm.Sequence {
+		// It's pinned by the model, then the sequence must match
+		return nil, fmt.Errorf("cannot use sequence %d of %q: model requires sequence %d",
+			vs.Sequence, vs.Unique(), vsm.Sequence)
+	}
+
+	// If the model does not have a sequence set, then we don't allow
+	// pinning through the manifest.
+	if vsm.Sequence <= 0 && vs.Pinned {
+		return nil, fmt.Errorf("pinning of %q is not allowed by the model", vs.Unique())
+	}
+
+	atSeq.Sequence = vs.Sequence
+	return atSeq, nil
+}
+
+func (w *Writer) fetchValidationSets(f SeedAssertionFetcher) error {
+	for _, vs := range w.model.ValidationSets() {
+		atSeq, err := w.finalValidationSetAtSequence(vs)
+		if err != nil {
+			return err
+		}
+		if err := f.FetchSequence(atSeq); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+// Start starts the seed writing, and fetches the necessary model assertions using
+// the provided SeedAssertionFetcher (See MakeSeedAssertionFetcher). The provided
+// fetcher must support the FetchSequence in case the model refers to any validation
+// sets. The seed-writer assumes that the snap assertions will end up in the given db
+// (writing assertions database). When the system seed directory is already present,
 // SystemAlreadyExistsError is returned.
-func (w *Writer) Start(db asserts.RODatabase, newFetcher NewFetcherFunc) (RefAssertsFetcher, error) {
+func (w *Writer) Start(db asserts.RODatabase, f SeedAssertionFetcher) error {
 	if err := w.checkStep(startStep); err != nil {
-		return nil, err
+		return err
 	}
 	if db == nil {
-		return nil, fmt.Errorf("internal error: Writer *asserts.RODatabase is nil")
+		return fmt.Errorf("internal error: Writer *asserts.RODatabase is nil")
 	}
-	if newFetcher == nil {
-		return nil, fmt.Errorf("internal error: Writer newFetcherFunc is nil")
+	if f == nil {
+		return fmt.Errorf("internal error: Writer fetcher is nil")
 	}
 	w.db = db
 
-	f := MakeRefAssertsFetcher(newFetcher)
-
 	if err := f.Save(w.model); err != nil {
 		const msg = "cannot fetch and check prerequisites for the model assertion: %v"
 		if !w.opts.TestSkipCopyUnverifiedModel {
-			return nil, fmt.Errorf(msg, err)
+			return fmt.Errorf(msg, err)
 		}
 		// Some naive tests including ubuntu-image ones use
 		// unverified models
@@ -480,18 +554,23 @@ func (w *Writer) Start(db asserts.RODatabase, newFetcher NewFetcherFunc) (RefAss
 		err := snapasserts.FetchStore(f, w.model.Store())
 		if err != nil {
 			if nfe, ok := err.(*asserts.NotFoundError); !ok || nfe.Type != asserts.StoreType {
-				return nil, err
+				return err
 			}
 		}
 	}
 
+	// fetch model validation sets if any
+	if err := w.fetchValidationSets(f); err != nil {
+		return err
+	}
+
 	w.modelRefs = f.Refs()
 
 	if err := w.tree.mkFixedDirs(); err != nil {
-		return nil, err
+		return err
 	}
 
-	return f, nil
+	return nil
 }
 
 // LocalSnaps returns a list of seed snaps that are local.  The writer
@@ -620,6 +699,11 @@ func (w *Writer) SetRedirectChannel(sn *SeedSnap, redirectChannel string) error
 
 }
 
+// Manifest returns the manifest for the current seed.
+func (w *Writer) Manifest() *Manifest {
+	return w.manifest
+}
+
 // snapsToDownloadSet indicates which set of snaps SnapsToDownload should compute
 type snapsToDownloadSet int
 
@@ -921,7 +1005,6 @@ func (w *Writer) considerForSnapdCarrying(sn *SeedSnap) {
 		return
 	}
 	w.noKernelSnap = true
-	return
 }
 
 // An AssertsFetchFunc should fetch appropriate assertions for the snap sn, it
@@ -1007,6 +1090,7 @@ func (w *Writer) downloaded(seedSnaps []*SeedSnap, fetchAsserts AssertsFetchFunc
 		w.availableSnaps = naming.NewSnapSet(nil)
 		w.availableByMode = make(map[string]*naming.SnapSet)
 		w.availableByMode["run"] = naming.NewSnapSet(nil)
+		w.byModeSnaps = make(map[string][]*SeedSnap)
 	}
 
 	for _, sn := range seedSnaps {
@@ -1020,7 +1104,11 @@ func (w *Writer) downloaded(seedSnaps []*SeedSnap, fetchAsserts AssertsFetchFunc
 				byMode = naming.NewSnapSet(nil)
 				w.availableByMode[mode] = byMode
 			}
+			if byMode.Contains(sn) {
+				continue
+			}
 			byMode.Add(sn)
+			w.byModeSnaps[mode] = append(w.byModeSnaps[mode], sn)
 		}
 	}
 
@@ -1060,13 +1148,6 @@ func (w *Writer) downloaded(seedSnaps []*SeedSnap, fetchAsserts AssertsFetchFunc
 		if err := w.checkBase(info, modes); err != nil {
 			return err
 		}
-		// error about missing default providers
-		for dp := range snap.NeededDefaultProviders(info) {
-			if !w.policy.checkAvailable(naming.Snap(dp), modes, w.availableByMode) {
-				// TODO: have a way to ignore this issue on a snap by snap basis?
-				return fmt.Errorf("cannot use snap %q without its default content provider %q being added explicitly%s", info.SnapName(), dp, errorMsgForModesSuffix(modes))
-			}
-		}
 	}
 
 	return nil
@@ -1142,9 +1223,81 @@ func (w *Writer) Downloaded(fetchAsserts AssertsFetchFunc) (complete bool, err e
 	if err := w.ensureARefs(nil, fetchAsserts); err != nil {
 		return false, err
 	}
+
+	if err := w.checkPrereqs(); err != nil {
+		return false, err
+	}
+
 	return true, nil
 }
 
+func (w *Writer) checkPrereqs() error {
+	// as we error on the first problem we want to check snaps mode by mode
+	// in a fixed order; we start with run then
+	// ephemeral as snaps marked as such need to be self-contained
+	// then specific modes sorted
+	modes := make([]string, 0, len(w.availableByMode))
+	modes = append(modes, "run")
+	fixed := 1
+	if _, ok := w.availableByMode["ephemeral"]; ok {
+		modes = append(modes, "ephemeral")
+		fixed = 2
+	}
+	for m := range w.availableByMode {
+		if m == "run" || m == "ephemeral" {
+			continue
+		}
+		modes = append(modes, m)
+	}
+	sort.Strings(modes[fixed:])
+	for _, m := range modes {
+		if err := w.checkPrereqsInMode(m); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+func (w *Writer) checkPrereqsInMode(mode string) error {
+	nmode := len(w.byModeSnaps[mode])
+	nephemeral := len(w.byModeSnaps["ephemeral"])
+	var snaps []*snap.Info
+	if mode != "run" && mode != "ephemeral" {
+		// mode is a concrete ephemeral mode
+		// (not run, not ephemeral which is the set of snaps
+		//  shared by all ephemeral modes)
+		// so we include snap marked for any ephemeral mode
+		snaps = make([]*snap.Info, 0, nmode+nephemeral)
+		for _, sn := range w.byModeSnaps["ephemeral"] {
+			snaps = append(snaps, sn.Info)
+		}
+	} else {
+		snaps = make([]*snap.Info, 0, nmode)
+	}
+	for _, sn := range w.byModeSnaps[mode] {
+		snaps = append(snaps, sn.Info)
+	}
+	warns, errs := snap.ValidateBasesAndProviders(snaps)
+	if errs != nil {
+		var errPrefix string
+		// XXX TODO: return an error that subsumes all the errors
+		if mode == "run" {
+			errPrefix = "prerequisites need to be added explicitly"
+		} else {
+			errPrefix = fmt.Sprintf("prerequisites need to be added explicitly for relevant mode %s", mode)
+		}
+		return fmt.Errorf("%s: %v", errPrefix, errs[0])
+	}
+	wfmt := "%v"
+	if mode != "run" {
+		wfmt = fmt.Sprintf("prerequisites for mode %s: %%v", mode)
+	}
+	for _, warn := range warns {
+		w.warningf(wfmt, warn)
+	}
+	return nil
+}
+
 func (w *Writer) checkPublisher(sn *SeedSnap) error {
 	if sn.local && sn.aRefs == nil {
 		// nothing to do
@@ -1191,6 +1344,97 @@ func (w *Writer) Warnings() []string {
 	return w.warnings
 }
 
+func (w *Writer) resolveValidationSetAssertion(seq *asserts.AtSequence) (asserts.Assertion, error) {
+	if seq.Sequence <= 0 {
+		hdrs, err := asserts.HeadersFromSequenceKey(seq.Type, seq.SequenceKey)
+		if err != nil {
+			return nil, err
+		}
+		return w.db.FindSequence(seq.Type, hdrs, -1, seq.Type.MaxSupportedFormat())
+	}
+	return seq.Resolve(w.db.Find)
+}
+
+func (w *Writer) validationSetAsserts() (map[*asserts.AtSequence]*asserts.ValidationSet, error) {
+	vsAsserts := make(map[*asserts.AtSequence]*asserts.ValidationSet)
+	vss := w.model.ValidationSets()
+	for _, vs := range vss {
+		atSeq, err := w.finalValidationSetAtSequence(vs)
+		if err != nil {
+			return nil, fmt.Errorf("internal error: %v", err)
+		}
+		a, err := w.resolveValidationSetAssertion(atSeq)
+		if err != nil {
+			return nil, fmt.Errorf("internal error: cannot resolve validation-set: %v", err)
+		}
+		vsAsserts[atSeq] = a.(*asserts.ValidationSet)
+	}
+	return vsAsserts, nil
+}
+
+func (w *Writer) validationSets() (*snapasserts.ValidationSets, error) {
+	vss, err := w.validationSetAsserts()
+	if err != nil {
+		return nil, err
+	}
+
+	valsets := snapasserts.NewValidationSets()
+	for _, vs := range vss {
+		valsets.Add(vs)
+	}
+	return valsets, nil
+}
+
+func (w *Writer) installedSnaps() []*snapasserts.InstalledSnap {
+	installedSnap := func(snap *SeedSnap) *snapasserts.InstalledSnap {
+		return snapasserts.NewInstalledSnap(snap.SnapName(), snap.ID(), snap.Info.Revision)
+	}
+
+	var installedSnaps []*snapasserts.InstalledSnap
+	for _, sn := range w.snapsFromModel {
+		installedSnaps = append(installedSnaps, installedSnap(sn))
+	}
+	for _, sn := range w.extraSnaps {
+		installedSnaps = append(installedSnaps, installedSnap(sn))
+	}
+	return installedSnaps
+}
+
+func (w *Writer) checkStepCompleted(step writerStep) bool {
+	// expectedStep is the next step it needs to perform. If that
+	// is higher (as they are int based values), then the step we
+	// are checking against has completed.
+	return w.expectedStep > step
+}
+
+// CheckValidationSets validates all snaps that are to be seeded against any
+// specified validation set. Info for all seed snaps must have been derived prior
+// to this call.
+func (w *Writer) CheckValidationSets() error {
+	// It makes no sense to check validation-sets before all required snaps
+	// have been resolved and downloaded. Ensure that this is not called before
+	// the Downloaded step has completed.
+	if !w.checkStepCompleted(downloadedStep) {
+		return fmt.Errorf("internal error: seedwriter.Writer cannot check validation-sets before Downloaded signaled complete")
+	}
+
+	valsets, err := w.validationSets()
+	if err != nil {
+		return err
+	}
+
+	// Check for validation set conflicts first, then we check all
+	// the seeded snaps against them
+	if err := valsets.Conflict(); err != nil {
+		return err
+	}
+
+	// Make one aggregated list of snaps we are seeding, then check that
+	// against the validation sets
+	installedSnaps := w.installedSnaps()
+	return valsets.CheckInstalledSnaps(installedSnaps, nil)
+}
+
 // SeedSnaps checks seed snaps and copies local snaps into the seed using copySnap.
 func (w *Writer) SeedSnaps(copySnap func(name, src, dst string) error) error {
 	if err := w.checkStep(seedSnapsStep); err != nil {
@@ -1230,6 +1474,11 @@ func (w *Writer) SeedSnaps(copySnap func(name, src, dst string) error) error {
 				// record final destination path
 				sn.Path = dst
 			}
+			if !info.Revision.Unset() {
+				if err := w.manifest.MarkSnapRevisionSeeded(sn.Info.SnapName(), sn.Info.Revision); err != nil {
+					return fmt.Errorf("cannot record snap for manifest: %s", err)
+				}
+			}
 		}
 		return nil
 	}
@@ -1244,12 +1493,36 @@ func (w *Writer) SeedSnaps(copySnap func(name, src, dst string) error) error {
 	return nil
 }
 
+func (w *Writer) markValidationSetsSeeded() error {
+	vsm, err := w.validationSetAsserts()
+	if err != nil {
+		return err
+	}
+	for seq, vs := range vsm {
+		if err := w.manifest.MarkValidationSetSeeded(vs, seq.Pinned); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
 // WriteMeta writes seed metadata and assertions into the seed.
 func (w *Writer) WriteMeta() error {
 	if err := w.checkStep(writeMetaStep); err != nil {
 		return err
 	}
 
+	if w.opts.ManifestPath != "" {
+		// Mark validation sets seeded in the manifest if the options
+		// are set to produce a manifest.
+		if err := w.markValidationSetsSeeded(); err != nil {
+			return err
+		}
+		if err := w.manifest.Write(w.opts.ManifestPath); err != nil {
+			return err
+		}
+	}
+
 	snapsFromModel := w.snapsFromModel
 	extraSnaps := w.extraSnaps
 
@@ -1263,7 +1536,7 @@ func (w *Writer) WriteMeta() error {
 // query accessors
 
 func (w *Writer) checkSnapsAccessor() error {
-	if w.expectedStep < seedSnapsStep {
+	if !w.checkStepCompleted(downloadedStep) {
 		return fmt.Errorf("internal error: seedwriter.Writer cannot query seed snaps before Downloaded signaled complete")
 	}
 	return nil
